/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2002, 2003, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with epsilon; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. */

/* This module provides the powerful GNU-style (i.e. POSIX with GNU extensions)
   command-line parsing facility , but does not necessarily require GNU libc.
   As such it may be preferrable for portability (especially towards the free
   operating systems derived from BSD, which I would like to support as
   secondary targets, right after GNU systems). */

#ifndef EAM_COMMAND_LINE_H_
#define EAM_COMMAND_LINE_H_

#include "common_text.h"

/* To do: test for unicity of short and long options
   To do: forbid long options beginning with "no-" */

/* Functions to set strings to be shown on --help or --version.
   By default all strings are set to "". */
void eam_set_program_name (char *string);
void eam_set_synopsis_string (char *string);
void eam_set_general_help_message (char *string);
void eam_set_version_string (char *string);
void eam_set_copyright_string (char *string);
void eam_set_license_message (char *string);
void eam_set_bug_reporting_message (char *string);

/* Functions to define options. All string parameters must be non-NULL. 
   The options -h / --help and -V / --version are automatically defined. 
   For long options "" means 'no long name'.
   For short options '\0' means 'no short name'.
   The description parameter is used for --help.
   Don't write the "--" for long options or the "-" for short options;
   they are automatically inserted at the beginning.

   Toggle options have always either value (char*)0 or (char*)1. */
void eam_add_toggle_option (const char *long_option,
                        const char short_name,
                        const long initial_value,
                        const char *description);
void eam_add_option_with_nonoptional_parameter (const char *long_name,
                                            const char short_name,
                                            const char *initial_value,
                                            const char *description);
void eam_add_option_with_optional_parameter (const char *long_name,
                                         const char *initial_value,
                                         const char *value_with_no_parameter,
                                         const char *description);

/* Call this one to parse the arguments in argv. Note that it modifies both
   argc and argv, eliminating all options and leaving only non-options, in
   the same order they were passed in. --help and --version are managed here.*/
void eam_parse_command_line (int *argc, char **argv);

/* Functions to access the values passed via options. These can be called
   only after eam_parse_command_line(). */
char *eam_short_option_to_value (char name);
char *eam_long_option_to_value (char *name);

/* This function shows the value of all options. It's useful for debugging. */
void eam_show_options_status (void);

/* To do: this should be moved somewhere else? */
/* Output a message including the program name to stderr, then exit with
   failure: the arguments are the same as for printf(): */
void eam_fatal(const char* format, ...) __attribute__((noreturn));

/* Show the message, suggest to use --help and exit with failure. The
   arguments are the same as for printf(): */
void eam_command_line_fatal(const char* format, ...);

/* Usually you don't need to call these directly: */
void eam_show_help (void);          /* The effect of passing --help */
void eam_show_version (void);       /* The effect of passing --version */

#endif /* #ifndef EAM_COMMAND_LINE_H_ */
