/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include <stdio.h>
#include <assert.h>
#include "doubly_linked_list.h"
#include "malloc.h"

/* List header: the user always uses this to refer a whole list or to begin 
   scanning it, from either end. */
struct eam_doubly_linked_list{
  /* Pointer to the first and last element. They can be NULL. */
  struct eam_doubly_linked_list_element *eam_first_element;
  struct eam_doubly_linked_list_element *eam_last_element;
  
  /* The lenght of the list; this allows to compute length in O(1).  */
  int length;
}; // struct

struct eam_doubly_linked_list_element* eam_first_element(const struct eam_doubly_linked_list *list){
  return list->eam_first_element;
}

struct eam_doubly_linked_list_element* eam_last_element(const struct eam_doubly_linked_list *list){
  return list->eam_last_element;
}

/* Each list element: */
struct eam_doubly_linked_list_element{
  /* Pointers to the previous and the next elements. They can be NULL. */
  struct eam_doubly_linked_list_element *previous;
  struct eam_doubly_linked_list_element *next;
  
  /* Pointer to the datum. It can be NULL: */
  void *datum;
}; // struct

struct eam_doubly_linked_list_element* eam_previous(const struct eam_doubly_linked_list_element *e){
  return e->previous;
}
struct eam_doubly_linked_list_element* eam_next(const struct eam_doubly_linked_list_element *e){
  return e->next;
}
void* eam_datum(const struct eam_doubly_linked_list_element *e){
  return e->datum;
}

struct eam_doubly_linked_list* eam_create_doubly_linked_list(void){
  /* Make a new header. No elements yet: */
  struct eam_doubly_linked_list *new_header = 
    (struct eam_doubly_linked_list*)
    eam_xmalloc(sizeof(struct eam_doubly_linked_list));
  new_header->eam_first_element = NULL;
  new_header->eam_last_element = NULL;
  new_header->length = 0;

  /* Return the header we built: */
  return new_header;
}

int eam_doubly_linked_list_length(const struct eam_doubly_linked_list *list){
  return list->length;
}

int eam_is_doubly_linked_list_empty(const struct eam_doubly_linked_list *list){
  return list->length == 0;
}

void eam_append_to_doubly_linked_list(void *datum,
                                  struct eam_doubly_linked_list *list){
  /* Make a new element and set its pointers: */
  struct eam_doubly_linked_list_element *new_element = 
    (struct eam_doubly_linked_list_element*)
    eam_xmalloc(sizeof(struct eam_doubly_linked_list_element));
  new_element->datum = datum;
  new_element->next = NULL;
  new_element->previous = list->eam_last_element;
  
  /* Make the old extreme element (if any) point to the new element: */
  if(list->eam_last_element != NULL)
    list->eam_last_element->next = new_element;
  
  /* Update pointers to extreme elements and element count in the header: */
  if(list->eam_first_element == NULL)
    list->eam_first_element = new_element;
  list->eam_last_element = new_element;
  list->length++;
}

void eam_prepend_to_doubly_linked_list(void *datum,
                                   struct eam_doubly_linked_list *list){
  /* Make a new element and set its pointers: */
  struct eam_doubly_linked_list_element *new_element = 
    (struct eam_doubly_linked_list_element*)
    eam_xmalloc(sizeof(struct eam_doubly_linked_list_element));
  new_element->datum = datum;
  new_element->next = list->eam_first_element;
  new_element->previous = NULL;
  
  /* Make the old extreme element (if any) point to the new element: */
  if(list->eam_first_element != NULL)
    list->eam_first_element->previous = new_element;
  
  /* Update pointers to extreme elements and element count in the header: */
  if(list->eam_last_element == NULL)
    list->eam_last_element = new_element;
  list->eam_first_element = new_element;
  list->length++;
}

static void destroy_doubly_linked_list(struct eam_doubly_linked_list *list,
                                       const int destroy_elements)
  __attribute__(( always_inline ));
static void destroy_doubly_linked_list(struct eam_doubly_linked_list *list,
                                       const int destroy_elements){
  /* Destroy each element, and also the data if requested: */
  struct eam_doubly_linked_list_element *p;
  for(p = list->eam_first_element; p != NULL; ){
    struct eam_doubly_linked_list_element *saved_next = p->next;
    if(destroy_elements)
      free(p->datum);
    free(p);
    p = saved_next;
  } // for
  
  /* Destroy the header: */
  free(list);
}

void eam_destroy_doubly_linked_list_structure(struct eam_doubly_linked_list *list){
  destroy_doubly_linked_list(list, 0);
}

void eam_destroy_whole_doubly_linked_list(struct eam_doubly_linked_list *list){
  destroy_doubly_linked_list(list, 1);
}

void* eam_detach_first_from_doubly_linked_list(struct eam_doubly_linked_list *list){
  /* Check for non emptyness only when debugging. This code must be fast. */
#ifdef ENABLE_ASSERTIONS
  assert(list->length != 0);
#endif
  
  /* Keep a pointer to the datum to be returned and to the list element we'll
     need to destroy: */
  struct eam_doubly_linked_list_element *element = list->eam_first_element;
  void *datum = element->datum;
  
  /* Update the header and the element next to the one we are detaching: */
  list->eam_first_element = element->next;
  if(element->next == NULL)
    list->eam_last_element = NULL;
  else
    element->next->previous = NULL;
  list->length--;
  
  /* Destroy the element, which will not be reachable anymore: */
  free(element);
  
  /* Return the pointer we kept: */
  return datum;
}

void* eam_detach_last_from_doubly_linked_list(struct eam_doubly_linked_list *list){
  /* Check for non emptyness only when debugging. This code must be fast. */
#ifdef ENABLE_ASSERTIONS
  assert(list->length != 0);
#endif
  
  /* Keep a pointer to the datum to be returned and to the list element we'll
     need to destroy: */
  struct eam_doubly_linked_list_element *element = list->eam_last_element;
  void *datum = element->datum;
  
  /* Update the header and the element next to the one we are detaching: */
  list->eam_last_element = element->previous;
  if(element->previous == NULL)
    list->eam_first_element = NULL;
  else
    element->previous->next = NULL;
  list->length--;
  
  /* Destroy the element, which will not be reachable anymore: */
  free(element);
  
  /* Return the pointer we kept: */
  return datum;
}

struct eam_doubly_linked_list* eam_concatenate_doubly_linked_lists(
    struct eam_doubly_linked_list *list_1,
    struct eam_doubly_linked_list *list_2){
#ifdef ENABLE_ASSERTIONS
  /* Check that list_1 and list_2 are distinct: */
  assert(list_1 != list_2);
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* Make the header for the new list: */
  struct eam_doubly_linked_list *new_header = 
    (struct eam_doubly_linked_list*)
    eam_xmalloc(sizeof(struct eam_doubly_linked_list));
  if(list_1->length != 0)
    new_header->eam_first_element = list_1->eam_first_element;
  else
    new_header->eam_first_element = list_2->eam_first_element;
  if(list_2->length != 0)
    new_header->eam_last_element = list_2->eam_last_element;
  else
    new_header->eam_last_element = list_1->eam_last_element;
  new_header->length = list_1->length + list_2->length;
  
  /* Update the pointers linking the last element of list_1 and the first
     one of list_2: */
  if(list_2->eam_first_element != NULL)
    list_2->eam_first_element->previous = list_1->eam_last_element;
  if(list_1->eam_last_element != NULL)
    list_1->eam_last_element->next = list_2->eam_first_element;  
  
  /* Destroy the old headers: */
  free(list_1);
  free(list_2);
  
  /* Return the header for the new list: */
  return new_header;
}

/* A test. */
#if 0
#define LOOP
#include <stdio.h>

struct eam_doubly_linked_list* make_interval(const int from, const int to){
  int i;
  struct eam_doubly_linked_list *r = eam_create_doubly_linked_list();
  for(i = from; i <= to; i++){
    int *datum = (int*)eam_xmalloc(sizeof(int));
    *datum = i;    
    eam_append_to_doubly_linked_list(datum, r);
  }
  //return r;
  eam_append_to_doubly_linked_list(NULL, r);

  eam_prepend_to_doubly_linked_list(NULL, r);
  void *p = eam_detach_last_from_doubly_linked_list(r);
  //printf("[%p]\n", p);
  p = eam_detach_first_from_doubly_linked_list(r);
  //printf("[%p]\n", p);

  return eam_concatenate_doubly_linked_lists(
            eam_create_doubly_linked_list(),
            eam_concatenate_doubly_linked_lists(r,
                                            eam_create_doubly_linked_list()));
}

int main(){
  struct eam_doubly_linked_list *list;  
  struct eam_doubly_linked_list_element *p;
#ifdef LOOP
 q:
#endif
  list = eam_concatenate_doubly_linked_lists(make_interval(1, 10),
                                         make_interval(11, 20));
    //make_interval(1, 10);
  for(p = list->eam_first_element; p != NULL; p = eam_next(p))
    printf("%i ", *(int*)(eam_datum(p)));
  printf("(length %i)\n", eam_doubly_linked_list_length(list));
  for(p = list->eam_last_element; p != NULL; p = eam_previous(p))
    printf("%i ", *(int*)(eam_datum(p)));
  printf("(length %i)\n", eam_doubly_linked_list_length(list));
  
  eam_destroy_whole_doubly_linked_list(list);
#ifdef LOOP
  goto q;
#endif
  return 0;
}

#endif
