/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef EAM_DOUBLY_LINKED_LIST_H_
#define EAM_DOUBLY_LINKED_LIST_H_

/* This compilation unit supports doubly linked list operations using a
   variant of Lisp-style "conses", which *point* to data (with generic
   void* pointers) instead of including them in a struct together with
   'next' and 'previous' pointers.
   In contrast with the Lisp style, here list operations are destructive,
   i.e. modify the pointer structure of the list header and elements (even if
   they never touch data). 
   Circular lists are not allowed. */

/* List header: the user always uses this to refer a whole list or to begin 
   scanning it, from either end. It's an opaque data structure for the user: */
struct eam_doubly_linked_list;

/* Selectors, useful for beginning a list scan; note that they can return NULL: */
struct eam_doubly_linked_list_element* eam_first_element(const struct eam_doubly_linked_list *list)
  __attribute__ (( pure ));
struct eam_doubly_linked_list_element* eam_last_element(const struct eam_doubly_linked_list *list)
  __attribute__ (( pure ));

/* Each list element. It's an opaque data structure for the user: */
struct eam_doubly_linked_list_element;

/* Selectors: note that eam_previous() and eam_next() return NULL when the whole list
   has been scanned. Of course eam_datum() returns NULL if the datum is null. All
   selectors are O(1). */
struct eam_doubly_linked_list_element* eam_previous(const struct eam_doubly_linked_list_element *e)
  __attribute__ (( pure ));
struct eam_doubly_linked_list_element* eam_next(const struct eam_doubly_linked_list_element *e)
  __attribute__ (( pure ));
void* eam_datum(const struct eam_doubly_linked_list_element *e)
  __attribute__ (( pure ));

/* Create a new empty doubly-linked list: */
struct eam_doubly_linked_list* eam_create_doubly_linked_list(void)
  __attribute__ (( malloc ));

/* Return the size of the given doubly-linked-list. O(1). */
int eam_doubly_linked_list_length(const struct eam_doubly_linked_list *list)
  __attribute__ (( pure ));

/* Return a nonzero value iff the list is empty. O(1). */
int eam_is_doubly_linked_list_empty(const struct eam_doubly_linked_list *list)
  __attribute__ (( pure ));

/* These destructively update the list header and elements. O(1). */
void eam_append_to_doubly_linked_list(void *datum,
                                  struct eam_doubly_linked_list *list);
void eam_prepend_to_doubly_linked_list(void *datum,
                                   struct eam_doubly_linked_list *list);

/* These destructively updates the list header and elements (an element is
   also destroyed). Executing these operations on an empty list is checked
   for (and a fatal error is issued in that case) *only* when ENABLE_ASSERTIONS is
   defined, for efficiency's sake. The relevant datum is returned, and as
   it's always the case with data it's the user responsibility to destroy it
   when needed. O(1). */
void* eam_detach_first_from_doubly_linked_list(struct eam_doubly_linked_list *list);
void* eam_detach_last_from_doubly_linked_list(struct eam_doubly_linked_list *list);

/* Return a new list with the elements of list_1 followed by the elements of
   list_2. Note that both list_1 and list_2 are invalidated by this operation,
   which destroys their headers. list_1 and list_2 must be distinct. O(1). */
struct eam_doubly_linked_list* eam_concatenate_doubly_linked_lists(
    struct eam_doubly_linked_list *list_1,
    struct eam_doubly_linked_list *list_2)
  __attribute__ (( malloc ));

/* Destroy the header and all the elements of the given doubly-linked list.
   Data are not touched. This is useful for lists where *all* the data are
   shared with other structures. */
void eam_destroy_doubly_linked_list_structure(struct eam_doubly_linked_list *list);

/* Destroy the header, all the elements and all the data of the given
   doubly-linked list (which must have all been malloc()ed). No other
   data-dependant finalization is performed. */
void eam_destroy_whole_doubly_linked_list(struct eam_doubly_linked_list *list);

#endif // #ifndef EAM_DOUBLY_LINKED_LIST_H_
