/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include "list.h"
#include "filesystem.h"
#include "types.h"
#include "malloc.h"
#include "command_line.h"

int eam_exists_and_is_readable(const char *file_name){
  /* Try to open the file: */
  FILE *f = fopen(file_name, "r");
  if(f == NULL)
    /* Opening the file for reading failed: */
    return 0; 
  else{
    /* Ok, opening the file succeeded. We can close the file and tell the
       caller it's readable: */
    fclose(f);
    return 1;
  } // else
}

int eam_exists_and_is_readable_and_is_not_older_than(const char *file_name,
						 const char *other_file_name){
  if(! eam_exists_and_is_readable(other_file_name)){
    fprintf(stderr, "About the file \"%s\":\n", other_file_name);
    eam_fatal("other_file_name does not exist");
  } // if
  return (eam_exists_and_is_readable(file_name) &&
	  (eam_file_name_to_last_modification_time(file_name) >=
	   eam_file_name_to_last_modification_time(other_file_name)));
}

/* This implementation may be crude, but it's enough for now. I can write
   something more elaborate if performance requires it: */
void eam_remove_directory_and_all_its_content(const char *directory){
  /* Remember the working directory; we need it because we're about to
     chdir(): */
  char *working_directory = eam_get_working_directory();
  
  /* Enter the directory to be deleted and destroy everything it contains: */
  chdir(directory);
  system("rm -rf -- * .[^.]*");
  
  /* Go back to the previous working directory, destroy its name which we
     saved before, and finally remove the directory to be destroyed, which now
     should be empty: */
  chdir(working_directory);
  free(working_directory);
  const int rmdir_exit_code = rmdir(directory);
  assert(rmdir_exit_code == 0);
  //fprintf(stderr, "[Removed the directory %s]\n", directory);
}

/* The directories to be destroyed at exit() time: */
static struct eam_list *directories_to_destroy = NULL;

/* The function to be called at exit() time, destroying all temporary
   directories with their content: */
static void destroy_all_temporary_directories(void){
  /* We don't bother destroying the list structure or the pointed strings; we're
     exiting now, and deallocating would only be a waste of time. */
  while(! eam_null(directories_to_destroy)){
    fprintf(stderr, "Deleting the temporary directory %s...\n",
	    (char*)eam_car(directories_to_destroy));
    
    /* Destroy the first referred directory: */
    eam_remove_directory_and_all_its_content((char*)eam_car(directories_to_destroy));
    
    /* Go on with the rest of the list: */
    directories_to_destroy = eam_cdr(directories_to_destroy);
  } // while
}

char* eam_create_temporary_directory_within(const char *parent_directory,
                                            const int destroy_at_exit_time){
  /* Create a new buffer holding a template for name of the directory we are
     about to create: */
  char *directory_template = (char*)eam_xmalloc(strlen(parent_directory) + 100);
  sprintf(directory_template, "%s/eam-temporary-XXXXXX", parent_directory);
  
  /* Make the directory, or return NULL if that's not possible: */
  if(mkdtemp(directory_template) == NULL) // this overwrites directory_template
    return NULL;
  
  /* Ok, we succeeded in creating the directory. Now remember we have to
     destroy it at exit() time: */
  static int did_we_call_atexit = 0; // trick to call atexit() exactly once
  if(destroy_at_exit_time){
    directories_to_destroy = eam_cons(directory_template, directories_to_destroy);
    if(! did_we_call_atexit){
      atexit(destroy_all_temporary_directories);
      did_we_call_atexit = 1;
    } // if
  } // if
  
  /* Return the name of the temporary directory: */
  return directory_template;
}

char* eam_create_temporary_directory(const int destroy_at_exit_time){
  /* Decide *where* to create the temporary directory: first check whether the
     TMPDIR environmnt variable is defined; otherwise default to /tmp: */
  const char *parent_directory = getenv("TMPDIR");
  if(parent_directory == NULL)
    parent_directory = "/tmp";
  
  return eam_create_temporary_directory_within(parent_directory,
                                               destroy_at_exit_time);
}

struct eam_list* eam_parse_search_path(const char *search_path){
  /* Start making the list empty: */
  struct eam_list *r = eam_nil;
  
  /* Start from the end of the search path, looking for path separators,
     so that we can build the list with cons respecting the order: */
  int i;
  int index_past_the_end_of_the_current_directory = strlen(search_path);
  for(i = index_past_the_end_of_the_current_directory - 1; i >= -1; i--)
    if((search_path[i] == ':') || (i == -1)){ // i == -1 marks the first name
      /* Ok, we found a directory name, starting at i + 1 and ending at
	 index_past_the_end_of_the_current_directory - 1: make a copy of the
	 name and cons it to the list; also update
	 index_past_the_end_of_the_current_directory, so that we remember
	 where the *next* directory name we found ends: */
      const int length = index_past_the_end_of_the_current_directory - i - 1;
      char *directory = (char*)eam_xmalloc(length + 1);
      strncpy(directory, search_path + i + 1, length);
      directory[length] = '\0';
      //fprintf(stderr, "[%s]\n", directory);
      r = eam_cons(directory, r);
      index_past_the_end_of_the_current_directory = i;
    } // if
  
  /* Return the list we built: */
  return r;
}

void eam_destroy_search_path(struct eam_list *search_path){
  eam_destroy_whole_list(search_path);
}

static char* locate_relative_in_search_path(const char *relative_file_name,
					    const struct eam_list *search_path){
  /* Look at each directory in the search path: */
  const struct eam_list *rest_of_the_search_path = search_path;
  const size_t relative_file_name_length = (int)strlen(relative_file_name);
  while(! eam_null(rest_of_the_search_path)){
    /* Check for an hypothetical file in the first directory: */
    const char *directory = (char*)eam_car(rest_of_the_search_path);
    char *full_path_name = (char*)eam_xmalloc(strlen(directory) + 
					 relative_file_name_length +
					 2);
    sprintf(full_path_name, "%s/%s", directory, relative_file_name);
    //fprintf(stderr, "[Trying with %s...]\n", full_path_name);
    
    /* If the hypothetical file exists and is readable then return its name;
       else destroy the name, and go on looking at the next directory: */
    if(eam_exists_and_is_readable(full_path_name))
      return full_path_name;
    else
      free(full_path_name);
    
    /* Work with the rest of the search path: */
    rest_of_the_search_path = eam_cdr(rest_of_the_search_path);
  } // while
  
  /* If we arrived here then we didn't find the file: */
  return NULL;
}

char* eam_locate_in_search_path(const char *file_name,
			    const struct eam_list *search_path){
  /* If the given pathname is absolute then just return a copy of it,
     else search for it in the search path: */
  if(eam_is_path_name_absolute(file_name)){
    char *copy = (char*)eam_xmalloc(strlen(file_name) + 1);
    strcpy(copy, file_name);
    return copy;
  }
  else
    return locate_relative_in_search_path(file_name, search_path);
}

FILE* eam_fopen_in_search_path(const char *relative_file_name,
			   const char *open_type,
			   const struct eam_list *search_path){
  /* Obtain the name of the first matching file, if any; fail immediately
     if no file matches: */
  char *full_path_name = eam_locate_in_search_path(relative_file_name, search_path);
  if(full_path_name == NULL)
    return NULL;

  /* If we got here then the file exists and is at least readable; fopen() it
     for the caller and destroy its name, which the caller should not see.
     Note that it's possible that we return NULL now; the caller should check
     in every case: */
  FILE *r = fopen(full_path_name, open_type);
  free(full_path_name);
  return r;
}

/* Adapted from the GNU libc manual. This function implements the GNU libc
   getcwd() behaviour, which is a nice extension to the standard, also on 
   non-GNU systems: */
char* eam_get_working_directory(){
  size_t size = 100;
  
  while(1){
    char *buffer = (char*)eam_xmalloc(size);
    if(getcwd(buffer, size) == buffer)
      return buffer;
    free(buffer);
    if(errno != ERANGE)
      eam_fatal("getcwd() failed with ERANGE");
    size *= 2;
  } // while
}

char* eam_file_name_to_extension(const char *file_name){
  const size_t length = strlen(file_name);
  int i;
  for(i = length - 1; i >= 1; i--) // a leading dot doesn't mark an extension...
    if(file_name[i] == '.'){
      /* We found the beginning of the (last) extension: */
      const size_t extension_length = length - i - 1;
      char *r = (char*)eam_xmalloc(extension_length + 1);
      strncpy(r, file_name + i + 1, extension_length);
      return r;
    }
  
  /* If we arrived here then the given file name has no extension: */
  return NULL;
}

char* eam_replace_extension(const char *file_name, const char *new_extension){
  const size_t length = strlen(file_name);
  const size_t new_extension_length = strlen(new_extension);
  int i, extension_beginning_index = -1;
  for(i = length - 1; i >= 0; i--)
    if(file_name[i] == '.'){
      /* We found the beginning of the (last) extension: */
      extension_beginning_index = i + 1;
      break;
    } // if
  
  /* Make a new large enough string; if file_name has no extension then just 
     stick the given extension to it into the new string; else concatenate the
     first part of file_name with new_extension into the new string: */
  char *r;
  if(extension_beginning_index == -1){ // no extension
    r = (char*)eam_xmalloc(length + new_extension_length + 2);
    sprintf(r, "%s.%s", file_name, new_extension);
  }
  else{
    r = (char*)eam_xmalloc(i + new_extension_length + 2);
    strncpy(r, file_name, i + 1);
    strcpy(r + i + 1, new_extension);
  } // else
  
  /* Return the string we built: */
  return r;
}

time_t eam_file_name_to_last_modification_time(const char *file_name){
  struct stat stats;
  
  /* Retrieve data about the file: */
  if(stat(file_name, &stats) != 0){
    fprintf(stderr, "About the file %s:\n", file_name);
    eam_fatal("stat() failed. Probably the file does not exist or isn't readable");
  } // if
  else
    /* Return just the part we're interested in: */
    return stats.st_mtime;
}

int eam_is_path_name_relative(const char *path_name){
  return ! eam_is_path_name_absolute(path_name);
}

int eam_is_path_name_absolute(const char *path_name){
  if(strlen(path_name) < 1){
    fprintf(stderr, "is_pathname_relative(): about \"%s\"\n", path_name);
    eam_fatal("the given pathname is too short");
  } // if
  
  return (path_name[0] == '/') || (path_name[0] == '~');
}

char* eam_path_name_to_file_part(const char *path_name){
  const size_t length = strlen(path_name);
  /* Particular cases: the "." and "~" pathnames: */
  if(length == 1)
    if((path_name[0] == '.') || (path_name[0] == '~')){
      char *r = (char*)eam_xmalloc(1);
      r[0] = '\0';
      return r;
    } // inner if
  
  int i;
  for(i = length - 1; i >= 0; i--)
    if(path_name[i] == '/'){
      /* We found the end of the directory part: */
      const size_t file_name_length = length - i - 1;
      char *r = (char*)eam_xmalloc(file_name_length + 1);
      strcpy(r, path_name + i + 1);
      return r;
    } // if
  
  /* If the first character is '~' and we're here then the entire pathname is
     a home directory such as "~john": */
  if(length > 0)
    if(path_name[0] == '~'){
      char *r = (char*)eam_xmalloc(1);
      r[0] = '\0';
      return r;
    } // inner if
  
  /* If we arrived here then there is no directory part: the whole string is
     a file name: */
#ifdef ENABLE_ASSERTIONS
  assert(i == -1);
#endif // #ifdef ENABLE_ASSERTIONS
  char *r = (char*)eam_xmalloc(length + 1);
  strcpy(r, path_name);
  return r;
}

char* eam_path_name_to_directory_part(const char *path_name){
  const size_t length = strlen(path_name);
  /* Particular cases: the "." and "~" pathnames: */
  if(length == 1)
    if((path_name[0] == '.') || (path_name[0] == '~')){
      char *r = (char*)eam_xmalloc(2);
      strcpy(r, path_name);
      return r;
    } // inner if

  int i;
  for(i = length - 1; i >= 0; i--)
    if(path_name[i] == '/'){
      /* Particular case: "/": */
      if(i == 0){
        char *r = (char*)eam_xmalloc(2);
        strcpy(r, "/");
        return r;
      }
      else{
        /* We found the end of the directory part: */
        const size_t directory_name_length = i;
        char *r = (char*)eam_xmalloc(directory_name_length + 1);
        strncpy(r, path_name, directory_name_length);
        r[directory_name_length] = '\0';
        return r;
      } // else
    } // if
#ifdef ENABLE_ASSERTIONS
  assert(i == -1);
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* If the first character is '~' and we're here then the entire pathname is
     a home directory such as "~john": */
  if(length > 0)
    if(path_name[0] == '~'){
      char *r = (char*)eam_xmalloc(length + 1);
      strcpy(r, path_name);
      return r;
    } // inner if
  
  /* If we arrived here then there is no directory part: the whole string is
     a file name */
  char *r = (char*)eam_xmalloc(2);
  strcpy(r, ".");
  return r;
}

/* Return the length of the given file, or -1 on failure: */
long eam_file_length(const char *path_name){
  struct stat s;
  if(stat(path_name, &s) != 0)
    return -1;
  else
    return (long)s.st_size;
}

/* Return a new malloc()ed string filled with the whole content of the given
   file, or NULL in case of errors. Of course this shouldn't be used with
   very large files whose content doesn't fit in core. */
char* eam_get_the_full_content_of(const char *path_name){
  /* Get the file size, so that we know how long a string we need to make;
     if this fails, we also fail: */
  long file_length = eam_file_length(path_name);
  if(file_length == -1)
    return NULL;
  
  /* Open the file; if this fails we also fail; otherwise create the string: */
  FILE *f = fopen(path_name, "r");
  if(f == NULL)
    return NULL;
  char *string = eam_xmalloc(file_length);
  
  /* Read the string: if we fail now deallocate and close before failing: */
  if(fread(string, (size_t)1, (size_t)file_length, f) != file_length){
    free(string);
    fclose(f);
    return NULL;
  }
  /* Of, everything went smooth if we arrived here. Finalize and return our
     shiny new string: */
  fclose(f);
  return string;
}


#if 0
int main(int argc, char **argv){
  fprintf(stderr, "%i\n", 
	  (int)eam_file_name_to_last_modification_time("/tmp/k"));

  return 0;

  int i;
  for(i = 0; i < 1; i++){
    char *directory = eam_create_temporary_directory();

    //printf("Created the directory %s\n", directory);
  }

  struct eam_list *search_path = eam_parse_search_path(getenv("PATH"));
  const char *file_name = argv[1];//"emacs";
  char *path_name = eam_locate_in_search_path(file_name, search_path);
  if(path_name == NULL)
    fprintf(stderr, "[%s was not found]\n", file_name);
  else
    fprintf(stderr, "%s\n", path_name);
  
  fprintf(stderr, "{%s}\n", eam_file_name_to_extension("test.jpg"));
  fprintf(stderr, "{%s}\n", eam_replace_extension("test.gif", "png"));
  
  return 0;
}
#endif
