/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2003 Luca Saiu
Minor updates by Luca Saiu, 2006

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with epsilon; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. */


/* A simple and efficient implementation of a stack of pointers made with
   macros. In the garbage-collector speed is crucial. */

#ifndef EAM_STACK_H_
#define EAM_STACK_H_

#include "types.h"

/* To be used in .c files: */
#define EAM_DEFINE_STACK(name) \
  word_t* name ## _stack; \
  integer_t name ## _stack_size; \
  integer_t name ## _stack_elements;

/* To be used in headers: */
#define EAM_DECLARE_STACK(name) \
  extern word_t* name ## _stack; \
  extern integer_t name ## _stack_size; \
  extern integer_t name ## _stack_elements;

/* Allocate storage and initialize index: */
#define eam_initialize_stack(name) { \
  name ## _stack_elements = 0; \
  name ## _stack_size = EAM_INITIAL_STACK_SIZE; \
  name ## _stack = (word_t*) eam_xmalloc(sizeof(word_t) * EAM_INITIAL_STACK_SIZE); \
}

/* Free storage: */
#define eam_destroy_stack(name) \
  free(name ## _stack)

/* Push a pointer onto a stack: */
#define eam_push_onto_stack(pointer, name) \
  (EAM_PUSH_ONTO_STACK(pointer, name ## _stack, name ## _stack_size, name ## _stack_elements))

/* Pop pointer from a stack: */
#define eam_pop_from_stack(name) \
  (EAM_POP_FROM_STACK(name ## _stack, name ## _stack_elements))

/* Returns the top of a stack, without popping: */
#define eam_top_of_stack(name) \
  (name ## _stack [name ## _stack_elements - 1])

/* Returns 0 iff the stack is empty: */
#define eam_is_stack_empty(name) \
  EAM_IS_STACK_EMPTY(name ## _stack_elements)  

/* Return the number of the elements in the given stack: */
#define eam_stack_elements_no(name) \
  ((integer_t)(name ## _stack_elements))

/* The following part should not be needed to the user: */


#define EAM_INITIAL_STACK_SIZE 1024

#define EAM_PUSH_ONTO_STACK(pointer, stack, stack_size, stack_elements) { \
  /* Enlarge the GC stack if needed: */ \
  \
  if(stack_elements == stack_size){ \
    /*fprintf(stderr, "<STACK GROWS TO "INTEGER_T_FORMAT" ELEMENTS>", stack_elements);*/ \
    stack_size *= 2; \
    stack = (word_t*) \
      realloc(stack, sizeof(word_t) * stack_size); \
    if(stack == NULL) \
      eam_fatal("could not enlarge stack"); \
  } \
  \
  /* Push element: */ \
  stack[stack_elements++] = pointer; \
}

#define EAM_POP_FROM_STACK(stack, stack_elements) \
  (stack[--stack_elements])

#define EAM_IS_STACK_EMPTY(stack_elements) \
  (stack_elements == 0)

#endif /* #ifndef EAM_STACK_H_ */
