/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2002, 2003, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with epsilon; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. */

#include <stdio.h>
#include <string.h>
#include "string_map.h"
#include "malloc.h"

#define EAM_STRING_MAP_DEFAULT_SIZE 512

struct eam_sm_element
{
  char *key;
  void *value;

  struct eam_sm_element *next;
};

struct eam_string_map
{
  struct eam_sm_element **buckets;
  size_t size;
  size_t size_of_value;
  eam_string_map_datum_finalizer datum_finalizer;
};

static inline int
string_hash_function (const char *x)
{
  long i, r = 0;
  for (i = 0; x[i] != '\0'; i++)
    r += x[i] * (i + 1);

  return r;
}

static void
destroy_sm_bucket (struct eam_sm_element *b,
		   const eam_string_map_datum_finalizer datum_finalizer)
{
  struct eam_sm_element *next;
  struct eam_sm_element *current = b;

  while (current != NULL)
    {
      datum_finalizer(current->value);
      //fprintf(stderr,"%s ",current->key);
      next = current->next;
      free (current->key);
      free (current->value);
      free (current);
      current = next;
    }				/* while */
}

struct eam_string_map *
eam_create_string_map (size_t size_of_value,
                       const eam_string_map_datum_finalizer datum_finalizer_)
{
  return eam_create_string_map_of_size(size_of_value,
				   EAM_STRING_MAP_DEFAULT_SIZE,
				   datum_finalizer_);
}

struct eam_string_map *
eam_create_string_map_of_size (size_t size_of_value_,
			   const int size_,
			   const eam_string_map_datum_finalizer datum_finalizer_)
{
  int i;
  struct eam_string_map *r =
    (struct eam_string_map *) eam_xmalloc (sizeof (struct eam_string_map));

  r->size = size_;
  r->size_of_value = size_of_value_;
  r->buckets = (struct eam_sm_element **)
    eam_xmalloc(sizeof(struct eam_sm_element *) * size_);
  
  for (i = 0; i < size_; i++)
    r->buckets[i] = NULL;
  
  r->datum_finalizer = datum_finalizer_;
  
  return r;
}

void
eam_destroy_string_map (struct eam_string_map *m)
{
  int i;
  int fb = 0;
  for (i = 0; i < m->size; i++)
    {
      if (m->buckets[i] != NULL)
	fb++;
      destroy_sm_bucket (m->buckets[i], m->datum_finalizer);
    }

  free (m->buckets);
  free (m);
}

struct eam_sm_element *
search_in_sm_bucket (struct eam_sm_element *bucket, const char *k)
{
  struct eam_sm_element *x = bucket;

  //fprintf(stderr,"\tStart searching for >%s<...\n",k);
  while (x != NULL)
    {
      //fprintf(stderr,"\tSearching for >%s< in bucket list... \n",k);
      //fprintf(stderr,"reached >%s<.\n",x->key);
      if (!strcmp (x->key, k))
	{
	  //fprintf(stderr,"found >%s<.\n",x->key);
	  return x;		/* We have found k */
	}
      else
	x = x->next;
    }
  //fprintf(stderr,"\t>%s< not found\n",k);
  /* x is NULL and we have not found k: */
  return NULL;
}

void
show_bucket (struct eam_string_map *m, int bucket)
{
  struct eam_sm_element *x = m->buckets[bucket];

  while (x != NULL)
    {
      fprintf (stderr, "%s->%p ", x->key, x->value);
      x = x->next;
    }
}

struct eam_sm_element *
remove_from_bucket (struct eam_sm_element *bucket, const char *k,
		    const eam_string_map_datum_finalizer datum_finalizer)
{
  struct eam_sm_element *x = bucket;

  if (bucket == NULL)
    return 0;
  else if (!strcmp (bucket->key, k))
    {
      struct eam_sm_element *next = bucket->next;

      datum_finalizer(bucket->value);
      free (bucket->key);
      free (bucket->value);
      free (bucket);
      return next;
    }

  while (x->next != NULL)
    {
      if (!strcmp (x->next->key, k))
	{
	  struct eam_sm_element *next_of_next = x->next->next;

	  datum_finalizer(x->next->value);
	  free (x->next->key);
	  free (x->next->value);
	  free (x->next);
	  x->next = next_of_next;
	  return bucket;
	}
      x = x->next;
    }

  return bucket;
}

void
eam_insert_into_string_map (struct eam_string_map *m, const char *k, void *v)
{
  int bucketNo = string_hash_function (k) % m->size;

  if (search_in_sm_bucket (m->buckets[bucketNo], k) == NULL)
    eam_multinsert_into_string_map (m, k, v);
  else{
    fprintf(stderr,
	    "%s: WARNING: eam_insert_into_string_map(\"%s\") is *not* replacing\n",
	    __PRETTY_FUNCTION__, k);
  } // else
}

void
eam_multinsert_into_string_map (struct eam_string_map *m, const char *k, void *v)
{
  int bucketNo = string_hash_function (k) % m->size;

  struct eam_sm_element *e =
    (struct eam_sm_element *) eam_xmalloc (sizeof (struct eam_sm_element));

  e->key = (char*) eam_xmalloc (strlen(k) + 1);
  strcpy (e->key, k);

  e->value = eam_xmalloc (m->size_of_value);
  memcpy (e->value, v, m->size_of_value);

  e->next = m->buckets[bucketNo];
  m->buckets[bucketNo] = e;
}

void
eam_remove_from_string_map (struct eam_string_map *m, const char *k)
{
  int bucket_no = string_hash_function (k) % m->size;

  m->buckets[bucket_no] = remove_from_bucket (m->buckets[bucket_no], k, m->datum_finalizer);
  //fprintf(stderr,"eam_remove_from_string_map(): [");
  //show_bucket(m,bucket_no);
  //fprintf(stderr,"]\n");
}

void *
eam_access_string_map (struct eam_string_map *m, const char *k)
{
  int bucket_no = string_hash_function (k) % m->size;
  struct eam_sm_element *e;

  if ((e = search_in_sm_bucket (m->buckets[bucket_no], k)) == NULL)
    return NULL;
  else
    return e->value;
}

int eam_is_string_map_defined_on (struct eam_string_map *map, const char *key)
{
  return eam_access_string_map(map, key) != NULL;
}

void eam_default_string_map_datum_finalizer(const void *datum){
  /* Do nothing. */
}
