;(define last-used-label 0)
(load "utility.scm")

(define new-label
  (let ((last-used-label 0))
    (lambda ()
      (set! last-used-label (+ last-used-label 1))
      `(label ,last-used-label))))

(define code '()) ; the list is stored in reversed order
(define globals '())

(define primitives
  '(null? cons? number? boolean? string? symbol? atom? array? character?
    primitive? closure?
    string->symbol symbol->string number->string
    string-ref
    eq? =
    + - * / ; To do: check division by zero
    cons car cdr
    array-ref array-set! array->list list->array make-array
    display-character
))

(defmacro define-on-guile x
  `(define ,@x))

(define-on-guile default-display
  (if (defined? 'default-display)
      default-display
      display))
(define-on-guile display-character
  (lambda (c) (default-display c)))
(define-on-guile display-boolean
  (lambda (b) (display-string (if b "#t" "#f"))))
(define-on-guile character?
  char?)
(define-on-guile cons?
  pair?)
(define-on-guile atom?
  (lambda (x)
    (and (not (cons? x))
         (not (array? x)))))


(define-on-guile primitive?
  (lambda (x) (and (procedure? x)
                   (not (closure? x)))))
(define-on-guile /
  (lambda (x y)
    (inexact->exact (floor (/ x y)))))

(define (display-symbol s)
  (display-string (symbol->string s)))

(define (display-number n)
  (display-string (number->string n)))

(define (newline)
  (display-character #\newline))

(define (display-string-starting-at s i)
  (if (= i (string-length s))
      #t
      (begin
        (display-character (string-ref s i))
        (display-string-starting-at s (+ i 1)))))
(define (display-string s)
  (display-string-starting-at s 0))

(define (display-cons x)
  (if (null? x)
      #t
      (begin
        (display-character #\()
        (display (car x))
        (display-string " . ")
        (display (cdr x))
        (display-character #\)))))

(define (display-list-with-no-parentheses x)
  (if (null? x)
      #t
      (begin
        (display (car x))
        (cond ((cons? (cdr x))
               (begin
                 (display-character #\space)
                 (display-list-with-no-parentheses (cdr x))))
              ((null? (cdr x))
               #t)
              (#t
               (begin
                 (display-string " . ")
                 (display (cdr x))))))))
(define (display-list x)
  (display-character #\()
  (display-list-with-no-parentheses x)
  (display-character #\)))

(define (display-array x)
  (display-character #\#)
  (display-list (array->list x)))

(define (display x)
  (cond ((null? x)
         (display-string "()"))
        ((number? x)
         (display-number x))
        ((symbol? x)
         (display-symbol x))
        ((boolean? x)
         (display-boolean x))
        ((character? x)
         (display-character x))
        ((string? x)
         (display-string x))
        ((closure? x)
         (display-string "#<closure>"))
        ((primitive? x)
         (display-string "#<primitive>"))
        ((cons? x)
         (display-list x))
        ((array? x)
         (display-array x))
        (#t
         (display-string "#<unknown>"))))

(define (emit-code x)
  (set! code (cons x code)))

(define (emit-label label)
  (emit-code "label-")
  (emit-code label)
  (emit-code ":"))

(define (emit-code-newline)
  (set! code (cons "\n" code)))

(define (emit-global global)
  (set! globals (cons global globals)))


(define (cond-conditions->if conditions)
  (if (null? conditions)
      #f
      (cons 'if 
            (cons (caar conditions)
                  (cons (cadar conditions)
                        (cons (cond-conditions->if (cdr conditions))
                              '()))))))

(define (let-subforms->application bindings body)
  (let ((formals (map (lambda (pair) (car pair)) bindings))
        (actuals (map (lambda (pair) (cadr pair)) bindings)))
    (cons (cons 'lambda
                (cons formals
                      body))
          actuals)))

(define (list-arguments->cons arguments)
  (if (null? arguments)
      ''()
      (cons 'cons
            (cons (car arguments)
                  (cons (list-arguments->cons (cdr arguments)) 
                        '())))))

(define (and-clauses->if clauses)
  (if (null? clauses)
      #t
      (list 'if (car clauses) (and-clauses->if (cdr clauses)) #f)))

(define (or-clauses->if clauses)
  (if (null? clauses)
      #f
      (list 'if (car clauses) #t (or-clauses->if (cdr clauses)))))

(define (compile-boolean b)
  (if b
      (emit-code "  \"push\"(constant(\"boolean\", \"-1\")) # boolean constant #t\n")
      (emit-code "  \"push\"(constant(\"boolean\", \"0\")) # boolean constant #f\n")))

(define (compile-number n)
  (emit-code "  \"push\"(constant(\"integer\", \"")
  (emit-code n)
  (emit-code "\")) # numeric constant ")
  (emit-code n)
  (emit-code-newline))

(define (compile-string s)
  (emit-code "  \"push\"(constant(\"string\", \"")
  (emit-code s)
  (emit-code "\")) # string constant \"")
  (emit-code s)
  (emit-code "\"")
  (emit-code-newline))

(define (compile-variable v)
  (emit-code "  if \"is-unbound?\"(constant(\"string\", \"")
  (emit-code v)
  (emit-code "\")) goto fail: # fail if unbound\n")
  (emit-code "  \"push\"(\"lookup\"(\"symbol\"(constant(\"string\", \"")
  (emit-code v)
  (emit-code "\")))) # lookup variable ")
  (emit-code v)
  (emit-code-newline))

(define (compile-when condition then)
  (let ((then-label      (new-label))
        (end-of-if-label (new-label)))
    (compile-form condition)
    (emit-code "  if \"pop\"() goto ") (emit-label then-label) (emit-code-newline)
    (emit-code "  push constant(\"boolean\", \"0\") # one-way if condition is #f\n")
    (emit-code "  goto ") (emit-label end-of-if-label) (emit-code-newline)
    (emit-label then-label) (emit-code-newline)
    (compile-form then)
    (emit-label end-of-if-label) (emit-code-newline)))

(define (compile-if condition then else)
  (let ((then-label      (new-label))
        (end-of-if-label (new-label)))
    (compile-form condition)
    (emit-code "  if \"pop\"() goto ") (emit-label then-label) (emit-code-newline)
    (compile-form else)
    (emit-code "  goto ") (emit-label end-of-if-label) (emit-code-newline)
    (emit-label then-label) (emit-code-newline)
    (compile-form then)
    (emit-label end-of-if-label) (emit-code-newline)))

(define (compile-bind-formals formals)
  (if (null? formals)
      #t
      (begin
        (emit-code "  \"bind-local\"(\"symbol\"(constant(\"string\", \"")
        (emit-code (car formals))
        (emit-code "\")), \"pop\"())\n")
        (compile-bind-formals (cdr formals)))))

(define (compile-lambda formals body)
  (let ((lambda-label       (new-label))
        (after-lambda-label (new-label)))
    (emit-code "  goto ") (emit-label after-lambda-label) (emit-code-newline)
    (emit-label lambda-label) (emit-code-newline)
    (compile-bind-formals (reverse formals))
    (compile-forms body)
    (emit-code "  %result := \"pop\"() # keep the result\n")
    (emit-code "  %local-environment := \"pop\"() # restore local environment\n")
    (emit-code "  goto \"pop\"() # return\n")
    (emit-label after-lambda-label) (emit-code-newline)
    (emit-code "  \"push\"(\"make-closure\"(") 
    (emit-label lambda-label)
    (emit-code ", %local-environment)) # push a closure\n")))

(define (compile-application function arguments)
  (let ((return-label (new-label)))
    (emit-code "  \"push\"(")
    (emit-label return-label)
    (emit-code ") # save return address\n")
    (emit-code "  \"push\"(%local-environment) # save caller local environment\n")
    (emit-code "### Compile the arguments:\n")
    (map compile-form arguments)
    (emit-code "### Compile the function ") (emit-code function) (emit-code ":\n")
    (compile-form function)
    (emit-code "### Apply the function to the arguments:\n")
    (emit-code "  %local-environment := \"cdr\"(\"top\"())\n")
    (emit-code "  goto \"car\"(\"pop\"())\n")
    (emit-label return-label) (emit-code-newline)
    (emit-code "  \"push\"(%result) # push what the procedure returned\n")))

(define (compile-array elements length)
  (if (null? elements)
      (begin
        (emit-code "  \"push\"(\"make-array\"(constant(\"integer\", ")
        (emit-code length)
        (emit-code "))) # make an array of ")
        (emit-code length)
        (emit-code " elements\n"))
      (begin
        (compile-quoted (car elements) #f)
        (compile-array (cdr elements) length))))

(define (compile-forms forms)
  (if (null? (cdr forms))
      (compile-form (car forms))
      (begin
        (compile-form (car forms))
        (emit-code "  \"pop\"() # this is not the last statement in a sequence\n")
        (compile-forms (cdr forms)))))

(define (compile-form form)
  (emit-code "### Translation of ") (emit-code form) (emit-code-newline)
  (cond ((number? form)
         (compile-number form))
        ((boolean? form)
         (compile-boolean form))
        ((string? form)
         (compile-string form))
        ((symbol? form)
         (compile-variable form))
        ((array? form)
         (let ((array-elements (array->list form)))
           (compile-array array-elements (length array-elements))))
        ((eq? (car form) 'if)
         (if (null? (cdddr form))
             (compile-when (cadr form) (caddr form))
             (compile-if (cadr form) (caddr form) (cadddr form))))
        ((eq? (car form) 'quote)
         (compile-quoted (cadr form) #f))
        ((eq? (car form) 'quasiquote)
         (compile-quoted (cadr form) #t))
        ((eq? (car form) 'let)
         (compile-form (let-subforms->application (cadr form) (cddr form))))
        ((eq? (car form) 'cond)
         (compile-form (cond-conditions->if (cadr form))))
        ((eq? (car form) 'list)
         (compile-form (list-arguments->cons (cdr form))))
        ((eq? (car form) 'and)
         (compile-form (and-clauses->if (cdr form))))
        ((eq? (car form) 'or)
         (compile-form (or-clauses->if (cdr form))))
        ((eq? (car form) 'lambda)
         (compile-lambda (cadr form) (cddr form)))
        ((eq? (car form) 'begin)
         (compile-forms (cdr form)))
        (#t
         (compile-application (car form) (cdr form)))
;;           (#t
;;            (emit-code "### FATAL: don't know how to compile ")
;;            (emit-code form)
;;            (emit-code-newline)
;;            (exit -1))
))

(define (expand-define header definition)
  (list 'define
        (car header)
        (cons 'lambda (cons (cdr header) definition))))

(define (compile-toplevel-define name definition)
  (emit-global name)
  (emit-code "## Definition of the global ") (emit-code name) (emit-code "\n")
  (compile-form definition)
  (emit-code "  global(\"")
  (emit-code name)
  (emit-code"\") := \"pop\"()\n"))

(define (compile-toplevel-expression expression)
  (compile-form expression)
  (emit-code "  \"pop\"() # we don't need the result\n"))

(define (compile-toplevel-form form)
  (emit-code "### Translation of the toplevel form ") (emit-code form) (emit-code-newline)
  (cond ((or (atom? form)
             (array? form))
         (compile-toplevel-expression form))
        ((eq? (car form) 'define)
         (if (cons? (cadr form))
             (compile-toplevel-form (expand-define (cadr form) (cddr form)))
             (compile-toplevel-define (cadr form) (caddr form))))
        (#t
         (compile-toplevel-expression form))))

(define (compile-toplevel-forms forms)
  (map compile-toplevel-form forms))

(define (generate-header)
  (display "compilation-unit \"abc\"\n\n"))

(define (generate-globals)
  (map (lambda (g)
         (display "global \"")
         (display g)
         (display "\"\n"))
       globals)
  (newline))

(define (generate-error-handler)
  (emit-code "\n\nfail:\n")
  (compile-form '(display "Fatal error: fuck off"))
  (emit-code "  \"pop\"() # we don't need the result of this final display\n")
  (emit-code "  \"exit-with-failure\"()\n"))

(define (generate-code)
  (display "code\n\n")
;  (generate-error-handler)
  (map display 
       (reverse code))
  (display "\nend code\n\n"))

(define (generate-footer)
  (display "end compilation-unit\n"))

(define (generate)
  (generate-header)
  (generate-globals)
  (generate-code)
  (generate-footer))
           
;(compile 
;'(
;  (define x 10)
 ;(cond ((#f 10) (#f 20) (#f 30) (#t 40)))
 ;'(if #t 10 20)
 ;''(1 . 2)
 ;'(+ 1 2)
 ;'(let ((a 1) (b 2)) (display a) (emit-code-newline) (+ a b))
 ;'(lambda (x y) (* x y))
 ;'(lambda (x y) 1 2 x)
 ;'((lambda (x) 10) 20)
 ;'(lambda (x) 42)
 ;'(list ''1 2)
 ;'(f 10)
 ;'(lambda (x) 1)
 ;'(begin (lambda (x) x) 42)
 ;'(if (< 1 2) 'a)
 ;'#(10 20 a)
;))

;; (compile-toplevel-forms
;;  '(
;;    (define n 1)
;;    (define fact
;;      (lambda (n) (if (zero? n) 1 (* n (fact (- n 1))))))
;;    (define (fact n)
;;      (if (zero? n) 1 (* n (fact (- n 1)))))
;;    (define (newline)
;;      (display-character #\newline))
;;    42
;; ))

;(begin (compile-toplevel-form '1)
;       (generate-code))

(define (expand-quoted form)
  (cond ((number? form) form)
        ((boolean? form) form)
        ((character? form) form)
        ((null? form) '(empty-list-immediate))
        ((symbol? form) `(symbol-immediate ,(symbol->string form)))
        ((array? form) `(list->array ,(expand-quote (array->list form))))
        ((cons? form) `(cons ,(expand-quoted (car form))
                             ,(expand-quoted (cdr form))))))

(define (expand-quasiquoted form)
  (cond ((number? form) form)
        ((boolean? form) form)
        ((character? form) form)
        ((null? form) '(empty-list-immediate))
        ((symbol? form) `(symbol-immediate ,form))
        ((array? form) `(list->array ,(expand-quote (array->list form))))
        ((eq? (car form) 'unquote)
         (cadr form))
        (#t
         `(cons ,(expand-quasiquoted (car form))
                ,(expand-quasiquoted (cdr form))))))

(define (expand-array array-form)
  `(list->array 1 ; number of dimensions. Ugly, but needed for guile and R5RS
                ,(expand-quoted (array->list array-form))))

(define (compilation-error . optional-message)
  (display "Compilation error")
  (if (not (null? optional-message))
      (begin
        (display ": ")
        (display (car optional-message))))
  (newline)
  (exit -1))

(define temporary-registers-no 5)
(define temporary-registers
  '((temporary-register 0)
    (temporary-register 1)
    (temporary-register 2)
    (temporary-register 3)
    (temporary-register 4)))

(define maximum-used-temporary -1)
(define (first-not-in xs ys)
  (cond ((null? xs)
         #f) ; failure
        ((is-in? (car xs) ys)
         (first-not-in (cdr xs) ys))
        (#t
         (car xs))))

(define (minimum-not-in start xs)
  (if (is-in? start xs)
      (minimum-not-in (+ start 1) xs)
      start))

(define reset-maximum-used-temporary!
  (set! maximum-used-temporary -1))

(define (first-free-temporary alive-temps)
  (let ((index (minimum-not-in 0 alive-temps)))
    (if (< maximum-used-temporary index)
        (set! maximum-used-temporary index))
    `(temporary ,index)))


(define last-used-pseudo 0)
(define (new-pseudo)
  (set! last-used-pseudo (+ last-used-pseudo 1))
  `(pseudo ,last-used-pseudo))

(define (get-pseudo-allocator-state)
  last-used-pseudo)
(define (set-pseudo-allocator-state! x)
  (set! last-used-pseudo x))
(define (reset-pseudo-allocator-state!)
  (set! last-used-pseudo 0))


(define instructions
  '(()))
(define linearized-instructions
  '())
(define (push-context!)
  (set! instructions (cons '() instructions))
  (emit! '(begin-context)))
(define (pop-context!)
  (emit! '(end-context))
  (set! linearized-instructions
        ;(append (car instructions) linearized-instructions))
        (append linearized-instructions (car instructions)))
  (set! instructions (cdr instructions)))
(define (emit! instruction)
  (set-car! instructions (cons instruction (car instructions))))
(define (emit-instructions! . instructions)
  (map emit! instructions))

; To do: compile lambda out-of-order, using a queue

(define (compile-form form handler temps locals non-locals globals)
  ;(emit! `(begin-form ,form))
  (cond ((number? form)
         `(make-number ,form))
        ((boolean? form)
         `(global ,(if form "scheme-true" "scheme-false")))
        ((string? form)
         `(primitive "make-string" ,form))
        ((symbol? form)
         (cond ((is-in? form locals)
                `(access-local ,(index-of form locals)))
               ((is-in? form non-locals)
                `(access-non-local ,(index-of form non-locals)))
               ((is-in? form globals)
                `(global ,(symbol->string form)))
               ((is-in? form primitives)
                `(global ,(symbol->string form)))
               (#t
                (compilation-error `(unbound variable ,form)))))
        ((eq? (car form) 'symbol-immediate)
         `(symbol-immediate ,(cadr form)))
        ((eq? (car form) 'quote)
         (compile-form (expand-quoted (cadr form)) 
                       handler temps locals non-locals globals))
        ((eq? (car form) 'quasiquote)
         (compile-form (expand-quasiquoted (cadr form)) 
                       handler temps locals non-locals globals))
        ((eq? (car form) 'if)
         (let* ((then-label (new-label))
                (end-of-if-label (new-label))
                (target (new-pseudo)))
           (emit! `(if-goto ,(compile-form (cadr form) handler temps
                                       locals non-locals
                                       globals)
                     ,then-label))
           (emit! `(:= ,target 
                        ,(compile-form (cadddr form) handler temps locals
                                       non-locals globals)))
           (emit! `(goto ,end-of-if-label))
           (emit! `(label ,then-label))
           (emit! `(:= ,target 
                        ,(compile-form (caddr form) handler temps locals
                                       non-locals globals)))
           (emit! `(label ,end-of-if-label))
           target))
        ((eq? (car form) 'begin)
         (if (null? (cddr form))
             (compile-form (cadr form) handler temps locals non-locals
                           globals)
             (begin
               (emit! (compile-form (cadr form) handler temps locals non-locals
                                     globals))
               (compile-form (cons 'begin (cddr form)) handler temps locals
                             non-locals globals))))
        ((eq? (car form) 'lambda)
         (let ((procedure-label (new-label))
               (after-procedure-label (new-label))
               (saved-pseudo-allocator-state (get-pseudo-allocator-state)))
           (push-context!)
           (reset-pseudo-allocator-state!) ;;
           (emit! `(label ,procedure-label))
           (emit! `(:= result-register
                        ,(compile-form (cons 'begin (cddr form))
                                       handler
                                       empty-temps
                                       (cadr form) ; locals
                                       (append locals non-locals)
                                       globals)))
           (emit! `(return))
           (set-pseudo-allocator-state! saved-pseudo-allocator-state) ;;
           (pop-context!)
           `("make-closure"
             ,procedure-label
             ("make-nl" ,(length locals) ,(length non-locals))
             ,(length (cadr form))))) ; number of parameters
        (#t
         (compilation-error `(unrecognized form ,form))))
         
;  (emit! `(end-form ,form))
)

          
(define (output-linearized-instructions)
  ;; Pop all started context, adding their instructions to the linearized
  ;; instruction list:
  (while (not (null? (car instructions)))
         (pop-context!))
  ;; Now linearized-instructions contains the concatentation of all contexts,
  ;; reversed.
  (map (lambda (i)
         (cond ((eq? (car i) 'label)
                (display i)
                (newline))
               ((eq? (car i) 'begin-context)
                (display ""))
               ((eq? (car i) 'end-context)
                (display "\n"))
               (#t
                (display "    ")
                (display i)
                (newline))))
       (reverse linearized-instructions)))


(define empty-temps
  '())

(push-context!)

(emit!
 (compile-form '(lambda (x) (if 10 20 30) x)
               ;'(begin (lambda (x y) (if x 120 y)) (if 10 20 30))
               ;'(if 10 (if 20 30 40) (if 50 60 70))
               ;'(lambda (x y z) 1 2 y);(if x y z))
               '(label error)
               empty-temps
               '(a b)        ; locals
               '(nl-1 nl-2)  ; non-locals
               '(gl-1 gl-2)) ; globals
)

(pop-context!)
(output-linearized-instructions)
