/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2003, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with epsilon; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. */


#ifndef EAMGC_HOMOGENEOUS_H_
#define EAMGC_HOMOGENEOUS_H_

#include "gc.h"

struct eamgc_homogeneous_pool;

/* A possibly user-supplied function tracing an object of some fixed kind.
   Pointer-finding may be conservative, but of course *at least* all real
   pointers must be marked. Note that the object itself should not be
   traced by this function; all its direct children should. */
typedef void (*eamgc_homogeneous_trace_function_t)(const word_t);

typedef struct eamgc_homogeneous_page_header* eamgc_homogeneous_page_t;

/* Reasonable defaults for objects of the functions types above:
   trace_everything_function may be used for conservative pointer finding;
   trace_nothing_function may be used for leaf objects. */
extern const eamgc_homogeneous_trace_function_t
  eamgc_homogeneous_trace_nothing_function;    // trace nothing
extern const eamgc_homogeneous_trace_function_t
  eamgc_homogeneous_trace_everything_function; // trace every word


/* Create a new empty homogeneous page of the kind of the given pool. This
   function doesn't attach the page to the pool. */
eamgc_homogeneous_page_t eamgc_create_homogeneous_page(struct eamgc_homogeneous_pool *pool)
  __attribute__ (( malloc )); 

/* Free the storage for the homogeneous page p (if possible on this system),
   and de-register p as a homogeneous page. */
void eamgc_destroy_homogeneous_page(const eamgc_homogeneous_page_t p);

unsigned_integer_t eamgc_eamgc_homogeneous_page_to_object_size(const eamgc_homogeneous_page_t p)
  __attribute__ (( pure ));

/* Returns NULL on failure: */
word_t eamgc_allocate_from_homogeneous_page(const eamgc_homogeneous_page_t p)
  __attribute__ (( malloc ));

/* object must belong to p. Returns the GC bit as it was before: */
integer_t eamgc_mark_in_homogeneous_page(const word_t object, const eamgc_homogeneous_page_t p);

/* object must belong to p */
void eamgc_unmark_in_homogeneous_page(const word_t object, const eamgc_homogeneous_page_t p);

/* object must belong to p */
integer_t eamgc_is_marked_in_homogeneous_page(const word_t object,
                                           const eamgc_homogeneous_page_t p)
  __attribute__ (( pure ));

inline integer_t eamgc_is_homogeneous_page_full(const eamgc_homogeneous_page_t p)
  __attribute__ (( pure ));

/* This is only a computation on pointers; it is fast, but does *not* check
   whether object actually belongs to a homogeneous page. */
inline eamgc_homogeneous_page_t eamgc_object_to_hypothetical_homogeneous_page(
    const word_t object)
  __attribute__ (( const ));

/* Returns NULL if the object does not belong to an homogeneous page,
   else return its page: */
eamgc_homogeneous_page_t eamgc_object_to_actual_homogeneous_page(const word_t object)
  __attribute__ (( pure ));

/* Given an object from a homogeneous page (we don't actually check
   whether it's really a homogeneous page!) return its tag: */
inline unsigned_integer_t eamgc_homogeneous_object_to_tag(const word_t object)
  __attribute__ (( pure ));

/* Given an object from a homogeneous page (we don't actually check
   whether it's really a homogeneous page!) return its pool-relative
   data: */
inline word_t eamgc_homogeneous_object_to_data(const word_t object)
  __attribute__ (( pure ));

/* Return the data of the given homogeneous pool: */
inline word_t eamgc_homogeneous_pool_to_data(const struct eamgc_homogeneous_pool *pool)
  __attribute__(( pure ));

/* Return the object size, in words, of the given homogeneous pool: */
inline unsigned_integer_t eamgc_homogeneous_pool_to_object_size(const struct eamgc_homogeneous_pool *pool)
  __attribute__(( pure ));

/* Given an object from a homogeneous page (we don't actually check
   whether it's really a homogeneous page!) return its allocated size
   in words: */
inline unsigned_integer_t eamgc_homogeneous_object_to_size(const word_t object)
  __attribute__ (( pure ));

/* Given a homogeneous page (of course we don't actually check that
   it's really a homogeneous page) return the size (in words) of one
   of its objects: */
inline unsigned_integer_t eamgc_homogeneous_page_to_object_size(
    const eamgc_homogeneous_page_t page) __attribute__ (( pure ));

/* Given a homogeneous page (of course we don't actually check that
   it's really a homogeneous page) return the tag of its objects: */
inline unsigned_integer_t eamgc_homogeneous_page_to_tag(
    const eamgc_homogeneous_page_t page) __attribute__ (( pure ));

/* Given a homogeneous page (of course we don't actually check that
   it's really a homogeneous page) return the data of its objects: */
inline word_t eamgc_homogeneous_page_to_data(
    const eamgc_homogeneous_page_t page) __attribute__ (( pure ));

/* 'Reset' a possibly full homogeneous page, so that at the next allocation
   we start back considering the first object. All alive objects in the page
   must be already marked when this function is called. O(1). */
void eamgc_reset_homogeneous_page(const eamgc_homogeneous_page_t page);

/* Unmark all the objects in the homogeneous page. */
void eamgc_unmark_all_objects_in_homogeneous_page(const eamgc_homogeneous_page_t page);

/* Return a nonzero value iff all the objects in the given homogeneous page
   are *not* marked: */
integer_t eamgc_are_all_objects_unmarked_in_homogeneous_page(
   const eamgc_homogeneous_page_t page)
  __attribute__ (( pure ));

/* Call the appropriate trace function for object's children; object must 
   be an homogeneous object belonging to the given page (which is not checked
   for): */
void eamgc_call_trace_function_for_homogeneous_object_children(
   const word_t object,
   const eamgc_homogeneous_page_t page);

/* Write the given pattern over all words of the given homogeneous object.
   This should only be needed for debugging. */
extern const word_t eamgc_just_created_pattern;
extern const word_t eamgc_uninitialized_pattern;
extern const word_t eamgc_dead_pattern;
void eamgc_overwrite_homogeneous_object(const word_t object, 
                                     const word_t pattern);

/* Overwrite all unmarked objects in the page with the 'dead' pattern. Useful 
   for debugging. Of course this should be called only after a complete
   marking. */
void eamgc_overwrite_dead_objects_in_homogeneous_page(const eamgc_homogeneous_page_t p);

/* Dump all the words of a homogeneous object to the given open file. This
   should only be needed for debugging: */
void eamgc_dump_homogeneous_object(const word_t object, FILE *file);

/* Given an object size, return the number of objects of that size fitting in
   a single homogeneous page: */
inline integer_t eamgc_homogeneous_object_size_to_object_no(
    const unsigned_integer_t object_size)
  __attribute__ (( const ));

/* Return the total memory space (in words) used by homogeneous page payloads.
   Headers, GC bits and other data structures are not counted: */
inline unsigned_integer_t eamgc_get_homogeneous_heap_size(void)
  __attribute__ (( pure ));

/* Return the number of currently existing homogeneous pages: */
integer_t eamgc_get_homogeneous_pages_no(void) __attribute__(( pure ));

/* For every existing homogeneous page (from any pool) call the given function
   passing it the page as its only parameter. An easy way to iterate (higher
   order programming in C... I must be *way* too tired). */
void eamgc_call_for_every_homogeneous_page(
   void (*function)(eamgc_homogeneous_page_t p));

/* Same as above, either just for possibly nonfull pages, or just for full
   pages, or just for current pages: */
void eamgc_call_for_every_possibly_nonfull_homogeneous_page( // includes current pages
    void (*function)(eamgc_homogeneous_page_t p));
void eamgc_call_for_every_full_homogeneous_page(
    void (*function)(eamgc_homogeneous_page_t p));
void eamgc_call_for_every_current_homogeneous_page(
    void (*function)(eamgc_homogeneous_page_t p));

/* Initialize internal data structures needed to manage homogeneous pools. The
   user never needs to directly call this. */
void eamgc_initialize_homogeneous_objects_structures(void);

/* Finalize internal data structures needed to manage homogeneous pools. The
   user never needs to directly call this. */
void eamgc_finalize_homogeneous_objects_structures(void);

/* Convert a possibly internal pointer to an existing homogeneous object into
   a pointer to the first word of the same homogeneous object:  */
word_t eamgc_possibly_internal_pointer_to_homogeneous_object(
    const word_t possibly_internal_pointer);

/* A pool is an opaque data structure for the user: */
struct eamgc_homogeneous_pool;

/* Create a homogeneous pool with the given parameters. Note how no facility
   is currently provided for destroying pools. */
struct eamgc_homogeneous_pool* eamgc_create_homogeneous_pool(
    const unsigned_integer_t object_size, // in words
    const unsigned_integer_t tag,
    const word_t pool_data,
    const eamgc_homogeneous_trace_function_t trace_function)
  __attribute__(( malloc ));


/* Allocate a new object from the given pool. This may trigger a collection. */
word_t eamgc_allocate_from_homogeneous_pool(struct eamgc_homogeneous_pool *pool)
  __attribute__(( malloc ))
;

/* Unmark all the objects in all the homogeneous pages of the pool. */
void eamgc_unmark_all_objects_in_homogeneous_pool(struct eamgc_homogeneous_pool *pool);

/* Return a nonzero value iff all the objects in all the pages of the given
   homogeneous pool are *not* marked: */
integer_t eamgc_are_all_objects_unmarked_in_homogeneous_pool(
     const struct eamgc_homogeneous_pool *pool)
  __attribute__ (( pure ));

/* Go back considering all allocated space as potentially containing dead
   objects. This must be called only by the collector right after marking,
   so that no information about which objects are still alive is lost.
   O(1). */
void eamgc_reset_homogeneous_pool(struct eamgc_homogeneous_pool *pool);

/* Overwrite all unmarked objects in the page with the given pattern. Useful 
   for debugging. Pages are not moved from full to possibly non-full.
   Of course this should be called only after a complete marking.  */
void eamgc_overwrite_dead_objects_in_homogeneous_pool(
   struct eamgc_homogeneous_pool *pool,
   const word_t pattern);

/* For every existing pool call the given function passing it the pool
   as its only parameter. An easy way to iterate (higher order programming
   in C... I must be *way* too tired). */
void eamgc_call_for_every_homogeneous_pool(
    void (*function)(struct eamgc_homogeneous_pool *pool));

/* Return a pointer to a static doubly-linked list containing all the
   homogeneous pools. The user should not modify the returned object in
   any way: */
struct eam_doubly_linked_list* eamgc_get_all_homogeneous_pools(void)
  __attribute__ (( pure ));

#endif /* #ifndef EAMGC_HOMOGENEOUS_H_ */
