/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "roots.h"
#include "../../common/c_context.h"
#include "trace.h"
#include "../../common/list.h"
#include "../../common/stack.h"
#include "../../common/malloc.h"

/* What we have to remember about a root [heap root] buffer (we only deal 
   with buffers internally: single-word roots [heap roots] are identified 
   with buffers of length 1): */
struct eam_buffer_descriptor{
  /* The word where the root [heap root] begins: */
  word_t beginning;
  
  /* How many words it's long: */
  unsigned_integer_t length;
}; // struct

/* Lists of all the root buffers to trace at the beginning of each 
   collection: elements have type struct eam_buffer_descriptor*: */
static struct eam_list *root_buffers;

/* Lists of all the heap root buffers to trace at the beginning of each 
   collection: elements have type struct eam_buffer_descriptor*: */
static struct eam_list *heap_root_buffers;

/* Lists of all the root tracing functions to call at the beginning of
   each collection: elements have type eamgc_root_tracing_function_t: */
static struct eam_list *root_tracing_functions;

/* We need the declaration of tracing stacks here. We prefer not to put
   global declarations such as this in a header... */
//EAM_DECLARE_STACK(homogeneous)
//EAM_DECLARE_STACK(large)

/* Initialize data structures for the roots filesystem. Not for the user: */
void eamgc_initialize_roots_structures(void){
  /* Set all lists to nil; if this is not the first time the GC is initialized
     the lists may still point somewhere (dangling, of course): */
  root_buffers = eam_nil;
  heap_root_buffers = eam_nil;
  root_tracing_functions = eam_nil;
}

/* Initialize data structures for the roots filesystem. Not for the user: */
void eamgc_finalize_roots_structures(void){
  /* Destroy the lists of root_buffers, and their contents. It's safe because 
     they were just xmalloc()ed: */
  eam_destroy_whole_list(root_buffers);
  eam_destroy_whole_list(heap_root_buffers);
  
  /* The cars of this one instead are pointers to functions: */
  eam_destroy_list_structure(root_tracing_functions);
}

/* Only useful for debugging: */
static void dump_root_buffers(void){
  fprintf(stderr, "root_buffers is now:\n");
  struct eam_list *p;
  for(p = root_buffers; ! eam_null(p); p = eam_cdr(p)){
    fprintf(stderr, 
            "* [%p, %u]\n",
            (void*)((struct eam_buffer_descriptor*)eam_car(p))->beginning,
            (unsigned)((struct eam_buffer_descriptor*)eam_car(p))->length);
  } // for
  fprintf(stderr, "\n");
}


/* Register a user-defined root, to be traced at the beginning of every 
   collection. Note that a *pointer* to the beginning of the roots
   must be supplied. The scanned words will be:
   - *pointer_to_root_beginning
   - *((void*)pointer_to_root_beginning + 1 * sizeof(word_t))
   - *((void*)pointer_to_root_beginning + 2 * sizeof(word_t))
   - *((void*)pointer_to_root_beginning + 3 * sizeof(word_t))
   - ...
   root_length must be in words. */
void eamgc_register_buffer_of_roots(const word_t first_pointer_to_root,
                                 const unsigned_integer_t roots_no){
  /* Create a new buffer descriptor: */
  struct eam_buffer_descriptor *buffer = (struct eam_buffer_descriptor*)
    eam_xmalloc(sizeof(struct eam_buffer_descriptor));
  buffer->beginning = first_pointer_to_root;
  buffer->length = roots_no;
  
  /* Insert it into the list: */
  root_buffers = eam_cons(buffer, root_buffers);
}

/* Register a single pointer to the heap as a root. It's equivalent to
   eamgc_register_root_buffer() with 1u as the second parameter. The heap
   pointer which is traced is (the future value of) *root_pointer. */
void eamgc_register_root(const word_t root_pointer){
  eamgc_register_buffer_of_roots(root_pointer, (unsigned_integer_t)1u);
}

/* Register a buffer already in the heap as a root. The heap pointers
   which are traced start at root (not at *root, even if that is of
   course also traced). The length is in words. */
void eamgc_register_buffer_of_heap_roots(const word_t root_beginning,
                                      const unsigned_integer_t root_length){
  /* Create a new buffer descriptor: */
  struct eam_buffer_descriptor *buffer = (struct eam_buffer_descriptor*)
    eam_xmalloc(sizeof(struct eam_buffer_descriptor));
  buffer->beginning = root_beginning;
  buffer->length = root_length;
  
  /* Insert it into the list: */
  heap_root_buffers = eam_cons(buffer, heap_root_buffers);
}

/* Register a single pointer already in the heap as a root. The heap pointer
   which is traced is root (not only *root, even if that is of course
   also traced). This is equivalent to eamgc_register_buffer_of_heap_roots()
   with 1u as the second parameter: */
void eamgc_register_heap_root(const word_t root){
  eamgc_register_buffer_of_heap_roots(root, (unsigned_integer_t)1u);
}

/* Register a function so that it's called at the beginning of every
   collection to trace some user-defined roots. */
void eamgc_register_root_tracing_function(const eamgc_root_tracing_function_t f){
  /* Add the function pointer to the list: */
  root_tracing_functions = eam_cons(f, root_tracing_functions);
}

/* A pointer which may be loaded from memory byte-by-byte. This may be
   inefficient, but such a solution is actually needed for word-aligned
   architectures such as Sparc. */
union byte_by_byte_potential_pointer{
  char bytes[sizeof(word_t)]; // an array of bytes as large as a word
  word_t *pointer;            // the same word, as a pointer
}; // union
                                                                                
/* Trace as a root pointer every word between the two given boundaries
   (the second one is included only if the whole word fits in the closed
   interval). first_word must be less or equal to last_word -- This is
   essential when tracing C stacks, which may grow in either direction
   depending on the architecture. There are no requirements on
   alignment. */
static void trace_words_between(word_t first_word, word_t last_word){
  /* Try to trace every word as a potential pointer: */ 
  char* p;
  for(p = (char*)first_word; p <= (char*)last_word; p += sizeof(word_t)){
    /* Load *(word_t)p, byte by byte. See the comment before the union
       definition for the reason we have to do this. I'd like GCC to
       be able to replace the loop with a simple load on non
       word-aligned architectures, but it looks pretty hard to to,
       even in the future. */
    union byte_by_byte_potential_pointer star_p_byte_by_byte;
    int i;
    for(i = 0; i < sizeof(word_t); i++)
      star_p_byte_by_byte.bytes[i] = *(p + i); // set the i-th byte
                                                                                
    /* Ok, now we have *(word_t*)p, which might be a heap pointer. Try to
       trace it: */
    eamgc_trace_object(star_p_byte_by_byte.pointer);
  } // outer for
}                                                                                

/* Trace as a root pointer every word, for every possible alignment
   which is multiple of a byte and less or equal to a word, between
   the two given boundaries (included): this is useful for conservative
   pointer-finding where we can't be sure about pointer alignment.
   Note that boundaries are sorted internally (this accommodates for
   either stack growing direction): */
static void trace_words_with_any_alignment(const void* one_end,
                                           const void* other_end){
  /* Sort the ends: */
  void *first_word, *last_word;
  if(one_end <= other_end){
    first_word = (void*)one_end;
    last_word =  (void*)other_end;
  }
  else{
    first_word = (void*)other_end;
    last_word =  (void*)one_end;
  }

  /* For any possible alignment from 0 to (bytes per word) - 1...: */
  integer_t i;
  for(i = 0; i < (integer_t)sizeof(word_t); i++)
    /* Trace the whole buffer (save part of the first word): */
    trace_words_between((const word_t)(((const char*)first_word) + i),
                        (const word_t)last_word);
}

/* A root tracing function (which the user may register if he/she
   desires so) tracing roots from the machine registers and the C 
   stack. */
static void eamgc_trace_c_context_roots_(void){
  //fprintf(stderr, "[regs]... ");
  /* Trace pointers from (an image of) machine registers: */
  eam_call_with_first_register_and_last_register(trace_words_with_any_alignment);
  
  //fprintf(stderr, "done.\n[stack]... ");
  /* Trace pointers from the C stack: */
  eam_call_with_stack_bottom_and_stack_top(trace_words_with_any_alignment);
  //fprintf(stderr, "done.\n");
}
const eamgc_root_tracing_function_t eamgc_trace_c_context_roots = 
  eamgc_trace_c_context_roots_;

/* See the comment at the end of eamgc_trace_roots(): */
static void trace_homogeneous_pool_data(struct eamgc_homogeneous_pool *pool){
  const word_t data = eamgc_homogeneous_pool_to_data(pool);
#ifdef VERBOSE_DEBUG
  fprintf(stderr, "Tracing homogeneous pool data: %p\n", (void*)data);
#endif // #ifdef VERBOSE_DEBUG
  eamgc_trace_object(data);
}

/* See the comment at the end of eamgc_trace_roots(): */
static void trace_large_pool_data(struct eamgc_large_pool *pool){
  const word_t data = eamgc_large_pool_to_data(pool);
#ifdef VERBOSE_DEBUG
  fprintf(stderr, "Tracing large pool data: %p\n", (void*)data);
#endif // #ifdef VERBOSE_DEBUG
  eamgc_trace_object(data);
}

/* Trace all roots. This is called by the collector to start tracing. 
   Not for the user. */
void eamgc_trace_roots(void){
  const struct eam_list *p;
  integer_t i;
  
  /* Invoke all root tracing functions: */
  for(p = root_tracing_functions; ! eam_null(p); p = eam_cdr(p)){
    const eamgc_root_tracing_function_t f = eam_car(p);
    f();
  } // for
  
  /* Trace all roots: */
  //fprintf(stderr, "[rts] ");
  for(p = root_buffers; ! eam_null(p); p = eam_cdr(p)){
    const struct eam_buffer_descriptor *buffer_descriptor = eam_car(p);
    word_t *w = (word_t*)buffer_descriptor->beginning;
    for(i = 0; i < buffer_descriptor->length; i++, w++)
      eamgc_trace_object(*w);
  } // for
  
  /* Trace all heap roots: */
  //fprintf(stderr, "[hrts] ");
  for(p = heap_root_buffers; ! eam_null(p); p = eam_cdr(p)){
    const struct eam_buffer_descriptor *buffer_descriptor = eam_car(p);
    word_t *w = (word_t*)buffer_descriptor->beginning;
    for(i = 0; i < buffer_descriptor->length; i++, w++)
      eamgc_trace_object((word_t)w);
  } // for

  #ifdef VERBOSE_DEBUG
  fprintf(stderr, "(roots pointing the heap are %i) ", 
          (int)(eam_stack_elements_no(homogeneous) +
                eam_stack_elements_no(large)));
  #endif // #ifdef VERBOSE_DEBUG
  
  /* Trace all pool data. They are not registered as root because in the
     future we might allow to destroy pools at runtime, and that wouldn't
     be possible for objects containing roots. This is very cheap, only
     one pointer per *pool*: */
  eamgc_call_for_every_homogeneous_pool(trace_homogeneous_pool_data);
  eamgc_call_for_every_large_pool(trace_large_pool_data);
}
