/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdlib.h>
#include "../../common/types.h"
#include "homogeneous.h"
#include "large.h"
#include "statistics.h"
#include "compile_time_parameters.h"

/* This is useful to print statistics: */
static double words_to_megabytes(const unsigned_integer_t words){
  return words * sizeof(word_t) / 1024. / 1024.;
}


/* Size of the alive heap at the last GC, in words: */
static unsigned_integer_t alive_homogeneous_heap_size = 0u;
static unsigned_integer_t alive_large_heap_size = 0u;

/* Set the alive heap size. Called by the tracer while marking reachable 
   objects. The size is in words. */
inline void eamgc_zero_alive_heap_size(void){
  alive_homogeneous_heap_size = 0u;
  alive_large_heap_size = 0u;
}

/* Set the alive heap size. Called by the tracer while marking reachable 
   objects. The size is in words. */
inline void eamgc_increment_alive_homogeneous_heap_size(const unsigned_integer_t size_to_add){
  alive_homogeneous_heap_size += size_to_add;
}
inline void eamgc_increment_alive_large_heap_size(const unsigned_integer_t size_to_add){
  alive_large_heap_size += size_to_add;
}

/* Return the total number of words taken by the objects which were alive
   at the last collection: */
inline unsigned_integer_t eamgc_get_alive_heap_size(void){
  return eamgc_get_alive_homogeneous_heap_size() +
    eamgc_get_alive_large_heap_size();
}
inline unsigned_integer_t eamgc_get_alive_homogeneous_heap_size(void){
  return alive_homogeneous_heap_size;
}
inline unsigned_integer_t eamgc_get_alive_large_heap_size(void){
  return alive_large_heap_size;
}


static unsigned_integer_t minimum_homogeneous_heap_size =
  EAMGC_INITIAL_MINIMUM_HOMOGENEOUS_HEAP_SIZE;
unsigned_integer_t eamgc_get_minimum_homogeneous_heap_size(void){
  return minimum_homogeneous_heap_size;
}
void eamgc_set_minimum_homogeneous_heap_size(const unsigned_integer_t size){
  minimum_homogeneous_heap_size = size;
#ifdef ENABLE_ASSERTIONS
  fprintf(stderr, ">>>>>>>>>>> NEW MINIMUM HOMOGENEOUS HEAP SIZE: %.3gMb\n", words_to_megabytes(size));
#endif // #ifdef ENABLE_ASSERTIONS
}

static unsigned_integer_t maximum_homogeneous_heap_size =
  EAMGC_INITIAL_MAXIMUM_HOMOGENEOUS_HEAP_SIZE;
unsigned_integer_t eamgc_get_maximum_homogeneous_heap_size(void){
  return maximum_homogeneous_heap_size;
}
void eamgc_set_maximum_homogeneous_heap_size(const unsigned_integer_t size){
  maximum_homogeneous_heap_size = size;
}

static unsigned_integer_t maximum_large_heap_size =
  EAMGC_INITIAL_MAXIMUM_LARGE_HEAP_SIZE;
unsigned_integer_t eamgc_get_maximum_large_heap_size(void){
  return maximum_large_heap_size;
}
void eamgc_set_maximum_large_heap_size(const unsigned_integer_t size){
  maximum_large_heap_size = size;
}

unsigned_integer_t eamgc_get_heap_size(void){
  return eamgc_get_homogeneous_heap_size() +
         eamgc_get_large_heap_size();
}

/* Useful for debugging: */
void eamgc_dump_heap_size(void){
  const unsigned_integer_t homogeneous_heap_size = eamgc_get_homogeneous_heap_size();
  const unsigned_integer_t large_heap_size = eamgc_get_large_heap_size();
  const unsigned_integer_t total_heap_size = eamgc_get_heap_size();
  const unsigned_integer_t alive_homogeneous_heap_size =
    eamgc_get_alive_homogeneous_heap_size();
  const unsigned_integer_t alive_large_heap_size =
    eamgc_get_alive_large_heap_size();
  const unsigned_integer_t alive_heap_size =
    eamgc_get_alive_heap_size();
  fprintf(stderr, " (heap size is %.3gMb (%gMb homogeneous + %.3gMb large))\n",
          words_to_megabytes(total_heap_size),
          words_to_megabytes(homogeneous_heap_size),
          words_to_megabytes(large_heap_size));
  fprintf(stderr, "   (alive heap size is %.3gMb (%.3g%%) (%.3gMb homogeneous + %.3gMb large))\n",
          words_to_megabytes(alive_heap_size),
          100.0 * alive_heap_size / total_heap_size,
          words_to_megabytes(alive_homogeneous_heap_size),
          words_to_megabytes(alive_large_heap_size));
  int system_page_size = getpagesize();
  long int available_system_pages_no = sysconf(_SC_AVPHYS_PAGES);
  fprintf(stderr, "     (available physical memory is %.3gMb(%i %iKb pages))\n",
          (double)system_page_size * available_system_pages_no / 1024.0 / 1024.0,
          (int)available_system_pages_no,
          system_page_size / 1024);
}

/* Recompute statistics so that heuristics give more precise evaluations. This
   is a relatively costly procedure, thought to be called just once per GC; but
   what we compute here enable heuristics to be fast: */
void eamgc_recompute_statistics(void){
  /* Get the current heap size: */
  const unsigned_integer_t alive_homogeneous_heap_size =
    eamgc_get_alive_homogeneous_heap_size();  
  const unsigned_integer_t alive_large_heap_size =
    eamgc_get_alive_large_heap_size();  
  const unsigned_integer_t alive_heap_size __attribute__(( unused ))  =
    alive_homogeneous_heap_size + alive_large_heap_size;
  const unsigned_integer_t homogeneous_heap_size =
    eamgc_get_homogeneous_heap_size();  
  const unsigned_integer_t large_heap_size =
    eamgc_get_large_heap_size();
  const unsigned_integer_t heap_size =
    eamgc_get_heap_size();
  const double homogeneous_to_total_ratio __attribute__(( unused )) =
    (heap_size != 0) ?
    ((double)homogeneous_heap_size / (double)heap_size) :
    0.5;
  const double alive_homogeneous_to_total_alive_ratio = 
    (alive_heap_size != 0) ?
    ((double)alive_homogeneous_heap_size / (double)alive_heap_size) :
    0.5;
  
  /* Compute the optimal heap size we would get by keeping the alive/total
     ratio constant: */
  unsigned_integer_t best_homogeneous_heap_size = (unsigned_integer_t)
    ((double)alive_homogeneous_heap_size / (double)EAMGC_LIVE_TO_TOTAL_RATIO);
  unsigned_integer_t best_large_heap_size = (unsigned_integer_t)
    ((double)alive_large_heap_size / (double)EAMGC_LIVE_TO_TOTAL_RATIO);
  
#ifdef ENABLE_ASSERTIONS
  fprintf(stderr, "Current homogeneous heap size: %.3gMb\n",
          words_to_megabytes(homogeneous_heap_size));
  fprintf(stderr, "Current large heap size: %.3gMb\n",
          words_to_megabytes(large_heap_size));
  fprintf(stderr, "Best homogeneous heap size: %.3gMb\n",
          (double)alive_homogeneous_heap_size / (double)EAMGC_LIVE_TO_TOTAL_RATIO 
          * (double)(sizeof(word_t)) / 1024.0 / 1024.0);
  fprintf(stderr, "Best large heap size: %.3gMb\n",
          (double)alive_large_heap_size / (double)EAMGC_LIVE_TO_TOTAL_RATIO 
          * (double)(sizeof(word_t)) / 1024.0 / 1024.0);
#endif // #ifdef ENABLE_ASSERTIONS

  /* Ask the system how much we can still allocate before trashing: */
  const int system_page_size = getpagesize();
  const long int available_system_pages_no = sysconf(_SC_AVPHYS_PAGES);
  unsigned_integer_t still_available_size = (unsigned_integer_t)
    (available_system_pages_no * system_page_size / sizeof(word_t));
  
  /* Well, we prefer to have a safety margin: let's stop a bit *before*
     trashing... */
  if(still_available_size < EAMGC_GROWTH_SAFETY_MARGIN)
    still_available_size = 0u;
  else
    still_available_size -= EAMGC_GROWTH_SAFETY_MARGIN;
  
  /* We've just asked the system how much we can allocate before trashing. 
     Of course this assumes we aren't already trashing. Let's give a further
     upper limit to the heap space: the *total* heap size must not be greater
     than the whole physical space, minus a safety magin. Of course also this
     is an  approximation, because it might allow a heap to grow very large
     when there are other processes needing memory. But it seems to work well
     in practice. */
  const long int total_system_pages_no = sysconf(_SC_PHYS_PAGES);
  const unsigned_integer_t physical_memory_size_in_words =
    (unsigned_integer_t)
    (total_system_pages_no * system_page_size / sizeof(word_t));
  if(heap_size > (physical_memory_size_in_words -
                  EAMGC_MINIMUM_HEAP_SPACE_FOR_OTHER_PROCESSES))
    still_available_size = 0u;
  
  /* Divide the available space between the homogeneous and the large heap,
     according to their current need: */
  const unsigned_integer_t still_available_homogeneous_size = (unsigned_integer_t)
    (still_available_size * homogeneous_to_total_ratio);
  const unsigned_integer_t still_available_large_size = (unsigned_integer_t)
    (still_available_size * (1.0 - homogeneous_to_total_ratio));
  
#ifdef ENABLE_ASSERTIONS
  fprintf(stderr, "[Before trimming]: Current homogeneous heap size: %.3gMb\n",
          words_to_megabytes(homogeneous_heap_size));
  fprintf(stderr, "[Before trimming]: Current large heap size: %.3gMb\n",
          words_to_megabytes(large_heap_size));
  fprintf(stderr,
          "Alive homogeneous/alive total ratio is %f\n",
          alive_homogeneous_to_total_alive_ratio);
  fprintf(stderr,
          "Homogeneous/total ratio is %f\n",
          homogeneous_to_total_ratio);
  fprintf(stderr,
          "Total physical memory is %.3gMb\n",
          words_to_megabytes(physical_memory_size_in_words));
  fprintf(stderr,
          "Available heap space is %.3gMb (%.3gMb homogeneous + %.3gMb large)\n",
          words_to_megabytes(still_available_size),
          words_to_megabytes(still_available_homogeneous_size),
          words_to_megabytes(still_available_large_size));
  fprintf(stderr, "[Before trimming]: New maximum homogeneous heap size: %.3gMb\n",
          words_to_megabytes(best_homogeneous_heap_size));
  fprintf(stderr, "[Before trimming]: New maximum large heap size: %.3gMb\n",
          words_to_megabytes(best_large_heap_size));
#endif // #ifdef ENABLE_ASSERTIONS

  /* Trim the optimal heap size with respect to the "no trashing" limit: */
  if(((double)best_homogeneous_heap_size - (double)homogeneous_heap_size) >
     (double)still_available_homogeneous_size)
    best_homogeneous_heap_size = homogeneous_heap_size +
      still_available_homogeneous_size;
  if(((double)best_large_heap_size - (double)large_heap_size) > 
     (double)still_available_large_size)
    best_large_heap_size = large_heap_size + still_available_large_size;
  
#ifdef ENABLE_ASSERTIONS
  fprintf(stderr, "[After up-trimming]: New maximum homogeneous heap size: %.3gMb\n",
          words_to_megabytes(best_homogeneous_heap_size));
  fprintf(stderr, "[After up-trimming]: New maximum large heap size: %.3gMb\n",
          words_to_megabytes(best_large_heap_size));
#endif // #ifdef ENABLE_ASSERTIONS

  /* The optimal heap size as computed now may be zero if the whole heap is
     dead, or less than the minimum. Of course we don't want this: */
  const unsigned_integer_t minimum_homogeneous_heap_size =
    eamgc_get_minimum_homogeneous_heap_size();
  if(best_homogeneous_heap_size < minimum_homogeneous_heap_size)
    best_homogeneous_heap_size = minimum_homogeneous_heap_size;
  if(best_large_heap_size < EAMGC_MINIMUM_MAXIMUM_LARGE_HEAP_SIZE)
    best_large_heap_size = EAMGC_MINIMUM_MAXIMUM_LARGE_HEAP_SIZE;
  
#ifdef ENABLE_ASSERTIONS
  fprintf(stderr, "[After down-trimming]: New maximum homogeneous heap size: %.3gMb\n",
          words_to_megabytes(best_homogeneous_heap_size));
  fprintf(stderr, "[After down-trimming]: New maximum large heap size: %.3gMb\n",
          words_to_megabytes(best_large_heap_size));
#endif // #ifdef ENABLE_ASSERTIONS

#ifdef ENABLE_ASSERTIONS
  /* For no reason the new maximum heap size should be less than we are
     using *now* for alive objects: */
  assert(best_homogeneous_heap_size >= alive_homogeneous_heap_size);
  assert(best_large_heap_size >= alive_large_heap_size);
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* Ok, we computed a reasonable upper limit. Set it, so that heuristics
     use it: */
  eamgc_set_maximum_homogeneous_heap_size(best_homogeneous_heap_size);
  eamgc_set_maximum_large_heap_size(best_large_heap_size);
  
#ifdef ENABLE_ASSERTIONS
  fprintf(stderr, "New maximum homogeneous heap size: %.3gMb\n",
          words_to_megabytes(best_homogeneous_heap_size));
  fprintf(stderr, "New maximum large heap size: %.3gMb\n",
          words_to_megabytes(best_large_heap_size));
#endif // #ifdef ENABLE_ASSERTIONS

  /* With the current strategy the minimum heap size is never updated. This
     might change in the future. */
}
