/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef EAMGC_STATISTICS_H_
#define EAMGC_STATISTICS_H_

/* When we talk about "heap size" we always ignore data structures overhead;
   the homogeneous heap size, for example, is the sum of all the allocated
   homogeneous *payload*; this excludes page headers, list structures holding
   the pages together, hash tables, tracing stacks, pool structures and mark
   bits. */

#include "../../common/types.h"

/* Get or set new minimum heap size (in words). If a homogeneous page *can* be
   destroyed (because it's empty) and the heap size is greater than the current
   minimum size then the page is destroyed: */
unsigned_integer_t eamgc_get_minimum_homogeneous_heap_size(void)
  __attribute__(( pure ));
void eamgc_set_minimum_homogeneous_heap_size(const unsigned_integer_t size);

/* Get or set new maximum heap size (in words). If more space than 'size'
   words is allocated then a GC is triggered: */
unsigned_integer_t eamgc_get_maximum_homogeneous_heap_size(void)
  __attribute__(( pure ));
void eamgc_set_maximum_homogeneous_heap_size(const unsigned_integer_t size);
unsigned_integer_t eamgc_get_maximum_large_heap_size(void)
  __attribute__(( pure ));
void eamgc_set_maximum_large_heap_size(const unsigned_integer_t size);

/* Recompute statistics so that heuristics give more precise evaluations. This
   is a relatively costly procedure, thought to be called just once per GC; but
   what we compute here enables heuristics to be fast: */
void eamgc_recompute_statistics(void);

/* Return the total number of words taken by the objects which were alive
   at the last collection: */
inline unsigned_integer_t eamgc_get_alive_heap_size(void) __attribute__(( pure ));
inline unsigned_integer_t eamgc_get_alive_homogeneous_heap_size(void) __attribute__(( pure ));
inline unsigned_integer_t eamgc_get_alive_large_heap_size(void) __attribute__(( pure ));

/* Return the total number of words taken by the homogeneous and large
   payloads (not necessarily allocated or alive): */
inline unsigned_integer_t eamgc_get_heap_size(void)
  __attribute__(( pure ));

/* Set the alive heap size. Called by the tracer while marking reachable 
   objects. The size is in words. */
inline void eamgc_zero_alive_heap_size(void);
inline void eamgc_increment_alive_homogeneous_heap_size(const unsigned_integer_t size_to_add);
inline void eamgc_increment_alive_large_heap_size(const unsigned_integer_t size_to_add);

/* Useful for debugging: */
void eamgc_dump_heap_size(void);

#endif // #ifndef EAMGC_STATISTICS_H_
