/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include <stdio.h>
#include "gc.h"

static struct eamgc_homogeneous_pool *conses_pool;

struct my_cons{
  integer_t datum;
  struct my_cons *next;
};

void trace_my_cons(const word_t cons){
  //eamgc_trace_nth_word(cons, 0);
  eamgc_trace_nth_word(cons, 1);
}

struct my_cons *my_nil = NULL;

struct my_cons* my_cons(integer_t car, struct my_cons *cdr)
  __attribute__(( malloc ));
struct my_cons* my_cons(integer_t car, struct my_cons *cdr){
  struct my_cons *r =
    (struct my_cons*)eamgc_allocate_from_homogeneous_pool(conses_pool);
#ifdef ENABLE_ASSERTIONS
  assert(r->datum == (integer_t)0xffffffff);
  assert(r->next == (struct my_cons*)0xffffffff);
#endif
  r->datum = car;
  r->next = cdr;
#ifdef VERBOSE_DEBUG
  fprintf(stderr, "+ Created cons: ");
  eamgc_dump_homogeneous_object((word_t)r, stderr);
  fprintf(stderr, "\n");
#endif // #ifdef VERBOSE_DEBUG

  return r;
}

integer_t my_car(const struct my_cons *cons){
  return cons->datum;
}

struct my_cons* my_cdr(const struct my_cons *cons){
  return cons->next;
}

struct my_cons* my_interval_acc(integer_t a, integer_t b, struct my_cons *acc){
  if(a > b)
    return acc;
  else
    return my_interval_acc(a, b - 1, my_cons(b, acc));
}

struct my_cons* my_interval(integer_t a, integer_t b){
  return my_interval_acc(a, b, my_nil);
}

#define TTT 10 //1000 //10000

struct my_cons* my_make_list(integer_t n){
  struct my_cons *r = my_nil;
  integer_t i;
  static integer_t q = 1;
  for(i = n; i > 0; i--)
    r = my_cons(q, r);
  q++;
  return r;
}

struct my_cons* my_interval_from_1(integer_t n){
  struct my_cons *r = my_nil;
  integer_t i;
  for(i = n; i > 0; i--)
    r = my_cons(i, r);
  return r;
}

void check_my_list(struct my_cons *xs){
  if(xs == NULL){
    //fprintf(stderr, "+++ OK [empty list]\n");
    return;
  }
  const integer_t element = xs->datum;
  integer_t length = 0;
  do{
    if(xs->datum != element){
      fprintf(stderr, "!! [element is "INTEGER_T_FORMAT"("HEXADECIMAL_INTEGER_T_FORMAT") instead of "INTEGER_T_FORMAT"("HEXADECIMAL_INTEGER_T_FORMAT")] at index "INTEGER_T_FORMAT"\n",
              (integer_t)xs->datum, (unsigned_integer_t)xs->datum,
              (integer_t)element, (unsigned_integer_t)element,
              (integer_t)length);
      fprintf(stderr, "The problem is at "INTEGER_T_FORMAT" [%p]\n", (integer_t)xs, xs);
      eamgc_homogeneous_page_t page = eamgc_object_to_actual_homogeneous_page((word_t)xs);
      fprintf(stderr, "The homogeneous page is at "INTEGER_T_FORMAT" [%p]\n", (integer_t)page, page);
      assert(0);
    } // if
    xs = my_cdr(xs);
    length++;
    assert(length >= 0);
    assert(length < TTT);
  }while(xs != NULL);
  fprintf(stderr, "+++ OK [element is "INTEGER_T_FORMAT", length is "INTEGER_T_FORMAT"]\n", (integer_t)element, (integer_t)length);
}

void initialize(void){
  conses_pool =
    eamgc_create_homogeneous_pool(2u,
                               1u,
                               NULL,
                               trace_my_cons 
                               //eamgc_homogeneous_trace_everything_function
                               );
  eamgc_register_root_tracing_function(eamgc_trace_c_context_roots);
}

void destructive_my_interval(const integer_t a, const integer_t b,
                             struct my_cons **list){
  *list = my_nil;
  integer_t i;
  for(i = b; i >= a; i--)
    *list = my_cons(i, *list);
}

void check_interval(struct my_cons* list_){
  struct my_cons* list = list_;
  integer_t lower, length = 0;

#ifdef VERBOSE_DEBUG
  fprintf(stderr, "# The list is:\n");
#endif // #ifdef VERBOSE_DEBUG

  if(list == NULL){
#ifdef VERBOSE_DEBUG
    fprintf(stderr, "[Ok: the list is empty]\n");
#endif // #ifdef VERBOSE_DEBUG
    return;
  }
  lower = list->datum - 1;
  do{
#ifdef VERBOSE_DEBUG
    eamgc_dump_homogeneous_object((word_t)list, stderr);
    fprintf(stderr, "\n");
#endif // #ifdef VERBOSE_DEBUG
    if(list->datum != lower + 1){
      fprintf(stderr, "!! [element is "INTEGER_T_FORMAT"("HEXADECIMAL_INTEGER_T_FORMAT") instead of "INTEGER_T_FORMAT"("HEXADECIMAL_INTEGER_T_FORMAT")] at index "INTEGER_T_FORMAT"\n",
              (integer_t)list->datum, (unsigned_integer_t)list->datum,
              (integer_t)lower + 1, (unsigned_integer_t)lower + 1, 
              (integer_t)length);
      assert(0);
    } // if
    lower = list->datum;
    length ++;
    list = list->next;
  }while(list != NULL);
  //#ifdef VERBOSE_DEBUG
  //fprintf(stderr, "# Ok: the list length is "INTEGER_T_FORMAT".\n\n", (int)length);
  //#endif // #ifdef VERBOSE_DEBUG
}

void do_work(void){
  initialize();
  
  int i;
  for(i = 0; i < 200000000; i++)
    eamgc_allocate_from_homogeneous_pool(conses_pool);
}

int main_(){
  eamgc_initialize_garbage_collector_and_continue(do_work);
  return 0;
}

#include "generic_pools.h"

////////////////////////////
void __called_from_main(void);
void __what_the_user_thinks_its_main(void);

int main(int argc, char **argv){
  // x:
  eamgc_initialize_garbage_collector_and_continue(__called_from_main);
  eamgc_finalize_garbage_collector();
  //  goto x;
  return 0;
}

void __called_from_main(void){
  eamgc_register_root_tracing_function(eamgc_trace_c_context_roots);

  eamgc_create_generic_pools();
  __what_the_user_thinks_its_main();
}

#define malloc eamgc_malloc
#define calloc eamgc_calloc
#define realloc eamgc_realloc
#define free eamgc_free
#define main __what_the_user_thinks_its_main

/* Check that p is a chain of length lenght: */
void check_chain(void **p, int length){
  //fprintf(stdout, "\n-c- begin:");
  integer_t last;
  if(length != 0)
    last = (integer_t)p[1] + 1;
  while(length != 0){
    //fprintf(stderr, "chain element: "INTEGER_T_FORMAT" [%p]\n", (integer_t)p, p);
    //    fprintf(stdout, INTEGER_T_FORMAT" ", (((integer_t*)p)[1])); fflush(stdout);
    assert(p != NULL);
    if(eamgc_object_to_actual_homogeneous_page(p) == NULL){
      fprintf(stderr,"!1 About to crash: the supposed pointer is "INTEGER_T_FORMAT" ["UNSIGNED_INTEGER_T_FORMAT"]\n",
              (integer_t)p, (unsigned_integer_t)p);
      assert(p != (void*)0xdeaddead);
      assert(*p != (void*)0xdeaddead);
    }
    if(((integer_t)(p[1])) != last - 1){
      fprintf(stderr, "-c- ERROR in list checking: "INTEGER_T_FORMAT" "INTEGER_T_FORMAT"\n",
              (integer_t)last, (integer_t)(p[1])); fflush(stdout);
      fprintf(stderr,"-c- About to crash: the pointer is "INTEGER_T_FORMAT" ["HEXADECIMAL_INTEGER_T_FORMAT"]\n",
              (integer_t)p, (integer_t)p);
      assert(0);
    }
    last = (integer_t)(p[1]);
    p = (void**)(*p);
    length--;
  }
  if(p != NULL){
    fprintf(stderr,"!2 About to crash: the non-NULL pointer is "INTEGER_T_FORMAT" ["INTEGER_T_FORMAT"]\n",
            (integer_t)p, (integer_t)p);
    /*
    fprintf(stderr,"!2 Its index is "INTEGER_T_FORMAT".\n",
            (integer_t)
            eamgc_object_to_index_in_homogeneous_page(p,
                                                   eamgc_object_to_hypothetical_homogeneous_page(p)));
    */
    eamgc_dump_homogeneous_object(p, stderr);
  }
  assert(p == NULL);
  //fprintf(stdout, "\n-c- ok\n\n");
}

void **reverse_chain_tr(void **chain, void **acc){
  if(chain == NULL)
    return acc;
  else{
    void **new_acc = (void**)malloc(8);
    new_acc[0] = acc;
    new_acc[1] = chain[1];
    //if(rand() % 10 == 0) malloc(rand() % 16 + 1);
    return reverse_chain_tr((void**)chain[0], new_acc);
  }
}

void **reverse_chain(void **chain){
  return reverse_chain_tr(chain, NULL);
}

void my_large_finalize_function(const word_t object, const unsigned_integer_t size){
}

void main(void){
  int i;
  
#define N 100
  word_t q[N];
  eamgc_register_buffer_of_roots((word_t)q, N);
  for(i = 0; i < 3000000; i++){
    q[rand() % N] = 
      malloc(1 + i % (20 * 4 + 1));
  } // for
}

void main__(void){
  //int q; eam_save_stack_bottom(&q - 100);
  printf("Hello, world!\n");
  void **p = NULL;
  void **new_p;
  eamgc_register_root((void*)&p);
  eamgc_register_root((void*)&new_p);
  while(1){
    static int length = 0;
    static int n = 0;
    n++;
    length++;
    //const size_t more_bytes = 2 * 4;//rand() % (16 * 4);//5;//10;
    //fprintf(stderr, "ALLOC: making cons for "INTEGER_T_FORMAT"\n", (integer_t)n);
    new_p = (void**)malloc(8 + rand() % 16);//9);//sizeof(void*) + more_bytes);
    assert(eamgc_object_to_actual_homogeneous_page(new_p) != NULL);
    assert(!eamgc_is_marked_in_homogeneous_page(new_p,
                                             eamgc_object_to_hypothetical_homogeneous_page(new_p)));
    //(void**)calloc(sizeof(void*) + more_bytes);
#ifdef ENABLE_ASSERTIONS
    if(new_p[0] != eamgc_just_created_pattern){
      fprintf(stderr, "new_p[0] == "INTEGER_T_FORMAT" ["HEXADECIMAL_INTEGER_T_FORMAT"]\n", 
              ((integer_t*)new_p)[0], ((integer_t*)new_p)[0]);
      assert(0);
    } // if
#endif // #ifdef ENABLE_ASSERTIONS
    new_p[0] = p;
    new_p[1] = (word_t)n;
    p = new_p;
    /*
    if((rand() % 500000) == 0){
      //fprintf(stderr, "{CHECK-1}... ");
      check_chain(p, length);
      //fprintf(stderr, "{OK}\n{FORCED COLLECTION}");
      eamgc_garbage_collect();
      fprintf(stderr, "{CHECK-2}... ");
      check_chain(p, length);
      fprintf(stderr, "{OK}\n\n");
    }
    */
    /*
    if(0 == (rand() % 1000)){//100000)){//if((rand() % 10000) == 0){
      //fprintf(stderr, "{RANDOM CHECK BEFORE TRUNCATING}... ");
      check_chain(p, length);
      //fprintf(stderr, "{OK}\n{TRUNCATING}\n\n");
      length = 0; p = NULL;
    }
    */
    
    if(rand() % 200000 == 0){
      fprintf(stderr, "{DOUBLE REVERSE}\n");
      p = reverse_chain(reverse_chain(p));
    }
    if(rand() % 200000 == 0){
      //fprintf(stderr, "{RANDOM CHECK: "INTEGER_T_FORMAT"}... ", (integer_t)length);
      /*
      fprintf(stderr, "A chain of length "INTEGER_T_FORMAT" starts at: "INTEGER_T_FORMAT" [%p]\n",
              (integer_t)length, (integer_t)&p, &p);
      */
      check_chain(p, length);
      //fprintf(stderr, ".");
      fprintf(stderr, "{OK}\n");
    }
    
  } // while
}

////////////////////////////

/*
void main(){
  eamgc_initialize_garbage_collector_and_continue(do_work);
  
  return 0;
}
*/
