/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "trace.h"
#include "roots.h"
#include "../../common/stack.h"
#include "homogeneous.h"
#include "large.h"
#include "../../common/malloc.h"
#include "../../common/command_line.h"


/* We use two distinct stacks to hold pointers to the objects whose
   children we have to trace. We don't want to check whether an object
   is homogeneous or large at pop time, since we already checked before
   pushing. See also the initial comment in the header about recursion: */
static EAM_DEFINE_STACK(homogeneous)
static EAM_DEFINE_STACK(large)

/* Create internal data structures needed for tracing: */
void eamgc_initialize_trace_structures(void){
  eam_initialize_stack(homogeneous);
  eam_initialize_stack(large);
}

/* Destroy internal data structures needed for tracing: */
void eamgc_finalize_trace_structures(void){
  eam_destroy_stack(homogeneous);
  eam_destroy_stack(large);
}

/* Trace the children of the given object which *must* be an already marked
   homogeneous object (typically just popped from the appropriate stack): */
static void trace_homogeneous_object_children(const word_t object){
  /* Extract the page from the object pointer (this is extremely fast
     and requires no memory accesses): */
  const eamgc_homogeneous_page_t page =
    eamgc_object_to_hypothetical_homogeneous_page(object);
#ifdef ENABLE_ASSERTIONS
  /* If we are here then the object should be really homogeneous, and
     already marked: */
  assert(eamgc_object_to_actual_homogeneous_page(object) == page);
  assert(eamgc_is_marked_in_homogeneous_page(object, page));
#endif
#ifdef VERBOSE_DEBUG
  //fprintf(stderr, "-> children of %p\n", object);
#endif // #ifdef VERBOSE_DEBUG

  /* Trace the object's children (i.e. push pointers to the yet unmarked
     ones onto the right stack, and mark them). This is performed by the
     trace function in the homogeneous page: */
  eamgc_call_trace_function_for_homogeneous_object_children(object, page);
}

/* Trace the given object which *must* be an already marked large object
   (typically just popped from the appropriate stack): */
static void trace_large_object_children(const word_t object){
#ifdef ENABLE_ASSERTIONS
  /* If we are here then the object should be really homogeneous, and
     already marked: */
  assert(eamgc_is_an_actual_large_object(object));
  assert(eamgc_is_large_object_marked(object));
#endif
  
  /* Trace the object's children (i.e. push pointers to the yet unmarked
     ones onto the right stack, and mark them). This is performed by the
     large object trace function: */
  eamgc_call_trace_function_for_large_object_children(object);
}

/* Trace the given object which *must* be a homogeneous object: */
void eamgc_trace_homogeneous_object(const word_t object){
  /* Extract the page from the object pointer (this is extremely fast
     and requires no memory accesses): */
  const eamgc_homogeneous_page_t page =
    eamgc_object_to_hypothetical_homogeneous_page(object);
  
  /* If the object is already marked then do nothing: otherwise mark it
     now (we want it to enter the stack at most *once* per collection)
     and push it onto the appropriate stack. Its children will come 
     later, when we pop it. */
  if(eamgc_mark_in_homogeneous_page(object, page) == 0){
#ifdef VERBOSE_DEBUG
    fprintf(stderr, "-> ");
    eamgc_dump_homogeneous_object(object, stderr);
    fprintf(stderr, "\n");
#endif // #ifdef VERBOSE_DEBUG
    eam_push_onto_stack(object, homogeneous);
  } // if
  else{
#ifdef VERBOSE_DEBUG
    fprintf(stderr, "(ALREADY MARKED: ");
    eamgc_dump_homogeneous_object(object, stderr);
    fprintf(stderr, ")\n");
#endif // #ifdef VERBOSE_DEBUG
  }
}

/* Trace the given object which *must* be a large object: */
void eamgc_trace_large_object(const word_t object){
#ifdef ENABLE_ASSERTIONS
  /* Assure that object is actually a large object: */
  assert(eamgc_is_an_actual_large_object(object));
#endif
  
  /* If the object is already marked then do nothing: otherwise mark it
     now (we want it to enter the stack at most *once* per collection)
     and push it onto the appropriate stack. Its children will come 
     later, when we pop it. */
  if(eamgc_mark_large_object(object) == 0){
#ifdef VERBOSE_DEBUG
    fprintf(stderr, "-> [large at %p]\n", (void*)object);
#endif // #ifdef VERBOSE_DEBUG
    eam_push_onto_stack(object, large);
  } // if
  else{
#ifdef VERBOSE_DEBUG
    fprintf(stderr, "(ALREADY MARKED: [large at %p]\n", (void*)object);
#endif // #ifdef VERBOSE_DEBUG
  }
}

/* Trace the given object, if it's really a heap-allocated object; if it
   isn't then no action is performed. */
void eamgc_trace_object(const word_t potential_object){
  /* If potential_object's bit pattern can't express not a word-aligned
     pointer then don't even bother to look up tables: it just can't be
     a heap pointer. Note how this trivial check avoid potentially slow
     memory accesses, and only requires cheap operations on registers: */
  if((((unsigned_integer_t)potential_object) % 
      (unsigned_integer_t)sizeof(word_t)) != (unsigned_integer_t)0u)
    return;
  
  /* Check whether the object is homogeneous or large and call the
     appropriate tracing function. If it's neither do nothing: */
  if(eamgc_object_to_actual_homogeneous_page(potential_object) != NULL)
    eamgc_trace_homogeneous_object(potential_object);
  else if(eamgc_is_an_actual_large_object(potential_object))
    eamgc_trace_large_object(potential_object);
  /* else do nothing. */
}

/* Trace all reachable objects, starting from the roots which must already
   have been traced. When the function returns all alive objects in the heap
   are marked. There should be no need to call this for the user. */
void eamgc_trace(void){
  /* Keep popping an object and pushing its still unmarked children until
     we have *both* stacks empty. Note that a homogeneous object can refer
     a large object, and vice-versa, so we need to always examine *both*
     stacks: */
  do{
    /* Work with the homogeneous stack until it becomes empty (this may
       make the large stack grow): */
    while(! eam_is_stack_empty(homogeneous)){
      //fprintf(stderr, ",");
      trace_homogeneous_object_children(eam_pop_from_stack(homogeneous));
    }
    /* Work with the large stack until it becomes empty (this may make
       the homogeneous stack grow): */
    while(! eam_is_stack_empty(large)){
      //fprintf(stderr, ",");
      trace_large_object_children(eam_pop_from_stack(large));
    }

  /* The large stack is always empty if we get here, and that's the reason
     why we don't explicitly check that *both* stacks are empty here: */
#ifdef ENABLE_ASSERTIONS
    assert(eam_is_stack_empty(large));
#endif
  } while(! eam_is_stack_empty(homogeneous));
#ifdef ENABLE_ASSERTIONS
  assert(eam_is_stack_empty(homogeneous));
  assert(eam_is_stack_empty(large));
#endif
}

/* Mark the given number of distinct objects starting from the already
   marked ones. Return EAMGC_ALL_ALIVE_OBJECTS_WERE_MARKED or
   EAMGC_NOT_ALL_ALIVE_OBJECTS_WERE_MARKED. Note that this function does
   not automatically trace roots, nor implements any automatic write
   barrier. The function may mark one object more than requested. This
   is for efficiency's sake: we can save a conditional within the loop
   in this way. */
integer_t eamgc_incremental_trace(const integer_t how_many_objects_to_mark){
  /* Remember how many objects we still have to mark: */
  integer_t objects_still_to_mark = how_many_objects_to_mark;
  
  /* Keep popping an object and pushing its still unmarked children until
     we have *both* stacks empty or we reach the limit. Note that a
     homogeneous object can refer a large object, and vice-versa, so we 
     need to examine *both* stacks. In this case it's better to examine
     the two stacks together, since otherwise the limit might make us
     always stop at the first stack, while the other one continues to 
     grow: */
  while((! eam_is_stack_empty(homogeneous)) || (! eam_is_stack_empty(large))){
    /* Have we done enough work? If so stop now, even if we have not
       traced everything: */
    if(objects_still_to_mark == 0)
      return EAMGC_NOT_ALL_ALIVE_OBJECTS_WERE_MARKED;
    
    /* Mark an homogeneous object, if any: */
    if(! eam_is_stack_empty(homogeneous)){
      trace_homogeneous_object_children(eam_pop_from_stack(homogeneous));
      objects_still_to_mark--;
    } // first if

    /* Mark a large object, if any: */
    if(! eam_is_stack_empty(large)){
      trace_large_object_children(eam_pop_from_stack(large));
      objects_still_to_mark--;
    } // second if
  } // while
  
  /* Ok, we marked the whole heap. Inform the caller about it: */
  return EAMGC_ALL_ALIVE_OBJECTS_WERE_MARKED;
}
