/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef TRACE_H_
#define TRACE_H_

#include "gc.h"

/* "Tracing" a heap object means: 
   - doing nothing if the object is already marked.
   - otherwise marking it, and adding pointers to all the objects directly
     reachable from it to an appropriate data structure, so that they can
     also be traced, later.
   Note that tracing an object by itself is not a recursive process: tracing
   an object can only end up marking an object and adding its direct children
   to a data structure. It's important not to use recursion in tracing 
   functions, in order not to overflow the C stack. 
   The eamgc_trace() function follows a conceptually recursive logic even if
   it's implemented in an iterative fashion using explicit stacks, for the
   reason above. */

/* Create internal data structures needed for tracing: */
void eamgc_initialize_trace_structures(void);

/* Destroy internal data structures needed for tracing: */
void eamgc_finalize_trace_structures(void);

/* Trace all reachable objects, starting from the roots which must already
   have been traced. When the function returns all alive objects in the heap
   are marked. There should be no need to call this for the user. */
void eamgc_trace(void);

/* Trace the given object, if it's really a heap-allocated object; if it
   isn't then no action is performed. */
void eamgc_trace_object(const word_t pointer_to_potential_heap_object);

/* Trace the given object which *must* be a homogeneous object: */
void eamgc_trace_homogeneous_object(const word_t pointer_to_object);

/* Trace the given object which *must* be a large object: */
void eamgc_trace_large_object(const word_t pointer_to_object);

/* The values returned by eamgc_incremental_trace(): */
#define EAMGC_ALL_ALIVE_OBJECTS_WERE_MARKED 0
#define EAMGC_NOT_ALL_ALIVE_OBJECTS_WERE_MARKED 1

/* Mark the given number of distinct objects starting from the already
   marked ones. Return EAMGC_ALL_ALIVE_OBJECTS_WERE_MARKED or
   EAMGC_NOT_ALL_ALIVE_OBJECTS_WERE_MARKED. Note that this function does
   not automatically trace roots, nor implements any automatic write
   barrier. The function may mark one object more than requested. This
   is for efficiency's sake: we can save a conditional within the loop
   in this way. */
// To do: this can't possibly work with the current bitvector scheme.
// We can experiment with this after implementing a more mundane
// mark-sweep mode for the homogeneous heap.
integer_t eamgc_incremental_trace(const integer_t how_many_objects_to_trace);

/* A handy macro used to trace a given word of an object. This is useful for
   writing tracing functions referring an object's children, using either
   exact or conservative pointer finding. As it's usual in C the index of 
   the first word is 0. */
#define eamgc_trace_nth_word(object, word_index) \
  eamgc_trace_object(((word_t*)(object))[(integer_t)(word_index)])
  //eamgc_trace_object((word_t)(((word_t*)(object)) + (integer_t)(word_index)))

/*  */

#endif // #ifndef TRACE_H_
