/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef EAM_LIBARIES_FAST_STACK_H_
#define EAM_LIBARIES_FAST_STACK_H_

#include "../../common/types.h"
#include "../../common/malloc.h"
#include <assert.h>

/* A stack can be accessed either with a top pointer or a frame pointer; the
   caller may also use both, but in this case it's his/her responsibility to
   keep them coherent. In other words, no operation on the stack top pointer
   updates the frame pointer, and vice-versa. 
   
   The caller is free to store pointers in stacks, but it's not safe to use
   pointers to objects stored *within* the stack if the stack is resized,
   since resizing the stack may move all stores objects. Offsets should be
   stored instead.
   
   No bound check is performed. The caller should enstablish some reasonable
   policy for checking for overflow and resizing only when appropriate. 
   
   This macro-based implementation is less convenient than the one in
   memory_stack.h, but may have superior performance, especialy if pointers
   are kept in registers, as the eAM allows.
   
   To declare a stack named <name> you should declare/define the following
   C variables, with types word_t or integer_t (macros always cast to the
   correct type: this makes implementation of eAM primitives easier):
   - <name>_fast_stack
   - <name>_fast_stack_allocated_size
   And only one of:
   - <name>_fast_stack_top_pointer
   - <name>_fast_stack_frame_pointer
   All the macros in this module only require <name> as the stack name. */


#include "stack.h"
#include "../../common/types.h"
#include "../../common/malloc.h"

/* Create a new stack with the given size. The size is in words: */
#define eam_create_fast_stack_with_frame_pointer(size, name) {          \
    name ## _fast_stack_allocated_size = (word_t)(unsigned_integer_t)(size); \
    name ## _fast_stack = (word_t)                                      \
      eam_xmalloc(sizeof(word_t) *                                      \
                  (unsigned_integer_t)(name ## _fast_stack_allocated_size)); \
    name ## _fast_stack_frame_pointer = (word_t)(name ## _fast_stack);  \
}
#define eam_create_fast_stack_with_top_pointer(size, name) {            \
    name ## _fast_stack_allocated_size = (word_t)(unsigned_integer_t)(size); \
    name ## _fast_stack = (word_t)                                      \
      eam_xmalloc(sizeof(word_t) *                                      \
                  (unsigned_integer_t)(name ## _fast_stack_allocated_size)); \
    name ## _fast_stack_top_pointer = (word_t)(name ## _fast_stack);    \
}

/* Destroy an existing stack: */
#define eam_destroy_fast_stack(name) {                                  \
    free((void*)(name ## _fast_stack));                                 \
}

/* Query/set the stack allocated size, always expressed in words. Resizing
   a stack automatically updates top and frame pointers. Be sure to see the
   initial comment about resizing in presence of pointers within the stack. */
#define eam_get_fast_stack_allocated_size(stack)                \
  ((unsigned_integer_t)(stack ## _fast_stack_allocated_size))
#define eam_resize_fast_stack_with_top_pointer(new_size_in_words, stack){ \
    const word_t *old_fast_stack = (const word_t*)(stack ## _fast_stack); \
    (stack ## _fast_stack_allocated_size) = (word_t)(new_size_in_words);      \
    (stack ## _fast_stack) = (word_t)                                   \
      eam_xrealloc((void*)(stack ## _fast_stack),                       \
                   (unsigned_integer_t)(stack ## _fast_stack_allocated_size) * \
                   sizeof(word_t));                                     \
    (stack ## _fast_stack_top_pointer) = (word_t)                       \
      (((word_t*)(stack ## _fast_stack)) +                              \
       (((word_t*)(stack ## _fast_stack_top_pointer)) - old_fast_stack)); \
}
#define eam_resize_fast_stack_with_frame_pointer(new_size_in_words, stack){ \
    const word_t *old_fast_stack = (const word_t*)(stack ## _fast_stack); \
    (stack ## _fast_stack_allocated_size) = (word_t)(new_size_in_words); \
    (stack ## _fast_stack) = (word_t)                                   \
      eam_xrealloc((void*)(stack ## _fast_stack),                       \
                   (unsigned_integer_t)(stack ## _fast_stack_allocated_size) * \
                   sizeof(word_t));                                     \
    (stack ## _fast_stack_frame_pointer) = (word_t)                     \
      (((word_t*)(stack ## _fast_stack)) +                              \
       (((word_t*)(stack ## _fast_stack_frame_pointer)) - old_fast_stack)); \
}

/* Push, using and updating the top pointer: */
#define eam_push_word_onto_fast_stack(datum, stack){                    \
    word_t * const s = (word_t * const)(stack ## _fast_stack_top_pointer); \
    *s = (word_t)(word_t)(datum);                                 \
    stack ## _fast_stack_top_pointer = (word_t)(s + 1);                 \
}

/* Pop, using and updating the top pointer. It's convenient to use the GCC
   statement-as-expression extension here: */
#define eam_pop_word_from_fast_stack(stack)({                           \
      stack ## _fast_stack_top_pointer = (word_t)                       \
        (((word_t*)(stack ## _fast_stack_top_pointer)) - 1);            \
      *((word_t*)(stack ## _fast_stack_top_pointer));                   \
})

/* Return the object at the top, using the top pointer: */
#define eam_top_word_of_fast_stack(stack) \
  (*(((word_t*)stack ## _fast_stack_top_pointer) - 1))

/* Return the number of elements in the stack, from the bottom (included) to
   the top pointer (excluded): */
#define eam_fast_stack_elements_no(stack)                               \
  ((unsigned_integer_t)                                                 \
   (((word_t*)stack ## _fast_stack_top_pointer) -                       \
    ((word_t*)stack ## _fast_stack)))

/* Explicitly update the top pointer: not needed if only using push/pop: */
#define eam_set_fast_stack_top_pointer(new_top_pointer, stack){         \
    (stack ## _fast_stack_top_pointer) = (word_t)(word_t*)(new_top_pointer); \
}

/* Explicitly update the top pointer: not needed if only using push/pop. The
   offset (in words) may be zero, positive or negative: */
#define eam_move_fast_stack_top_pointer(offset, stack){                 \
    (stack ## _fast_stack_top_pointer) = (word_t)                       \
      (((word_t*)(stack ## _fast_stack_top_pointer)) + (integer_t)(offset)); \
}

/* Update the frame pointer: */
#define eam_set_fast_stack_frame_pointer(new_frame_pointer, stack){     \
    (stack ## _fast_stack_frame_pointer) = (word_t)(word_t*)(new_frame_pointer); \
}

/* Update the frame pointer: */
#define eam_move_fast_stack_frame_pointer(offset, stack){               \
    (stack ## _fast_stack_frame_pointer) = (word_t)                     \
      (((word_t*)(stack ## _fast_stack_frame_pointer)) + ((integer_t)offset)); \
}

/* Query the top pointer: */
#define eam_get_fast_stack_top_pointer(stack) \
  (word_t*)(stack ## _fast_stack_top_pointer)

/* Query the bottom pointer: */
#define eam_get_fast_stack_bottom_pointer(stack) \
  (word_t*)(stack ## _fast_stack)

/* Query the frame pointer: */
#define eam_get_fast_stack_frame_pointer(stack) \
  (word_t*)(stack ## _fast_stack_frame_pointer)

/* Load, using the frame pointer. The offset (in words) may be positive,
   zero or negative: */
#define eam_load_word_from_fast_stack(offset, stack) \
  *(((word_t*)(stack ## _fast_stack_frame_pointer)) + (integer_t)(offset));

/* Store, using the frame pointer. The offset (in words) may be positive,
   zero or negative: */
#define eam_store_word_into_fast_stack(offset, datum, stack){   \
    word_t *s = (word_t*)(stack ## _fast_stack_frame_pointer);  \
    *(s + (integer_t)(offset)) = (word_t)(datum);               \
}

#endif // #ifndef EAM_LIBARIES_FAST_STACK_H_
