/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


/* The bulk of the implementation is in the header. */
#include "memory_stack.h"

/* What a stack is made of: */
struct eam_memory_stack{
  /* The actual array of words: */
  word_t *stack;
  
  /* By convention the top pointer points to the first *still unused* word: */
  word_t *top_pointer;
  
  /* Pointer to some conventional datum, with respect to which other data are
     accessed: */
  word_t *frame_pointer;
  
  /* The stack allocated size, in words: */
  unsigned_integer_t allocated_size;
}; // struct

/* Create a new stack with the given size. The size is in words: */
struct eam_memory_stack* eam_create_memory_stack(const unsigned_integer_t allocated_size){
  struct eam_memory_stack *const r = eam_xmalloc(sizeof(struct eam_memory_stack));
  r->stack = (word_t*)eam_xmalloc(allocated_size * sizeof(word_t));
  r->top_pointer = r->stack;
  r->frame_pointer = r->stack;
  r->allocated_size = allocated_size;
  return r;
}

/* Resize an existing stack: */
void eam_resize_memory_stack(const unsigned_integer_t new_size_in_words,
                      struct eam_memory_stack *stack){
  const word_t *old_memory_stack = stack->stack;
  stack->allocated_size = new_size_in_words;
  stack->stack = eam_xrealloc(stack->stack, new_size_in_words * sizeof(word_t));
  
  /* Update the top and frame pointer, by keeping constant the offset between
     each of them and the stack beginning: */
  stack->top_pointer = stack->stack + (stack->top_pointer - old_memory_stack);
  stack->frame_pointer = stack->stack + (stack->frame_pointer - old_memory_stack);
}

/* Destroy the given stack: */
void eam_destroy_memory_stack(struct eam_memory_stack *stack){
  free(stack->stack);
  free(stack);
}

/* Return the stack allocated size, in words: */
inline unsigned_integer_t eam_get_memory_stack_allocated_size(const struct eam_memory_stack *stack){
  return stack->allocated_size;
}

/* Push, using and updating the top pointer: */
inline void eam_push_word_onto_memory_stack(const word_t datum, struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert(stack->top_pointer < (stack->stack + stack->allocated_size));
#endif // #ifdef ENABLE_ASSERTIONS
  *stack->top_pointer = datum;
  stack->top_pointer++;
}

/* Pop, using and updating the top pointer: */
inline word_t eam_pop_word_from_memory_stack(struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert(stack->top_pointer > stack->stack);
#endif // #ifdef ENABLE_ASSERTIONS
  stack->top_pointer--;
  return *(stack->top_pointer);
}

/* Return the object at the top, of course using the top pointer: */
inline word_t eam_top_word_of_memory_stack(const struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert(stack->top_pointer > stack->stack);
#endif // #ifdef ENABLE_ASSERTIONS
  return *(((word_t*)stack->top_pointer) - 1);
}

/* Return the number of objects, from the bottom (included) to the top pointer
   (excluded): */
inline unsigned_integer_t eam_memory_stack_elements_no(
    const struct eam_memory_stack *stack){
  return (unsigned_integer_t)
    (((word_t*)stack->top_pointer) - ((word_t*)stack->stack));
}

/* Explicitly update the top pointer: not needed if only using push/pop: */
inline void eam_set_memory_stack_top_pointer(const word_t *new_top_pointer,
                                      struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert(new_top_pointer < (stack->stack + stack->allocated_size));
  assert(new_top_pointer >= stack->stack);
#endif // #ifdef ENABLE_ASSERTIONS
  stack->top_pointer = (word_t*)new_top_pointer;
}

/* Explicitly update the top pointer: not needed if only using push/pop. The
   offset (in words) may be zero, positive or negative: */
inline void eam_move_memory_stack_top_pointer(const integer_t offset,
                                       struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert((stack->top_pointer + offset) < (stack->stack + stack->allocated_size));
  assert((stack->top_pointer + offset) >= stack->stack);
#endif // #ifdef ENABLE_ASSERTIONS
  stack->top_pointer += offset;
}

/* Update the frame pointer: */
inline void eam_set_memory_stack_frame_pointer(const word_t *new_frame_pointer,
                                        struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert(new_frame_pointer < (stack->stack + stack->allocated_size));
  assert(new_frame_pointer >= stack->stack);
#endif // #ifdef ENABLE_ASSERTIONS
  stack->frame_pointer = (word_t*)new_frame_pointer;
}

/* Update the frame pointer: */
inline void eam_move_memory_stack_frame_pointer(const integer_t offset,
                                         struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert((stack->frame_pointer + offset) < (stack->stack + stack->allocated_size));
  assert((stack->frame_pointer + offset) >= stack->stack);
#endif // #ifdef ENABLE_ASSERTIONS
  stack->frame_pointer += offset;
}

/* Query the top pointer: */
inline word_t* eam_get_memory_stack_top_pointer(const struct eam_memory_stack *stack){
  return stack->top_pointer;
}

/* Query the bottom pointer: */
inline word_t* eam_get_memory_stack_bottom_pointer(const struct eam_memory_stack *stack){
  return stack->stack;
}

/* Query the frame pointer: */
inline word_t* eam_get_memory_stack_frame_pointer(const struct eam_memory_stack *stack){
  return stack->frame_pointer;
}

/* The same, for multi-word buffers: */
inline void eam_push_buffer_onto_memory_stack(const word_t *data,
                                       const unsigned_integer_t size,
                                       struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert((stack->top_pointer + size) < (stack->stack + stack->allocated_size));
#endif // #ifdef ENABLE_ASSERTIONS
  unsigned_integer_t i;
  word_t * const top_pointer = stack->top_pointer;
  for(i = 0; i < size; i++)
    *(top_pointer + i) = data[i];
  stack->top_pointer += size;
}

/* The same, for multi-word buffers: */
inline void eam_pop_buffer_from_memory_stack(const word_t *data,
                                      const unsigned_integer_t size,
                                      word_t *where_to_put_the_copy,
                                      struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert((stack->top_pointer + size) >= stack->stack);
#endif // #ifdef ENABLE_ASSERTIONS
  integer_t i;
  const word_t * const top_pointer = stack->top_pointer;
  for(i = 0; i < size; i++)
    where_to_put_the_copy[i] = *(top_pointer - (integer_t)size + i);
  stack->top_pointer -= size;
}

/* Load, using the frame pointer. The offset (in words) may be positive,
   zero or negative: */
inline word_t eam_load_word_from_memory_stack(const integer_t offset,
                                       struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert((stack->frame_pointer + offset) >= stack->stack);
  assert((stack->frame_pointer + offset) < (stack->stack + stack->allocated_size));
#endif // #ifdef ENABLE_ASSERTIONS
  return *(stack->frame_pointer + offset);
}

/* Store, using the frame pointer. The offset (in words) may be positive,
   zero or negative: */
inline void eam_store_word_into_memory_stack(const integer_t offset,
                                      const word_t datum,
                                      struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert((stack->frame_pointer + offset) >= stack->stack);
  assert((stack->frame_pointer + offset) < (stack->stack + stack->allocated_size));
#endif // #ifdef ENABLE_ASSERTIONS
  *(stack->frame_pointer + offset) = datum;
}

/* Same as above, for multi-word buffers: */
inline void eam_load_buffer_from_memory_stack(const integer_t offset,
                                       const unsigned_integer_t size_in_words,
                                       word_t *where_to_put_the_copy,
                                       struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert((stack->frame_pointer + offset) >= stack->stack);
  assert((stack->frame_pointer + offset + size_in_words - 1) < (stack->stack + stack->allocated_size));
#endif // #ifdef ENABLE_ASSERTIONS
  integer_t i;
  const word_t * const frame_pointer = stack->frame_pointer; 
  for(i = 0; i < size_in_words; i++)
    where_to_put_the_copy[i] = *(frame_pointer + offset + i);
}

/* Same as above, for multi-word buffers: */
inline void eam_store_buffer_into_memory_stack(const integer_t offset,
                                        const word_t *data,
                                        const unsigned_integer_t size_in_words,
                                        struct eam_memory_stack *stack){
#ifdef ENABLE_ASSERTIONS
  assert((stack->frame_pointer + offset) >= stack->stack);
  assert((stack->frame_pointer + offset + size_in_words - 1) < (stack->stack + stack->allocated_size));
#endif // #ifdef ENABLE_ASSERTIONS
  integer_t i;
  word_t * const frame_pointer = stack->frame_pointer;
  for(i = 0; i < size_in_words; i++)
    *(frame_pointer + offset + i) = data[i];
}
