/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <setjmp.h>
#include <assert.h>

#include "driver.h"

/* The place where to jump back when calling longjmp() to exit the driver;
   there's no need to expose these complications to the caller; he/she should
   only call eam_eam_exit() to exit the driver: */
static jmp_buf eam_driver_place_to_jump_back_to;

/* This benefits from -funroll-all-loops, or even crude manual loop unrolling.
   To do: use either.
   This function shouldn't be inlined, so that we're sure that our stack bottom
   approximation is actually below the frame for this function; in this way our
   approximation should be safe enough not to be fooled by too smart compiler
   optimizations which might merge frames and reorder local variables on the
   stack: */
static void eam_driver_main_loop(const eam_continuation_t c)
  __attribute__(( nothrow, noinline, noreturn ));
static void eam_driver_main_loop(const eam_continuation_t c){
  /* Execute the given continuation which returns the next continuation to
     execute; execute it, and so on "for ever". The last continuation
     explicitly longjmp()s to exit the infinite loop and return to the
     driver (just to exit it): */
  eam_continuation_t continuation_register = c;
  do{
    //    printf(".");
    continuation_register = (eam_continuation_t)(continuation_register());
  }while(1);
}

/* The exit code given back by the eAM to the user when eam_driver() returns,
   by an eam_exit() call: */
static integer_t eam_exit_code = 0;

/* Exit the eAM driver and return to the procedure which entered it: */
void eam_exit(const integer_t exit_code){
  eam_exit_code = exit_code;
  longjmp(eam_driver_place_to_jump_back_to, // where to go back
	  1);                               // not used
}

/* The location of the object which we consider the stack bottom: */
static volatile void* stack_bottom = NULL; // this initialization prevents stupid errors

integer_t eam_driver(const eam_continuation_t initial_continuation){
  /* This is useful to locate (an approximation of) the stack bottom: */
  volatile word_t a_dummy_object_at_the_stack_bottom;
  stack_bottom = &a_dummy_object_at_the_stack_bottom;
  
  /* Set the place where the user is able to return: */  
  if(setjmp(eam_driver_place_to_jump_back_to)){
    /* We arrived back here from a longjmp(), i.e. the caller longjmp()ed away
       from the driver loop. We're done: */
#ifdef ENABLE_VERBOSE_DEBUG
    fprintf(stderr, "Back out of the eAM driver loop via longjmp()...\n");
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    stack_bottom = NULL; // let's prevent stupid errors...
    return eam_exit_code;
  }
  else{
    /* We arrived here right after the execution of the setjmp() above: */
#ifdef ENABLE_VERBOSE_DEBUG
    fprintf(stderr, "Entering the eAM driver loop right after a setjmp()...\n");
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    eam_driver_main_loop(initial_continuation);
  } // else
}

/* Return the stack bottom; it's useful to initialize the GC, when needed. This
   is only usable within the driver loop, because what is returned is the
   address of a local object which is only alive when the driver loop is
   active. */
void* eam_get_stack_bottom(void){
#ifdef ENABLE_ASSERTIONS
  assert(stack_bottom != NULL); // it's NULL when the driver isn't active
#endif // #ifdef ENABLE_ASSERTIONS
  return (void*)stack_bottom;
}
