/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "modules.h"
#include "../common/malloc.h"
#include "../common/string_map.h"
#include "../common/mangle.h"

/* An eAM module structure: */
struct eam_module{
  /* Was the module dynamically linked? */
  integer_t is_dynamically_linked;
  
  /* The module filename and name, in both mangled and unmangled form.
     They are strings xmalloc()ed for this purpose: */
  char *file_name;
  char *unmangled_name;
  char *mangled_name;
  
  /* The module handle: */
  lt_dlhandle handle;
  
  /* Globals: */
  integer_t globals_no;
  word_t *globals;
  
  /* Blocks: */
  integer_t blocks_no;
  eam_continuation_t *blocks;
  
  /* Initialization and finalization functions: */
  //void **initialization_function;
  void (*initialization_function)(void);
  void (*finalization_function)(void);
  
  /* The reference count: */
  integer_t pins_no;
}; // struct

/* Size of each hash map: */
#define HASH_MAPS_SIZE 2049

/* The filename->module and unmangled-name->module mappings:. */
static struct eam_string_map *file_name_to_eam_module;
static struct eam_string_map *unmangled_name_to_eam_module;

/* The conventional "filename" used when opening the executable itself: */
static const char *no_file_file_name = "<self>";

/* Initialize eAM modules data structures: */
void eam_initialize_eam_modules(void){
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "Initializing eAM module support.\n");
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  /* Initialize libltdl support: */
  int initialization_failures_no = lt_dlinit();
  assert(initialization_failures_no == 0);
  
  /* Create data structures: */
  file_name_to_eam_module =
    eam_create_string_map_of_size(sizeof(struct eam_module*),
                                  HASH_MAPS_SIZE,
                                  eam_default_string_map_datum_finalizer);
  unmangled_name_to_eam_module =
    eam_create_string_map_of_size(sizeof(struct eam_module*),
                                  HASH_MAPS_SIZE,
                                  eam_default_string_map_datum_finalizer);
}

/* Unload all modules and libraries, and finalize eAM modules data
   structures: */
void eam_finalize_eam_modules(void){
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "Finalizing eAM module support.\n");
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  /* Destroy data structures: */
  eam_destroy_string_map(file_name_to_eam_module);
  eam_destroy_string_map(unmangled_name_to_eam_module);
  
  /* Finalize libltdl support: */
  lt_dlexit();
}

/* Return the value associated to the symbol based on the given name (see the
   comment before eam_load_information_from_module_symbols()) in the given
   module: */
static lt_ptr eam_load_symbol(const char *name, const struct eam_module *module){
  char symbol_name[EAML_COMPILATION_UNIT_MANGLED_NAME_LENGTH + 100];
  sprintf(symbol_name, "eam_%s_%s", module->mangled_name, name);
  //  fprintf(stderr, "??? Looking for \"%s\" in %s...\n", symbol_name, module->unmangled_name);
  return lt_dlsym(module->handle, symbol_name);
}

/* Load information from the exported symbols in the given module, which must
   be already initialized except for globals and blocks. Return 0 on success.
   In case of failure destroy the module and remove references to it from all
   data structures, and return 0: */
static integer_t eam_load_information_from_module_symbols(struct eam_module *module){
  char **symbol_value;
  /* Check that this is really the module we think it is: */
  symbol_value = (char**) eam_load_symbol("unmangled_name", module);
  /* This ugly code is due to undefined evaluation order ir || clauses... I'd have
     preferred just having been able to write 
     if((symbol_value == NULL) || strcmp(*symbol_value, ...)): */
  static char* dummy_name = "<>";
  if(symbol_value == NULL)
    symbol_value = &dummy_name;
  if(strcmp(*symbol_value, module->unmangled_name)){
#ifdef ENABLE_VERBOSE_DEBUG
    fprintf(stderr, "The file %s does NOT contain the module %s\n",
            module->file_name, module->unmangled_name);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    /* The name doesn't match: */
    lt_dlclose(module->handle);
    eam_remove_from_string_map(file_name_to_eam_module, module->file_name);
    eam_remove_from_string_map(unmangled_name_to_eam_module, module->unmangled_name);
    free(module->file_name);
    free(module->unmangled_name);
    free(module->mangled_name);
    free(module);
    return -1;
  } // if
  
  /* Ok, if we arrived here then the module file is what we think it is. It makes
     no sense to also check the other symbols. The check we made is enough to
     prevent *inadvertite* errors, but malicious module are impossible to catch
     in any way, because they can execute arbitrary code. */
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "Ok, the file %s contains the module %s\n",
          module->file_name, module->unmangled_name);
#endif // #ifdef ENABLE_VERBOSE_DEBUG

  /* Load initialization and finalization functions: */
  module->initialization_function = eam_load_symbol("initialize", module);
  assert(module->initialization_function != NULL);
  module->finalization_function = eam_load_symbol("finalize", module);
  assert(module->finalization_function != NULL);
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "The initialization function is at %p\n",
          (void*)module->initialization_function);
  fprintf(stderr, "The finalization function is at %p\n",
          (void*)module->finalization_function);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  
  /* Load globals: */
  integer_t* globals_no = (integer_t*)eam_load_symbol("globals_no", module);
  assert(globals_no != NULL);
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "The module %s exports %i globals.\n",
          module->unmangled_name, (int)*globals_no);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  module->globals_no = *globals_no;
  module->globals = eam_xmalloc(sizeof(word_t) * (module->globals_no + 1));
  integer_t i;
  char global_index[100];
  for(i = 1; i <= module->globals_no; i++){
    sprintf(global_index, "global_%i", (int)i);
    word_t* global = (word_t*) eam_load_symbol(global_index, module);
    if(global != NULL){
      module->globals[i] = global;
#ifdef ENABLE_VERBOSE_DEBUG
      fprintf(stderr, "The global n.%i is at %p\n", (int)i, (void*)global);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    }
    else{
      module->globals[i] = NULL;
#ifdef ENABLE_VERBOSE_DEBUG
      fprintf(stderr, "The global n.%i is internal\n", (int)i);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    } // else
  } // for

  /* Load blocks: */
  integer_t* blocks_no = (integer_t*)eam_load_symbol("blocks_no", module);
  assert(blocks_no != NULL);
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "The module %s has %i blocks.\n",
          module->unmangled_name, (int)*blocks_no);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  module->blocks_no = *blocks_no;
  module->blocks = eam_xmalloc(sizeof(word_t) * (module->blocks_no + 1));
  char label_index[100];
  for(i = 1; i <= module->blocks_no; i++){
    sprintf(label_index, "label_%i", (int)i);
    eam_continuation_t block =
      (eam_continuation_t) eam_load_symbol(label_index, module);
    if(block != NULL){
      module->blocks[i] = block;
#ifdef ENABLE_VERBOSE_DEBUG
      fprintf(stderr, "The block n.%i is at %p\n", (int)i, (void*)block);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    }
#ifdef ENABLE_VERBOSE_DEBUG
    else
      fprintf(stderr, "The block n.%i is internal\n", (int)i);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  } // for
  
  return 0;
}

/* This is a workaround to what I supect to be a Libtool bug. See the comment
   in modules.h: */
static void call(void (*function)(void))
#ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
  __attribute__(( noinline ));
#else
  __attribute__(( always_inline ));
#endif // #ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
static void call(void (*function)(void)){
  function();
}

/* Create a module loading it from the needed file. Return the new module or
   NULL in case of errors. If unmangled_name is NULL then dlopen() the main
   executable, and make a statically linked module. This allows to use a
   single interface for all the code which *may* be dynamically loaded:  */
static struct eam_module* eam_actually_create_module(const char *unmangled_name,
                                                     const char *path_name)
#ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
  __attribute__(( noinline ));
#else
  __attribute__(( always_inline ));
#endif // #ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
static struct eam_module* eam_actually_create_module(const char *unmangled_name,
                                                     const char *path_name){
  //fprintf(stderr, "COMPLETE:  %s\n", path_name);
  //fprintf(stderr, "FILE:      %s\n", eam_path_name_to_file_part(path_name));
  //fprintf(stderr, "DIRECTORY: %s\n", eam_path_name_to_directory_part(path_name));
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "+ CREATING module \"%s\" from %s...\n", unmangled_name, path_name);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  
  /* Try to load the module, and return NULL immediately on failure: */
  lt_dlhandle handle;
  handle = lt_dlopen(path_name);
  if(handle == NULL){
#ifdef ENABLE_VERBOSE_DEBUG
    fprintf(stderr, "Failed lt_dlopen()ing the module %s\n",
            (path_name != NULL) ? path_name : no_file_file_name);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    return NULL;
  } // if
  
  /* Ok, loading succeeded; make a structure, link it to our global table, and
     return it: */
  struct eam_module *r = (struct eam_module*)
    eam_xmalloc(sizeof(struct eam_module));
  r->handle = handle;
  if(path_name != NULL)
    r->file_name = eam_path_name_to_file_part(path_name);
  else{
    r->file_name = (char*)eam_xmalloc(strlen(no_file_file_name) + 1);
    strcpy(r->file_name, no_file_file_name);
  } // else
  r->unmangled_name = (char*)eam_xmalloc(strlen(unmangled_name) + 1);
  strcpy(r->unmangled_name, unmangled_name);
  r->mangled_name =
    eam_mangle_with_limit(unmangled_name,
                          EAML_COMPILATION_UNIT_MANGLED_NAME_LENGTH);
  r->pins_no = 1;
  r->is_dynamically_linked = (path_name != NULL);
  eam_insert_into_string_map(file_name_to_eam_module, r->file_name, &r);
  eam_insert_into_string_map(unmangled_name_to_eam_module, r->unmangled_name, &r);
  if(eam_load_information_from_module_symbols(r) == 0){
#ifdef ENABLE_VERBOSE_DEBUG
    fprintf(stderr, "Calling the initialization function... (r is %p)\n", (void*)r);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    call(r->initialization_function);
#ifdef ENABLE_VERBOSE_DEBUG
    fprintf(stderr, "...ok, returned with success (r is %p).\n", (void*)r);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    return r;
  }
  else{ // symbols loading failed. Everything was already destroyed. Fail:
#ifdef ENABLE_VERBOSE_DEBUG
    fprintf(stderr, "FAILED loading module %s from %s\n", unmangled_name, path_name);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    return NULL;
  }
}

/* Destroy the given module: */
static void eam_actually_destroy_module(struct eam_module *module)
#ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
  __attribute__(( noinline ));
#else
  __attribute__(( always_inline ));
#endif // #ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
static void eam_actually_destroy_module(struct eam_module *module){
#ifdef ENABLE_ASSERTIONS
  assert(module != NULL);
  assert(module->pins_no == 0);
#endif // #ifdef ENABLE_ASSERTIONS
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "- DESTROYING module %s (module is at %p)...\n",
          module->unmangled_name, module);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  
  /* Call the finalization function: */
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "Calling the finalization function %p (module is at %p)...\n",
          (void*)module->finalization_function,
          (void*)module);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  call(module->finalization_function);
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "...ok, returned with success (module is at %p)\n", module);
#endif // #ifdef ENABLE_VERBOSE_DEBUG  
  
  /* Actually destroy the module: */
  lt_dlclose(module->handle);
  /* Remove the structure from our global tables: */
  eam_remove_from_string_map(file_name_to_eam_module, module->file_name);
  eam_remove_from_string_map(unmangled_name_to_eam_module, module->unmangled_name);
  
  /* Destroy data structures: */
  free(module->file_name);
  free(module->unmangled_name);
  free(module->mangled_name);
  free(module->globals);
  free(module->blocks);
  free(module);
}

/* Load the module found in the given file, if not already loaded. If path_name
   is NULL then load the module from the main executable symbol, and return a
   statically-linked module. Return the module (be it new or already existing),
   or NULL in case of failure. This allows to use a single interface for all the
   code which *may* be dynamically loaded: */
struct eam_module* eam_load_module(const char *unmangled_name,
                                   const char *path_name){
  /* Check whether we have already loaded the module: make a new one if
     we find nothing, otherwise just pin what we already have and return it: */
  struct eam_module **module = (struct eam_module **)
    eam_access_string_map(unmangled_name_to_eam_module, unmangled_name);
  if(module == NULL){
    struct eam_module *new_module =
      eam_actually_create_module(unmangled_name, path_name);
#ifdef ENABLE_VERBOSE_DEBUG
    if(new_module != NULL)
      fprintf(stderr, "About to return a NEW module at %p\n", new_module);
    else
      fprintf(stderr, "About to return NULL, since eam_actually_create_module() failed\n");
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    return new_module;
  }
  else{
#ifdef ENABLE_VERBOSE_DEBUG
    fprintf(stderr, " + PINNING %s\n", unmangled_name);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    (*module)->pins_no++;
#ifdef ENABLE_VERBOSE_DEBUG
    fprintf(stderr, "About to return an EXISTING module at %p\n", *module);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
    return *module;
  } // else
}
//#undef ENABLE_VERBOSE_DEBUG

/* Unload the given module, if not needed anymore */
void eam_unload_module(const struct eam_module *module_){
  struct eam_module *module = (struct eam_module*)module_;
  /* Unpin the module. If the reference count drops to zero then destroy it: */
  module->pins_no--;
  if(module->pins_no == 0)
    eam_actually_destroy_module(module);
#ifdef ENABLE_VERBOSE_DEBUG
  else
    fprintf(stderr, " - UNPINNING %s (in %s)\n",
            module->unmangled_name, module->file_name);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
}

/* Unload the given module, if not needed anymore */
void eam_unload_module_given_its_name(const char *unmangled_name){
  struct eam_module **module = (struct eam_module **)
    eam_access_string_map(unmangled_name_to_eam_module, unmangled_name);
  assert(module != NULL);
  eam_unload_module(*module);
}

/* Return the module name: */
char* eam_module_to_unmangled_name(const struct eam_module *module){
#ifdef ENABLE_ASSERTIONS
  assert(module != NULL);
#endif // #ifdef ENABLE_ASSERTIONS
  return module->unmangled_name;
}

/* Return the module name: */
char* eam_module_to_mangled_name(const struct eam_module *module){
#ifdef ENABLE_ASSERTIONS
  assert(module != NULL);
#endif // #ifdef ENABLE_ASSERTIONS
  return module->unmangled_name;
}

/* Call the block with the given index (which must be exported) in the given
   module. The resulting continuation is returned: */
eam_continuation_t eam_call_block(const struct eam_module *module,
                                  const integer_t index){
#ifdef ENABLE_ASSERTIONS
  assert(index > 0);
  assert(index < module->blocks_no);
#endif // #ifdef ENABLE_ASSERTIONS
  return module->blocks[index]();
}

/* Return a pointer (which may be updated) to the globlal with the given
   index (which must be exported) in the givne module: */
word_t* eam_access_global(const struct eam_module *module,
                          const integer_t index){
#ifdef ENABLE_ASSERTIONS
  assert(index > 0);
  assert(index < module->globals_no);
#endif // #ifdef ENABLE_ASSERTIONS
  return module->globals[index];
}
