/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef EAM_MODULES_H_
#define EAM_MODULES_H_

#include "../common/types.h"
#include "../common/filesystem.h"
#include <ltdl.h>

/* eAM modules are libraries of compiled native code which the eAM can load
   and unload at runtime. Dependencies are automatically resolved, and a 
   reference count is transparently kept. By saying "to load" a module we mean
   to create it if not already loaded, and to increment its reference count.
   Loading a module implies recursively loading all the modules and the
   libraries it depends on.
   When a module is "unloaded" its reference count is decremented. A module
   is actually destroyed, and the modules it depends to are unloaded, when
   its reference count drops to zero.
   Modules are automatically initialized at creation time and finalized at
   destruction time, by calling functions they contain for that purpose. */

/* An eAM module, as an abstract data type: */
struct eam_module;

/* Initialize eAM modules data structures: */
void eam_initialize_eam_modules(void);

/* Unload all modules and libraries, and finalize eAM modules data
   structures: */
void eam_finalize_eam_modules(void);

/* Load the module found in the given file, if not already loaded. If path_name
   is NULL then load the module from the main executable symbol, and return a
   statically-linked module. Return the module (be it new or already existing),
   or NULL in case of failure. This allows to use a single interface for all the
   code which *may* be dynamically loaded: */
struct eam_module* eam_load_module(const char *unmangled_name,
                                   const char *file_name)
#ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
  __attribute__(( noinline ));
#else
  __attribute__(( always_inline ));
#endif // #ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG

/* Unload the given module, if not needed anymore. In that case the module
   is also automatically finalized: */
void eam_unload_module(const struct eam_module *module)
#ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
  __attribute__(( noinline ));
#else
  __attribute__(( always_inline ));
#endif // #ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
void eam_unload_module_given_its_name(const char *unmangled_name)
#ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
  __attribute__(( noinline ));
#else
  __attribute__(( always_inline ));
#endif // #ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG

/* Call the block with the given index (which must be exported) in the given
   module. The resulting continuation is returned. The attribute is needed to
   work around what I suspect to be a Libtool bug. To do: investigate. */
eam_continuation_t eam_call_block(const struct eam_module *module,
                                  const integer_t index)
#ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG
  __attribute__(( noinline ));
#else
  __attribute__(( always_inline ));
#endif // #ifdef ENABLE_WORK_AROUND_LIBTOOL_BUG

/* Return a pointer (which may be updated) to the globlal with the given
   index (which must be exported) in the givne module: */
word_t* eam_access_global(const struct eam_module *module,
                          const integer_t index);

/* Return the module name: */
char* eam_module_to_unmangled_name(const struct eam_module *module)
  __attribute__(( pure ));
char* eam_module_to_mangled_name(const struct eam_module *module)
  __attribute__(( pure ));


#endif // #ifndef EAM_MODULES_H_
