#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <malloc.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/mman.h> // for mprotect()
#include <signal.h>

typedef void** word_t;

static size_t page_size;

static size_t allocated_size;
size_t eam_stack_elements_no;
word_t *stack;

size_t make_a_page_size_multiple(const size_t s){
  const double needed_pages_no = ceil((double)s  / (double)page_size);
  return ((size_t)needed_pages_no) * page_size;
}

void make_stack(const size_t new_size){
  word_t *buffer;

  fprintf(stderr, "Creating a stack of %i words (%i bytes)\n", 
	  (int)new_size,
	  (int)(new_size * sizeof(word_t)));
  /* Make a buffer for the stack, including the check pages: */
  allocated_size =
    make_a_page_size_multiple(new_size * sizeof(word_t) + page_size * 2);
  eam_stack_elements_no = new_size;
  fprintf(stderr, "Calling mmap() with size %i bytes\n", (size_t)allocated_size); 
  buffer = mmap(NULL,
		allocated_size,
		PROT_READ | PROT_WRITE | PROT_EXEC,
		MAP_PRIVATE | MAP_ANONYMOUS,
		0,
		0);
  
  /* Check that mmap() didn't fail: */
  perror("mmap()");
  assert(buffer != (word_t*)-1);
  
  /* Forbid reading and writing from check pages: */
  assert(mprotect(buffer,
		  page_size,
		  PROT_NONE) == 0);
  assert(mprotect(buffer + (allocated_size - page_size) / sizeof(word_t),
		  page_size,
		  PROT_NONE) == 0);
  
  /* Make the stack start right after the bottom check page: */
  stack = buffer + (page_size / sizeof(word_t));
}


void set_page_info(void){
  page_size = (size_t)sysconf(_SC_PAGESIZE);
  fprintf(stderr,
	  "The page size on this machine is %i bytes\n",
	  (int)page_size);
}

/* Oliver Laumann says on 11.8.1994:
 *
 * Determining the faulting address in a SIGSEGV or SIGBUS handler is
 * definitely system dependent; it is even impossible on some platforms.
 *
 * In some UNIX variants, the third parameter passed to the signal handler
 * is a `struct sigcontext *scp'; the address is available either as
 * scp->sc_badvaddr (Ultrix), as scp->sc_jmpbuf.jmp_context.except[3]
 * (AIX), as scp->sc_sl.sl_ss.ss_cr21 (HP-UX), or as scp->sc_traparg_a0
 * (DEC/Alpha OSF/1).
 *
 * Or the address is passed to the handler as an additional, fourth
 * argument (BSD, SunOS 4.x); or you can set the SA_SIGINFO flag in the
 * sa_flags field of the `struct sigaction' argument to the sigaction()
 * syscall to receive a `siginfo_t *' argument with a si_addr field in
 * the handler (System V Release 4, SunOS 5.x, SGI Irix 5.1).
 */

static void out_of_bounds_stack_handler(int signal_no,
					siginfo_t *info, // To do: not portable
					void *uap){      // To do: not portable
  fprintf(stderr,
	  "Signal %i raised accessing memory at %p\n", 
	  signal_no,
	  info->si_addr);
  //fprintf(stderr, "[Intercepted signal %i]\n", signal);
  fprintf(stderr, "Returning from signal handler.\n");
}

static void set_signal_handlers(void){
  /* To do: SIGBUS is thrown instead of SIGSEGV on some systems: */
  struct sigaction sa;
  
  /* Fetch the current action: */
  assert(sigaction(SIGSEGV, 0, &sa) == 0);

  /* Update our copy of the action: */
  sa.sa_sigaction = out_of_bounds_stack_handler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_SIGINFO;
  
  /* Install our updated copy: */
  assert(sigaction(SIGSEGV, &sa, 0) == 0); 
}

static void initialize(void){
  set_page_info();
  make_stack(4096);
  set_signal_handlers();
}

static void use_the_stack(){
  int i;
  fprintf(stderr, "Testing the stack... ");
  for(i = 0; i < eam_stack_elements_no; i++)
    stack[i] = (word_t)i;
  fprintf(stderr, "success.\n");
}

static void abuse_the_stack(){
  fprintf(stderr, "Abusing the stack (over the top)... ");
  stack[eam_stack_elements_no] = (word_t)1;
  fprintf(stderr, "success (this shouldn't happen).\n");

  fprintf(stderr, "Abusing the stack (below the bottom)... ");
  stack[-1] = (word_t)1;
  fprintf(stderr, "success (this shouldn't happen).\n");
  /*
  int i;
  fprintf(stderr, "Abusing the stack (over the top)... ");
  for(i = 0; i <= eam_stack_elements_no; i++)
    stack[i] = (word_t)i;
  fprintf(stderr, "success (this shouldn't happen).\n");
  
  fprintf(stderr, "Abusing the stack (below the bottom)... ");
  for(i = eam_stack_elements_no - 1; i >= -1; i--)
    stack[i] = (word_t)i;
  fprintf(stderr, "success (this shouldn't happen).\n");
  */
}

int main(){
  initialize();
  
  use_the_stack();
  abuse_the_stack();
  
  return 0;
}
