/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

%{
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "../common/types.h"
#include "eaml.h"
#include "eamlparser.h"
#include "../common/malloc.h"

/* Show a parse error: */
int yyerror(char* message);

/* Update yytext and yyleng to remove surrounding braces and unescape closed
   braces: */
static void unescape_embedded_code();

/* Update yytext and yyleng to remove surrounding double quotes and replace
   escape sequences with their expansion: */
static void unescape_string();
%}

%option yylineno

DIGIT   [0-9]
LOWER   [a-z]
UPPER   [A-Z]
LETTER  {LOWER}|{UPPER}
UNDERSCORE \_

COMMA     \,
SEMICOLON \;

SIGN       (\-|\+)

INTEGER    {SIGN}?{DIGIT}+

/* We accept as an identifier pretty much everything not containing whitespaces;
   we also forbid backslashes, because allowing them might lead users to think
   they can use escapes. They can't. For more elaborate identifier names users
   should write identifiers as strings. Raw identifiers can't end with a colon
   because otherwise they could be mistaken for labels: */
RAW_IDENTIFIER [^ \a\b\t\n\r\v\(\)\{\}\"\'%,\\]*[^ \a\b\t\n\r\v\(\)\{\}\"\'%,\\:]

/*CESCAPE    (\\.)|(\\{DIGIT})|(\\{DIGIT}{DIGIT})|(\\{DIGIT}{DIGIT}{DIGIT})*/
CESCAPE    (\\[\"\'\\abtnrv])
STRING     \"([^\n\"]|{CESCAPE})*\"

EMBEDDED_CODE \{([^}]|(\\\}))*\}

WHITESPACE       (" "|\t|\n|\r|\v)+
LINECOMMENT #[^\n\r]*(\n|\r)
/* An hack to recognize a line comment at the EOF without newline: */
LINECOMMENT_AT_EOF #[^\n\r]*
COMMENT       {LINECOMMENT}|{LINECOMMENT_AT_EOF}
LABEL         {RAW_IDENTIFIER}:

%%

{EMBEDDED_CODE} { unescape_embedded_code(); return EMBEDDED_CODE; }
{WHITESPACE}   { /* Do nothing */ }
{COMMENT}      { /* Do nothing */ }
{LINECOMMENT_AT_EOF} {
  if(! feof(yyin)){
    fprintf(stderr, "scan error (malformed line comment): this should not happen");
    assert(0);
  } // if
  /* else we are at EOF, and this is ok: do nothing */ }
%r({DIGIT}+)      { return INDEXED_REGISTER; }
\%{RAW_IDENTIFIER} { /* Remove the "%" prefix: */
                    yytext++; yyleng--;
                    return NAMED_REGISTER; }
{INTEGER}         { return NUMBER; }
\(                { return OPENPAR; }
\)                { return CLOSEPAR; }
\,                { return COMMA; }
":="              { return TAKES; }
constant          { return CONSTANT; }
exported          { return EXPORTED; }
global            { return GLOBAL; }
primitive         { return PRIMITIVE; }
code              { return CODE; }
compilation-unit  { return COMPILATION_UNIT; }
import            { return IMPORT; }
library           { return LIBRARY; }
initialization    { return INITIALIZATION; }
finalization      { return FINALIZATION; }
immediate         { return IMMEDIATE; }
c-name            { return C_NAME; }
c-declaration     { return C_DECLARATION; }
c-implementation  { return C_IMPLEMENTATION; }
c-procedure-body  { return C_PROCEDURE_BODY; }
end               { return END; }
arity             { return ARITY; }
define-register   { return DEFINE_REGISTER; }
as                { return AS; }
memory            { return MEMORY; }
non-tail-recursive-goto { return NON_TAIL_RECURSIVE_GOTO; }
goto              { return GOTO; }
if                { return IF; }
{LABEL}           { return LABEL; }
{STRING}          { unescape_string(); 
  //  fprintf(stderr, "[eAML scanner]: unescaped string is \"%s\"\n", yytext);
                    return IDENTIFIER; }
{RAW_IDENTIFIER}  { /* No unescaping takes place here */
  //  fprintf(stderr, "[eAML scanner]: unescaped string is \"%s\"\n", yytext);
                    return IDENTIFIER; }

%%

/* Update yytext and yyleng to remove surrounding braces and unescape closed
   braces: */
static void unescape_embedded_code(){
  /* Allocate a temporary buffer; its needed size will be yyleng + 1 in the
     worst case: */
  char *buffer = (char*)eam_xmalloc(yyleng + 1);
  int i, buffer_i = 0;
  for(i = 1; i < yyleng - 1; i++)
    switch(yytext[i]){
    case '\\':{
      /* If this is a known escape, then the *next* character should be copied;
	 otherwise simply copy the whole escape sequence as it is: */
      if((yytext[i + 1] == '\\') || (yytext[i + 1] == '}'))
	i++;
      /* We don't want the break here: go on with the next case. */
    }
    default:
      buffer[buffer_i++] = yytext[i];
    } // switch
  
  /* Copy the temporary buffer into yytext, its length into yyleng, and destroy
     the temporary buffer: */
  buffer[buffer_i] = '\0'; yyleng = buffer_i;
  strcpy(yytext, buffer);
#ifdef ENABLE_ASSERTIONS
  assert(strlen(yytext) == yyleng);
#endif // #ifdef ENABLE_ASSERTIONS
  free(buffer);
}

/* Update yytext and yyleng to remove surrounding double quotes and replace
   escape sequences with their expansion: */
static void unescape_string(){
#ifdef ENABLE_ASSERTIONS
  assert(yytext[0] == '"');
  assert(yytext[yyleng - 1] == '"');
#endif // #ifdef ENABLE_ASSERTIONS

  /* Allocate a temporary buffer; its needed size will be yyleng + 1 in the
     worst case: */
  char *buffer = (char*)eam_xmalloc(yyleng + 1);
  int i, buffer_i = 0;
  for(i = 1; i < yyleng - 1; i++)
    if(yytext[i] == '\\')
      switch(yytext[i + 1]){ // no breaks are intended.
      case '\\': buffer[buffer_i++] = '\\'; i++; break;
      case '"':  buffer[buffer_i++] = '"';  i++; break;
      case '\'': buffer[buffer_i++] = '\''; i++; break;
      case 'a':  buffer[buffer_i++] = '\a'; i++; break;
      case 'b':  buffer[buffer_i++] = '\b'; i++; break;
      case 't':  buffer[buffer_i++] = '\t'; i++; break;
      case 'n':  buffer[buffer_i++] = '\n'; i++; break;
      case 'r':  buffer[buffer_i++] = '\r'; i++; break;
      case 'v':  buffer[buffer_i++] = '\v'; i++; break;
      /* To do: also support numeric escape sequences? */
      default:
        fprintf(stderr,
                "eAML scanner: warning: unknown escape \\%c (expanded as '%c').\n",
                yytext[i + 1], yytext[i + 1]);
        fprintf(stderr,
                "eAML scanner:          Most probably this is not what you want.\n");
	//i += 1; // skip the *next* character
      } // switch
    else
      /* Copy the character: */
      buffer[buffer_i++] = yytext[i];
  
  /* Copy the temporary buffer into yytext, its length into yyleng, and destroy
     the temporary buffer: */
  buffer[buffer_i] = '\0'; yyleng = buffer_i;
  strcpy(yytext, buffer);
#ifdef ENABLE_ASSERTIONS
  assert(strlen(yytext) == yyleng);
#endif // #ifdef ENABLE_ASSERTIONS
  free(buffer);
}

int yywrap(){
  return 1;
}

/* An ugly global defined in the parser: */
extern char *file_name_being_parsed;

int yyerror(char* message){
  fprintf(stderr, "%s:%i: %s near \'%s\'\n",
	  file_name_being_parsed, yylineno, message, yytext);
  exit(EXIT_FAILURE);
}
