/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "eaml.h"
#include "../common/command_line.h"
#include "../common/malloc.h"
#include "symbol_table.h"
#include "eamlparser.h"
#include "generate_c.h"
#include "compile_time_parameters.h"

extern int yylineno; // the usual global generated by flex
extern char* yytext; // the usual global generated by flex

/* This is what we need for translating a compilation unit. Just a way to pass
   fewer parameters. */
struct eam_translation_context{
  /* The compilation unit we're translating: */
  struct eam_compilation_unit *compilation_unit;
  
  /* The symbol table holding *all* the symbols of the compilation unit we're
     translating, and only the *exported* symbols of the (perhaps indirectly-)
     included compilation units: supplied by the caller, and not created,
     updated or destroyed here. */
  struct eam_symbol_table *symbol_table;
  
  /* The stream where to emit the generated C code: */
  FILE *output_stream;
}; // struct

/* Return the number of definitions with the given type: */
static integer_t count_in_definitions(const struct eam_compilation_unit
                                      *compilation_unit,
                                      const enum eam_definition_type type){
  struct eam_list *p;
  integer_t r = 0;
  for(p = compilation_unit->definitions; ! eam_null(p); p = eam_cdr(p))
    if(((const struct eam_definition *)(eam_car(p)))->type == type)
      r++;
  return r;
}

/* Return the number of commands with the given type: */
static integer_t count_in_commands(const struct eam_compilation_unit
                                   *compilation_unit,
                                   const enum eam_command_type type){
  struct eam_list *p;
  integer_t r = 0;
  for(p = compilation_unit->commands; ! eam_null(p); p = eam_cdr(p))
    if(((const struct eam_command *)(eam_car(p)))->type == type)
      r++;
  return r;
}

/* Return the number of globals/primitives/blocks defined in the given
   compilation unit, counting both internal and exported symbols: */
static integer_t compilation_unit_to_globals_no(const struct eam_compilation_unit
                                                *compilation_unit){
  return count_in_definitions(compilation_unit,
                              eam_global_definition_type);
}
static integer_t compilation_unit_to_primitives_no(const struct eam_compilation_unit
                                                *compilation_unit){
  return count_in_definitions(compilation_unit,
                              eam_primitive_definition_type);
}
static integer_t compilation_unit_to_blocks_no(const struct eam_compilation_unit
                                                *compilation_unit){
  return count_in_commands(compilation_unit, eam_label_command_type);
}

/* Call eam_lookup_symbol_table_into_string() with the only symbol table we use
   here: */
static char* lookup_into_string(const char *symbol,
				const struct eam_translation_context *translation_context){
  return eam_lookup_symbol_table_into_string(translation_context->symbol_table,
					 symbol);
}

/* Note that the symbol table in the returned translation context is NULL. */
static struct eam_translation_context* initialize_translation(
				      const struct eam_compilation_unit *compilation_unit,
				      const char *output_directory){
  struct eam_translation_context *r;
  char *output_file_name = (char*) eam_xmalloc(strlen(output_directory) +
					  EAML_COMPILATION_UNIT_MANGLED_NAME_LENGTH +
					  2); // '/' and '\0'
  /* Try to open the file: if we fail then just return NULL, else make a structure: */
  sprintf(output_file_name, "%s/%s.c",
	  output_directory,
	  compilation_unit->mangled_name);
  FILE *output_file = fopen(output_file_name, "w");
  free(output_file_name);
  if(output_file == NULL)
    return NULL;
  else{
    r = (struct eam_translation_context*)eam_xmalloc(sizeof(struct eam_translation_context));
    r->compilation_unit = (struct eam_compilation_unit*)compilation_unit;
    r->symbol_table = NULL;
    r->output_stream = output_file;
    return r;
  } // else
}

static void finalize_translation(struct eam_translation_context *translation_context){
  /* Destroy everything referred by the translation context, except the symbol
     table, which was supplied by the caller and we don't touch: */
  eam_destroy_compilation_unit(translation_context->compilation_unit);
  fclose(translation_context->output_stream);
  
  /* Destroy the structure itself: */
  free(translation_context);
}

static void translate_expression(const struct eam_expression *expression,
				 const struct eam_translation_context *translation_context);

static integer_t translate_expressions_into_temporaries(
   struct eam_list *expressions,
   const struct eam_translation_context *translation_context){
  integer_t argument_index = 0;
  while(! eam_null(expressions)){
    /* We found one more argument: */
    argument_index++;
    
    /* Translate the first expression: */
    struct eam_expression *first = eam_car(expressions);
    fprintf(translation_context->output_stream,
            "  register word_t __"INTEGER_T_FORMAT" = (word_t)({\n  ",
            argument_index);
    translate_expression(first, translation_context);
    fprintf(translation_context->output_stream, "\n;});\n");
    
    /* Translate the remaining expressions: */
    expressions = eam_cdr(expressions);
  } // while
  return argument_index;
}

static void translate_primitive_call(const struct eam_primitive_call *primitive_call,
				     const struct eam_translation_context *translation_context){
  // To do: check arity
  integer_t i;
  const integer_t arguments_no = 
    translate_expressions_into_temporaries(primitive_call->operands,
                                           translation_context);
  fprintf(translation_context->output_stream,
          "%s(", lookup_into_string(primitive_call->operator,
                                    translation_context));
  for(i = 0; i < arguments_no; i++)
    fprintf(translation_context->output_stream,
            "__"INTEGER_T_FORMAT"%s",
            i + 1,
            ((i + 1) < arguments_no) ? ", " : "");
  fprintf(translation_context->output_stream, ")");
}

static void translate_left_expression(const struct eam_expression*,
                                      const struct eam_translation_context*);

static void translate_left_primitive_call(
   const struct eam_primitive_call *left_primitive_call,
   const struct eam_translation_context *translation_context){
  fprintf(translation_context->output_stream,
          "%s(", lookup_into_string(left_primitive_call->operator,
                                    translation_context));
  const struct eam_list *p;
  for(p = left_primitive_call->operands; ! eam_null(p); p = eam_cdr(p)){
    translate_left_expression((const struct eam_expression*)eam_car(p),
                              translation_context);
    fprintf(translation_context->output_stream,
            "%s",
            ((eam_null(eam_cdr(p))) ? "\n" : "\n,\n"));
  } // for
  fprintf(translation_context->output_stream, ")");
}

static void translate_global(const char *global,
			     const struct eam_translation_context *translation_context){
  fprintf(translation_context->output_stream, "%s", lookup_into_string(global, translation_context));
}

static void translate_register(
   const struct eam_register *register_,
   const struct eam_translation_context *translation_context){
  switch(register_->type){
  case eam_indexed_register_register_type:{
    /* Named registers must only be referred by their name: */
    if(eam_is_register_named_in_symbol_table(
          translation_context->symbol_table,
          register_->register_no))
      eam_fatal("the register %%r"INTEGER_T_FORMAT
                " should be referred by its name %%%s",
                register_->register_no,
                eam_register_to_name_in_symbol_table(
                   translation_context->symbol_table,
                   register_->register_no));
    fprintf(translation_context->output_stream,
            "eam_register_"INTEGER_T_FORMAT,
            register_->register_no);
    break;
  }
  case eam_named_register_register_type:{
    if(! eam_is_register_name_bound_in_symbol_table(
            translation_context->symbol_table,
            register_->register_name))
      eam_fatal("the register %%%s is unknown", register_->register_name);
    if(eam_name_to_register_in_symbol_table(translation_context->symbol_table,
                                            register_->register_name) == -1){
      /* The name is bound to a memory pseudo-register: */
      char *register_mangled_name = eam_mangle(register_->register_name);
      fprintf(translation_context->output_stream, "/*%s*/eam_register_%s",
              register_->register_name,
              register_mangled_name);
      free(register_mangled_name);
    }
    else{
      /* The name is bound to an indexed register */
      fprintf(translation_context->output_stream,
              "/*%s*/eam_register_"INTEGER_T_FORMAT,
              register_->register_name,
              eam_name_to_register_in_symbol_table(
                 translation_context->symbol_table,
                 register_->register_name));
    } //else
    break;
  }
  default:
    assert(0);
  }; // switch
}

static void translate_expression(const struct eam_expression *expression,
				 const struct eam_translation_context *translation_context){
  switch(expression->type){
  case eam_immediate_expression_type:{
    fprintf(translation_context->output_stream,
            "(word_t)(%s)",
	    expression->immediate->value);
    break;
  }
  case eam_register_expression_type:{
    translate_register(expression->register_, translation_context);
    break;
  }
  case eam_label_expression_type:{
    fprintf(translation_context->output_stream,
            "%s",
            lookup_into_string(expression->label, translation_context));
    break;
  }
  case eam_global_expression_type:{
    translate_global(expression->global, translation_context);
    break;
  }
  case eam_primitive_call_expression_type:{
    translate_primitive_call(expression->primitive_call, translation_context);
    break;
  }
  default:
    assert(0);
  } // switch
}

static void translate_left_expression(const struct eam_expression *left_expression,
				      const struct eam_translation_context *translation_context){
  switch(left_expression->type){
  case eam_register_expression_type:{
    translate_register(left_expression->register_, translation_context);
    break;
  }
  case eam_global_expression_type:{
    translate_global(left_expression->global, translation_context);
    break;
  }
  case eam_primitive_call_expression_type:{
    translate_left_primitive_call(left_expression->primitive_call,
                                  translation_context);
    break;
  }
  default:
    translate_expression(left_expression, translation_context);
  } // switch
}

/* Generate a fast jump when possible, otherwise generate a slow jump: */
static void translate_jump_to(
   const struct eam_expression *target_expression,
   const struct eam_translation_context *translation_context){
  if(target_expression->type == eam_label_expression_type){
    /* Jump to a known address. We may be able to make a fast jump in this
       case: */
    struct eam_symbol_value *symbol_value = 
      eam_lookup_symbol_table(translation_context->symbol_table,
			  target_expression->label);
    char *label_as_string =
      lookup_into_string(target_expression->label, translation_context);
    fprintf(translation_context->output_stream, "EAM_JUMP_TO_%s(%s);\n",
	    symbol_value->compilation_unit_mangled_name,
	    label_as_string);
  }
  else{
    /* Jump to an adress computed with an expression: we have to jump thru
       the driver in this case: */
    fprintf(translation_context->output_stream, "EAM_SLOW_JUMP(({");
    translate_expression(target_expression, translation_context);
    fprintf(translation_context->output_stream, "\n;}));\n");
  } // else
}

static void translate_command(const struct eam_command *command,
			      const struct eam_translation_context *translation_context){
  switch(command->type){
  case eam_label_command_type:{
    fprintf(translation_context->output_stream, "  EAM_FAST_JUMP(%s);  // proceed to the next block\n",
	    lookup_into_string(command->label_command.label,
			       translation_context));
    fprintf(translation_context->output_stream, "EAM_END_BLOCK\n");
    /* To do: this is a temporary kludge: begin. */
    if(! strcmp(command->label_command.label, "main:")){
      fprintf(translation_context->output_stream,
	      "integer_t eam_argc; // declared in runtime.h\nchar **eam_argv; // declared in runtime.h\n");
      fprintf(translation_context->output_stream,
              "EAM_BEGIN_INTERNAL_BLOCK(initialize_and_go_to_initial_block)\n  eam_%s_"EAM_INITIALIZE_ROOT"();\n  EAM_FAST_JUMP(%s_initial_block);\n",
	      translation_context->compilation_unit->mangled_name,
	      translation_context->compilation_unit->mangled_name);
      fprintf(translation_context->output_stream,
              "EAM_END_BLOCK\n");
      fprintf(translation_context->output_stream,
              "int main(int argc, char **argv){\n  eam_argc = (integer_t)argc;\n  eam_argv = argv;\n  integer_t eam_exit_code = eam_driver((const eam_continuation_t)initialize_and_go_to_initial_block);\n  eam_%s_"FINALIZE_ROOT"();\n  return eam_exit_code;\n}\n",
	      translation_context->compilation_unit->mangled_name);
    } // if
    /* To do: this is a temporary kludge: end. */
    fprintf(translation_context->output_stream, "EAM_BEGIN%sBLOCK(%s)\n",
	    (command->label_command.exported ? "_EXPORTED_" : "_INTERNAL_"),
	    lookup_into_string(command->label_command.label,
			       translation_context));
    break;
  }
  case eam_primitive_call_command_type:{
    fprintf(translation_context->output_stream, "{  // primitive call: begin\n");
    translate_primitive_call(command->primitive_call_command.primitive_call,
			     translation_context);
    fprintf(translation_context->output_stream, "\n;}; // primitive call: end\n");
    break;
  }
  case eam_assignment_command_type:{
    fprintf(translation_context->output_stream, "{ // assignment: begin\n  // left side: begin\n");
    translate_left_expression(command->assignment_command.left_expression,
			      translation_context);
    fprintf(translation_context->output_stream, "\n  // left side: end\n=\n(word_t)({ // right side: begin\n");
    translate_expression(command->assignment_command.right_expression,
			 translation_context);
    fprintf(translation_context->output_stream, "\n;});// right side: end\n} // assignment: end\n");
    break;
  }
  case eam_goto_command_type:{
    fprintf(translation_context->output_stream, "  ");
    translate_jump_to(command->goto_command.target_expression,
		      translation_context);
    break;
  }
  case non_tail_recursive_eam_goto_command_type:{
    fprintf(translation_context->output_stream, "  EAM_NON_TAIL_JUMP(({");
    translate_expression(command->non_tail_recursive_goto_command.target_expression,
			 translation_context);
    fprintf(translation_context->output_stream, "\n;}));\n");
    break;
  }
  case eam_if_command_type:{
    fprintf(translation_context->output_stream, "  if(({");
    translate_expression(command->if_command.guard_expression,
			 translation_context);
    fprintf(translation_context->output_stream, "\n;}))\n    ");
    translate_jump_to(command->if_command.target_expression,
		      translation_context);
    break;
  }
  default:
    assert(0);
  } // switch
}

static void translate_commands(struct eam_list *commands,
			       const struct eam_translation_context *translation_context){
  fprintf(translation_context->output_stream, "/* The conventional \"initial\" block: begin. */\n");
  fprintf(translation_context->output_stream,
	  "EAM_BEGIN_INTERNAL_BLOCK(%s_initial_block)\n",
	  translation_context->compilation_unit->mangled_name);
  while(! eam_null(commands)){
    /* Translate the first command: */
    struct eam_command *first = eam_car(commands);
    translate_command(first, translation_context);
    
    /* Translate the remaining commands: */
    commands = eam_cdr(commands);
  } // while
  fprintf(translation_context->output_stream, "EAM_END_BLOCK\n");
  fprintf(translation_context->output_stream, "/* The conventional \"initial\" block: end. */\n");
}

static void translate_register_naming(
   const struct eam_register_naming *naming,
   const integer_t declaration,
   const struct eam_translation_context *translation_context){
  switch(naming->definition->type){
  case eam_indexed_register_register_definition_type:{
    fprintf(translation_context->output_stream,
            "/* Nothing is needed for %%%s, defined as an indexed register */\n",
            naming->name);
    /* Do nothing: indexed registers have already been declared and defined
       in registers.h and registers.c.*/
    break;
  } // case
  case eam_memory_register_definition_type:{
    /* Memory pseudo-register: */
    char *register_mangled_name = eam_mangle(naming->name);
    fprintf(translation_context->output_stream,
            "%sword_t /* %%%s */eam_register_%s;\n",
            declaration ? "extern " : "",
            naming->name,
            register_mangled_name);
    free(register_mangled_name);
    break;
  } // case
  } // switch
}

static void translate_register_namings(
   const struct eam_list *namings,
   const integer_t declaration, // declaration or definition?
   const struct eam_translation_context *translation_context){
  const struct eam_list *p;
  fprintf(translation_context->output_stream, "/* Register naming %s: begin: */\n",
          declaration ? "declarations" : "definitions");
  for(p = namings; ! eam_null(p); p = eam_cdr(p))
    translate_register_naming(eam_car(p), declaration, translation_context);
  fprintf(translation_context->output_stream, "/* Register naming %s: end. */\n\n",
          declaration ? "declarations" : "definitions");
}

static void translate_primitive_declaration(const struct eam_primitive_definition *primitive,
					    const struct eam_translation_context *translation_context){
  fprintf(translation_context->output_stream,
	  "/* Declaration of the primitive %s: begin */\n",
	  primitive->name);
  if(strlen(primitive->c_declaration) != 0)
    fprintf(translation_context->output_stream, "%s\n", primitive->c_declaration);
  fprintf(translation_context->output_stream, "#define %s %s\n",
	  lookup_into_string(primitive->name, translation_context),
	  primitive->c_name);
  fprintf(translation_context->output_stream,
	  "/* Declaration of the primitive %s: end */\n\n",
	  primitive->name);
}

static void translate_primitive_implementation(const struct eam_primitive_definition *primitive,
					       const struct eam_translation_context *translation_context){
  fprintf(translation_context->output_stream, "/* eAML primitive %s implementation: begin */\n",
	  primitive->name);
  if(strlen(primitive->c_implementation) != 0)
    fprintf(translation_context->output_stream, "%s\n", primitive->c_implementation);
  fprintf(translation_context->output_stream, "/* eAML primitive %s implementation: end */\n\n",
	  primitive->name);
}

static void translate_global_definition_declaration(const struct eam_global_definition *global_definition,
						    const struct eam_translation_context *translation_context){
  /* If the global is not exported then use a static qualifier also here. The C
     language allows declaring a static variable more than once, when declarations
     don't conflict: */
  fprintf(translation_context->output_stream,
	  "/* Declaration of the global %s: */\n",
	  global_definition->name);
  fprintf(translation_context->output_stream,
	  "%s %sword_t %s;\n",
	  (global_definition->exported ? "extern" : "static"),
	  (global_definition->constant ? "const " : ""),	
	  lookup_into_string(global_definition->name, translation_context));
}

static void translate_global_definition_implementation(const struct eam_global_definition *global_definition,
						       const struct eam_translation_context *translation_context){
  fprintf(translation_context->output_stream,
	  "/* Definition of the global %s: begin */\n",
	  global_definition->name);
  fprintf(translation_context->output_stream,
	  "%s%sword_t %s;\n",
	  (global_definition->exported ? "" : "static "),
	  (global_definition->constant ? "const " : ""),
	  lookup_into_string(global_definition->name, translation_context));
  fprintf(translation_context->output_stream,
	  "/* Definition of the global %s: end */\n\n",
	  global_definition->name);
}

/* Pay attention to the name of the third parameter. It is 'exported', and not
   'only_exported'. The caller can choose to either translate only exported
   definitions, or only non-exported definitions. This is different from other
   procedures in this C file. */
static void translate_definition_declaration(const struct eam_definition *definition,
					     const struct eam_translation_context *translation_context,
					     const int exported){
  switch(definition->type){
  case eam_immediate_type_definition_type:{
    fprintf(translation_context->output_stream, "#error To do: translate constant type definition declarations\n");
    break;
  }
  case eam_global_definition_type:{
    /* I use nxor for comparison between booleans: */
    if(! (definition->global_definition->exported ^ exported))
      translate_global_definition_declaration(definition->global_definition,
					      translation_context);
    break;
  }
  case eam_primitive_definition_type:{
    /* I use nxor for comparison between booleans: */
    if(! (definition->primitive_definition->exported ^ exported))
      translate_primitive_declaration(definition->primitive_definition,
				      translation_context);
    break;
  }
  default:
    assert(0);
  } // switch
}

static void translate_definition_implementation(
   const struct eam_definition *definition,
   const struct eam_translation_context *translation_context){
  switch(definition->type){
  case eam_immediate_type_definition_type:{
    fprintf(translation_context->output_stream, "#error To do: translate immediate type definition implementation\n");
    break;
  }
  case eam_global_definition_type:{
    translate_global_definition_implementation(definition->global_definition,
					       translation_context);
    break;
  }
  case eam_primitive_definition_type:{
    translate_primitive_implementation(definition->primitive_definition,
				       translation_context);
    break;
  }
  default:
    assert(0);
  } // switch
}

static void output_header(const struct eam_translation_context *translation_context){
  /* Output initial comments, which are common to all translated compilation units: */
  fprintf(translation_context->output_stream, "/* This file is automatically generated. You should not edit it. */\n\n");
  fprintf(translation_context->output_stream, "/* Required non-eAM libraries are: \n");
  struct eam_list *p;
  for(p = translation_context->compilation_unit->non_eam_libraries; 
      ! eam_null(p);
      p = eam_cdr(p))
    fprintf(translation_context->output_stream, "* %s\n", (char*)eam_car(p));
  fprintf(translation_context->output_stream, "*/\n\n");

  fprintf(translation_context->output_stream, "/* This is the translation of the compilation unit \"%s\" into C; \n",
	  translation_context->compilation_unit->name);
  fprintf(translation_context->output_stream, "   the compilation unit mangled name is \"%s\". */\n\n",
	  translation_context->compilation_unit->mangled_name);
}

static void output_blocks_forward_declarations(const int exported,
					       const struct eam_translation_context *translation_context){
  const struct eam_list *commands = translation_context->compilation_unit->commands;
  if(exported)
    fprintf(translation_context->output_stream, "/* Forward declarations of exported blocks: begin. */\n");
  else
    fprintf(translation_context->output_stream, "/* Forward declarations of internal blocks: begin. */\n");

  /* Output the block declarations for each eAML label: */
  while(eam_neam_null(commands)){
    struct eam_command *command = eam_car(commands);
    if(command->type == eam_label_command_type){
      if(! (command->label_command.exported ^ exported)) // compare booleans with nxor
	fprintf(translation_context->output_stream,
		"EAM_DECLARE%sBLOCK(%s);\n",
		(exported ? "_EXPORTED_" : "_INTERNAL_"),
		lookup_into_string(command->label_command.label, translation_context));
    } // if
    commands = eam_cdr(commands);
  } // while
  
  if(exported)
    fprintf(translation_context->output_stream, "/* Forward declarations for exported blocks: end. */\n\n");
  else
    fprintf(translation_context->output_stream, "/* Forward declarations for internal blocks: end. */\n\n");
}

static void output_declarations_part(const struct eam_translation_context *translation_context){
  /* Protect the declarations part from multiple inclusion: begin: */
  fprintf(translation_context->output_stream, "/* ************* Here comes the declarations part ************* */\n");
  fprintf(translation_context->output_stream, "#ifndef _%s_h_\n", translation_context->compilation_unit->mangled_name);
  fprintf(translation_context->output_stream, "#define _%s_h_\n\n", translation_context->compilation_unit->mangled_name);
  
  /* The header we always include for compiling what should be linked to the 
     runtime; this takes care of types, registers, and the driver: */
  fprintf(translation_context->output_stream, "#include \"runtime.h\"\n\n");
  
  fprintf(translation_context->output_stream, "/* Of course we're compiling \"also\" this one: */\n#ifndef EAM_COMPILING_ALSO_%s\n#define EAM_COMPILING_ALSO_%s\n#endif\n\n",
          translation_context->compilation_unit->mangled_name,
          translation_context->compilation_unit->mangled_name);
  
  /* Declaration of a variable holding the unmangled compilation unit name. Useful
     for debugging: */
  fprintf(translation_context->output_stream, "/* The compilation unit unmangled name: */\nextern const char* eam_%s_unmangled_name; // \"%s\";\n\n",
	  translation_context->compilation_unit->mangled_name,
	  translation_context->compilation_unit->name);
  
  /* Force #inclusion of the declaration part of all imported compilation units. */
  fprintf(translation_context->output_stream, "/* Force imported compilation units to be used: begin */\n");
  const struct eam_list *rest = translation_context->compilation_unit->imports;
  while(eam_neam_null(rest)){
    struct eam_unmangled_and_mangled *unmangled_and_mangled =
      (struct eam_unmangled_and_mangled*) eam_car(rest);
    fprintf(translation_context->output_stream,
	    "#define _%s_c_ // Don't include the implementation part\n",
	    unmangled_and_mangled->mangled);
    fprintf(translation_context->output_stream, 
	    "#include \"%s.c\"\n", 
	    unmangled_and_mangled->mangled);
    rest = eam_cdr(rest);
  } // while
  fprintf(translation_context->output_stream, "/* Force imported compilation units to be used: end */\n\n");
  
  /* Declare named registers: */
  translate_register_namings(translation_context->compilation_unit->register_namings,
                             1, // declaration
                             translation_context);
  
  /* Output the declaration of two procedures to be called at load time and at
     unload time, respectively: */
  fprintf(translation_context->output_stream, "/* Load/unload procedures: begin */\n");
  fprintf(translation_context->output_stream, "void eam_%s_"EAM_INITIALIZE_ROOT"(void);\n", 
	  translation_context->compilation_unit->mangled_name);
  fprintf(translation_context->output_stream, "void eam_%s_"FINALIZE_ROOT"(void);\n",
	  translation_context->compilation_unit->mangled_name);
  fprintf(translation_context->output_stream, "/* Load/unload procedures: end */\n\n");

  /* Output declarations for exported blocks: */
  output_blocks_forward_declarations(1, translation_context);
  
  /* Output the C initialization declarations for this compilation unit: */
  fprintf(translation_context->output_stream, "/* Initialization code: begin. */\n");
  if(translation_context->compilation_unit->initialization != NULL)
    fprintf(translation_context->output_stream, "%s\n",
	    translation_context->compilation_unit->initialization->c_declaration);
  fprintf(translation_context->output_stream, "/* Initialization code: end. */\n\n");
  
  /* Output the C finalization declarations for this compilation unit: */
  fprintf(translation_context->output_stream, "/* Finalization code: begin. */\n");
  if(translation_context->compilation_unit->finalization != NULL)
    fprintf(translation_context->output_stream, "%s\n", 
	    translation_context->compilation_unit->finalization->c_declaration);
  fprintf(translation_context->output_stream, "/* Finalization code: end. */\n\n");
  
  /* Output definition declarations: */
  fprintf(translation_context->output_stream, "/* Definitions: begin. */\n");
  const struct eam_list *definitions = translation_context->compilation_unit->definitions;
  while(! eam_null(definitions)){
    /* Translate the first definition: */
    translate_definition_declaration(eam_car(definitions),
				     translation_context,
				     1); // only the exported part
    
    /* Translate the remaining stuff: */
    definitions = eam_cdr(definitions);
  } // while
  fprintf(translation_context->output_stream, "/* Definitions: end. */\n\n");
  
  /* End of the declaration part: close the #ifdef above which protected the 
     declaration part from multiple inclusion: */
  fprintf(translation_context->output_stream, "#endif //#ifndef _%s_h_\n\n", translation_context->compilation_unit->mangled_name);
}
  
static void output_implementation_part(const struct eam_translation_context *translation_context){
  /* Now we start with the implementation part, which is protected from multiple
     inclusion with a macro which is distinct from the one used for the declaration
     part. This is a simple extension of the usual C preprocessor trick used to
     prevent multiple inclusion of header files.
     The idea is saving us the pain of dealing with two different generated files; 
     we can just use the .c file also as a header, just #defining a macro before
     #including it: */
  fprintf(translation_context->output_stream, "/* ************* Here comes the implementation part ************* */\n");
  fprintf(translation_context->output_stream, "#ifndef _%s_c_\n", 
	  translation_context->compilation_unit->mangled_name);
  fprintf(translation_context->output_stream, "#define _%s_c_\n\n", 
	  translation_context->compilation_unit->mangled_name);
  
  /* Give a value to the name we declared above: */
  fprintf(translation_context->output_stream,
          "/* The compilation unit unmangled name: */\nconst char* eam_%s_unmangled_name =\n  \"%s\";\n\n",
	  translation_context->compilation_unit->mangled_name,
	  translation_context->compilation_unit->name);

  fprintf(translation_context->output_stream,
          "/* How many globals there are: */\nconst integer_t eam_%s_globals_no =\n  %i;\n\n",
          translation_context->compilation_unit->mangled_name,
          (int)(compilation_unit_to_globals_no(translation_context->compilation_unit)));

  fprintf(translation_context->output_stream,
          "/* How many blocks (i.e. labels) there are: */\nconst integer_t eam_%s_blocks_no =\n  %i;\n\n",
          translation_context->compilation_unit->mangled_name,
          (int)(compilation_unit_to_blocks_no(translation_context->compilation_unit)));

  fprintf(translation_context->output_stream, "/* How many primitives there are: */\nconst integer_t eam_%s_primitives_no =\n  %i;\n\n",
          translation_context->compilation_unit->mangled_name,
          (int)(compilation_unit_to_primitives_no(translation_context->compilation_unit)));

  
  /* Generate code to make jumps local or foreign, according to the way the other
     compilation units are linked: */
  fprintf(translation_context->output_stream,
	  "/* Make jumps local or foreign, according to the way the other compilation units\n"
	  "   are linked: */\n");
  const struct eam_list *compilation_units_names =
    eam_symbol_table_to_mangled_compilation_units(translation_context->symbol_table);
  const struct eam_list* cu;
  for(cu = compilation_units_names; ! eam_null(cu); cu = eam_cdr(cu)){
    char *mangled_name = (char*)eam_car(cu);
    fprintf(translation_context->output_stream,
	    "#ifdef EAM_COMPILING_ALSO_%s\n#define EAM_JUMP_TO_%s(label) EAM_FAST_JUMP(label)\n#else\n#define EAM_JUMP_TO_%s(label) EAM_SLOW_JUMP(label)\n#endif //#ifdef EAM_COMPILING_ALSO_%s\n",
	    mangled_name, mangled_name, mangled_name, mangled_name);
  } // for
  fprintf(translation_context->output_stream, "\n");

  /* Declare named registers: */
  translate_register_namings(translation_context->compilation_unit->register_namings,
                             0, // implementation
                             translation_context);
  
  /* Output the implementation part of the initialization code: */
  fprintf(translation_context->output_stream, "/* Initialization code: begin. */\n");
  if(translation_context->compilation_unit->initialization != NULL){
    fprintf(translation_context->output_stream, "%s\n", 
	    translation_context->compilation_unit->initialization->c_declaration);
    fprintf(translation_context->output_stream, "%s\n", 
	    translation_context->compilation_unit->initialization->c_implementation);
  }
  fprintf(translation_context->output_stream, "/* Initialization code: end. */\n\n");
  
  /* Output the implementation part of the finalization code: */
  fprintf(translation_context->output_stream, "/* Finalization code: begin. */\n");
  if(translation_context->compilation_unit->finalization != NULL){
    fprintf(translation_context->output_stream, "%s\n",
	    translation_context->compilation_unit->finalization->c_declaration);
    fprintf(translation_context->output_stream, "%s\n",
	    translation_context->compilation_unit->finalization->c_implementation);
  }
  fprintf(translation_context->output_stream, "/* Finalization code: end. */\n\n");

  /* Output the user-defined initialization and finalization procedures: */
  fprintf(translation_context->output_stream, "/* User-defined initialization procedure: */\nstatic void eam_%s_user_defined_initialization_procedure(void){\n",
          translation_context->compilation_unit->mangled_name);
  if(translation_context->compilation_unit->initialization != NULL)
    fprintf(translation_context->output_stream, "%s\n", 
	    translation_context->compilation_unit->initialization->c_procedure_body);
  fprintf(translation_context->output_stream, "}\n\n");
  fprintf(translation_context->output_stream, "/* User-defined finalization procedure: */\nstatic void eam_%s_user_defined_finalization_procedure(void){\n",
          translation_context->compilation_unit->mangled_name);
  if(translation_context->compilation_unit->finalization != NULL)
    fprintf(translation_context->output_stream, "%s\n", 
	    translation_context->compilation_unit->finalization->c_procedure_body);
  fprintf(translation_context->output_stream, "}\n\n");

  /* Output the implementation of the procedures to be called at load [unload]
     time, which force the loading [unloading] of the imported compilation units: */
  fprintf(translation_context->output_stream, "/* Force linking of imported compilation units: */\n");
  fprintf(translation_context->output_stream, 
	  "static integer_t %s"EAM_PINS_NO_ROOT" = (integer_t)0;\n",
	  translation_context->compilation_unit->mangled_name);
  fprintf(translation_context->output_stream, "void eam_%s_"EAM_INITIALIZE_ROOT"(void){\n", 
	  translation_context->compilation_unit->mangled_name);
  
  fprintf(translation_context->output_stream, 
	  "  /* Do nothing but add a pin if we already initialized this compilation\n"
	  "     unit: */\n"
	  "  if(++ %s"EAM_PINS_NO_ROOT" > (integer_t)1)\n    return;\n",
	  translation_context->compilation_unit->mangled_name);
  
  struct eam_list *rest = translation_context->compilation_unit->imports;
  while(eam_neam_null(rest)){
    struct eam_unmangled_and_mangled *unmangled_and_mangled =
      (struct eam_unmangled_and_mangled*) eam_car(rest);
    fprintf(translation_context->output_stream, 
	    "  eam_%s_"EAM_INITIALIZE_ROOT"();\n",
	    unmangled_and_mangled->mangled);
    rest = eam_cdr(rest);
  } // while
  fprintf(translation_context->output_stream,
          "  /* Call user-defined code: */\n  eam_%s_user_defined_initialization_procedure();\n",
          translation_context->compilation_unit->mangled_name);

#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(translation_context->output_stream,
          "  printf(\"%s was initialized.\\n\");\n", 
	  translation_context->compilation_unit->name);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  fprintf(translation_context->output_stream, "}\n");
  fprintf(translation_context->output_stream, "void eam_%s_"FINALIZE_ROOT"(void){\n",
	  translation_context->compilation_unit->mangled_name);
  /* ////////
  fprintf(translation_context->output_stream, 
	  "  printf(\"Pins for %s: %%i --> %%i\\n\", %s"EAM_PINS_NO_ROOT", %s"EAM_PINS_NO_ROOT" - 1);\n",
	  translation_context->compilation_unit->mangled_name,
	  translation_context->compilation_unit->mangled_name,
	  translation_context->compilation_unit->mangled_name);
  */ /////////

  fprintf(translation_context->output_stream, 
	  "  /* Do nothing but remove a pin if this compilation unit is still needed: */\n"
	  "  if(-- %s"EAM_PINS_NO_ROOT" > (integer_t)0)\n    return;\n",
	  translation_context->compilation_unit->mangled_name);
  fprintf(translation_context->output_stream,
          "  /* Call user-defined code: */\n  eam_%s_user_defined_finalization_procedure();\n",
          translation_context->compilation_unit->mangled_name);
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(translation_context->output_stream, "  printf(\"%s was finalized.\\n\");\n",
	  translation_context->compilation_unit->name);
#endif // #ifdef ENABLE_VERBOSE_DEBUG
  rest = eam_reverse(translation_context->compilation_unit->imports);
  while(eam_neam_null(rest)){
    struct eam_unmangled_and_mangled *unmangled_and_mangled =
      (struct eam_unmangled_and_mangled*) eam_car(rest);
    fprintf(translation_context->output_stream, "  eam_%s_"FINALIZE_ROOT"();\n",
	    unmangled_and_mangled->mangled);
    rest = eam_cdr(rest);
  } // while 
  eam_destroy_list_structure(rest);
  fprintf(translation_context->output_stream, "}\n\n");
  
  /* Output declarations for internal blocks: */
  output_blocks_forward_declarations(0, translation_context);
  
  /* Output declarations for non-exported definitions: we only declared exported
     definitions in the declaration part. */
  fprintf(translation_context->output_stream, "/* Non-exported definitions declaration: begin. */\n");
  const struct eam_list *definitions = 
    translation_context->compilation_unit->definitions;
  while(! eam_null(definitions)){
    /* Translate the first definition: */
    translate_definition_declaration(eam_car(definitions),
				     translation_context,
				     0); // only the non-exported part
    
    /* Translate the remaining stuff: */
    definitions = eam_cdr(definitions);
  } // while
  fprintf(translation_context->output_stream, "/* Non-exported definitions declaration: end. */\n\n");

  /* Output definitions implementation: */
  fprintf(translation_context->output_stream, "/* Definitions: begin. */\n");
  definitions = translation_context->compilation_unit->definitions;
  while(! eam_null(definitions)){
    /* Translate the first definition: */
    translate_definition_implementation(eam_car(definitions), translation_context);
    
    /* Translate the remaining stuff: */
    definitions = eam_cdr(definitions);
  } // while
  fprintf(translation_context->output_stream, "/* Definitions: end. */\n\n");
  
  /* Translate code: */
  translate_commands(translation_context->compilation_unit->commands,
		     translation_context);
  
  /* End of the definition part: close the #ifdef above: */
  fprintf(translation_context->output_stream, "\n#endif //#ifndef _%s_c_\n",
	  translation_context->compilation_unit->mangled_name);
}

/* Translate the given compilation unit in the given file using the given
   already-filled symbol table, into a C file to be written in the given
   directory.
   Fail with a fatal error on any error condition: */
void eam_generate_c_file(const struct eam_compilation_unit *compilation_unit,
		     const struct eam_symbol_table *symbol_table,
		     const char *output_directory_for_c_file){
  /* The compilation unit must be non-null: */
  if(compilation_unit == NULL)
    eam_fatal("eam_generate_c_file(): NULL compilation unit");
  
  /* Do the initialization stuff before translating, and fail with a fatal error
     if something goes wrong: */
  struct eam_translation_context *translation_context =
    initialize_translation(compilation_unit, output_directory_for_c_file);
  if(translation_context == NULL){
    fprintf(stderr, "About the translation of the compilation unit \"%s\":\n",
	    compilation_unit->name);
    eam_fatal("error initializing the translation");
  } // if
  
  /* Just copy the pointer to the given symbol table. It must contain all the
     needed symbols, non just the exported ones: */
  translation_context->symbol_table = (struct eam_symbol_table*)symbol_table;
  
  /* Output the compilation unit header: */
  output_header(translation_context);
  
  /* Output the declaration part: */
  output_declarations_part(translation_context);
  
  /* Output the implementation part: */
  output_implementation_part(translation_context);
  
  /* Destroy the temporary data structures we used for translation: */
  finalize_translation(translation_context);
}
