/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#include <stdio.h>
#include "../common/types.h"
#include "../common/filesystem.h"
#include "../common/malloc.h"
#include "../common/command_line.h"
#include "generate_including_c.h"
#include "compile_time_parameters.h"
#include <string.h>

void eam_generate_single_including_c_file(const struct eam_list *compilation_units,
                                          const char *temporary_directory){
  /* "e" is a name which is never obtained by mangling: we use it as the name
     of our 'artificial' main C file #including the other ones: */
  char *file_name = (char*)eam_xmalloc(strlen(temporary_directory) + 100);
  sprintf(file_name, "%s/e.c", temporary_directory);
  FILE *f = fopen(file_name, "w");
  if(f == NULL)
    eam_fatal("could not create the file %s", file_name);
  
  const struct eam_list *rest;
  fprintf(f, "/* This file is automatically generated. Don't edit it. */\n\n");
  fprintf(f, "/* All jumps to labels within #included compilation units should be fast: */\n");
  for(rest = compilation_units; ! eam_null(rest); rest = eam_cdr(rest))
    fprintf(f, "#define EAM_COMPILING_ALSO_%s\n",
	    ((struct eam_compilation_unit*)eam_car(rest))->mangled_name);
  fprintf(f, "\n/* #include declaration parts only: */\n");
  for(rest = compilation_units; ! eam_null(rest); rest = eam_cdr(rest))
    fprintf(f, "#define _%s_c_\n",
	    ((struct eam_compilation_unit*)eam_car(rest))->mangled_name);
  for(rest = compilation_units; ! eam_null(rest); rest = eam_cdr(rest))
    fprintf(f, "#include \"%s.c\"\n",
	    ((struct eam_compilation_unit*)eam_car(rest))->mangled_name);
  fprintf(f, "\n/* #include implementation parts (only): */\n");
  for(rest = compilation_units; ! eam_null(rest); rest = eam_cdr(rest))
    fprintf(f, "#undef _%s_c_\n",
	    ((struct eam_compilation_unit*)eam_car(rest))->mangled_name);
  for(rest = compilation_units; ! eam_null(rest); rest = eam_cdr(rest))
    fprintf(f, "#include \"%s.c\"\n",
	    ((struct eam_compilation_unit*)eam_car(rest))->mangled_name);
  
  fclose(f);
  free(file_name);
}

/* Output a C file #including the files needed for a single set. Used by 
   eam_generate_including_c_files(): */
static void output_set(const char *temporary_directory,
                       int set_index,
                       const struct eam_list *first_compilation_unit,
                       int set_length){
  const struct eam_list *p;
  int i;
  char *file_name = (char*)eam_xmalloc(strlen(temporary_directory) + 100);
  sprintf(file_name, "%s/_%i.c", temporary_directory, set_index);
  FILE *f = fopen(file_name, "w");
  if(f == NULL)
    eam_fatal("could not create the file %s", file_name);
  free(file_name);
  fprintf(f, "/* This file is automatically generated. Don't edit it. */\n\n");
  fprintf(f, "/* We're compiling in a single command line the following compilation units:\n");
  fprintf(f, "   Jumps among the blocks of the following compilation units should be fast: */\n");
  p = first_compilation_unit;
  for(i = 0; i < set_length; i++, p = eam_cdr(p)){
    struct eam_compilation_unit *first = eam_car(p);
    fprintf(f, "#define EAM_COMPILING_ALSO_%s\n", first->mangled_name);
  } // for

  fprintf(f, "\n/* Include declarations from all compilation units: */\n#include \"_.h\"\n\n");

  fprintf(f, "/* Include the implementation part of all the compilation units\n   we're compiling together: */\n");
  p = first_compilation_unit;
  for(i = 0; i < set_length; i++, p = eam_cdr(p)){
    struct eam_compilation_unit *first = eam_car(p);
    fprintf(f, "#undef _%s_c_\n", first->mangled_name);
  } // for
  p = first_compilation_unit;
  for(i = 0; i < set_length; i++, p = eam_cdr(p)){
    struct eam_compilation_unit *first = eam_car(p);
    fprintf(f, "#include \"%s.c\"\n", first->mangled_name);
  } // for
  fclose(f);
}

/* Generate C files #including other generated C files, like
   generate_including_c_file(), but limiting the size of each set of files
   which are compiled (and optimized) together. The number of such sets is
   returned. The generated file names are <temporary_directory>/_i.c, where
   i is an index ranging from 0 to the returned number minus 1: */
integer_t eam_generate_including_c_files(const struct eam_list *compilation_units,
                                   const char *temporary_directory){
  /* How many sets were generated till now: */
  int sets_no = 0;
  
  /* How many compilation units are in the current set: */
  int current_set_length = 0;
  
  /* Pointer to the first compilation unit belonging to the current set: */
  const struct eam_list *set_beginning = compilation_units;
  
  /* How big is the current set, in bytes (we take the sum of the size of
     the generated C files in the set as an heuristic for their "size"): */
  long current_set_size = 0;
  FILE *f; // this is used for all the files we write (always one at the time)
  const struct eam_list *rest; // what we use to scroll compilation_units
  /* Of course we use names which would never occour as mangled names. We
     share the same buffer for all file names: */
  char *file_name = (char*)eam_xmalloc(strlen(temporary_directory) + 100);

  /* First generate a single header file "_.h" with all declarations; this
     is made available to all sets: */
  sprintf(file_name, "%s/_.h", temporary_directory);
  f = fopen(file_name, "w");
  if(f == NULL)
    eam_fatal("could not create the file %s", file_name);
  fprintf(f, "/* This file is automatically generated. Don't edit it. */\n");
  fprintf(f, "\n/* #include declaration parts only: */\n");
  for(rest = compilation_units; ! eam_null(rest); rest = eam_cdr(rest))
    fprintf(f, "#define _%s_c_\n",
	    ((struct eam_compilation_unit*)eam_car(rest))->mangled_name);
  for(rest = compilation_units; ! eam_null(rest); rest = eam_cdr(rest))
    fprintf(f, "#include \"%s.c\"\n",
	    ((struct eam_compilation_unit*)eam_car(rest))->mangled_name);
  free(file_name);
  fclose(f);
  
  /* Ok, we generated the header. Now generate the sets: */
  const struct eam_list *remaining_compilation_units = compilation_units;
  
  /* While we still have files to be put into sets... */
  while(! eam_null(remaining_compilation_units)){
    /* Gather data about the next file: */
    const struct eam_compilation_unit *first =
      eam_car(remaining_compilation_units);
    char *first_file_name = (char*)eam_xmalloc(strlen(temporary_directory) +
                                               strlen(first->mangled_name) +
                                               100);
    sprintf(first_file_name, "%s/%s.c", temporary_directory,
            first->mangled_name);
    const long first_file_size = eam_file_length(first_file_name);
    assert(first_file_size != -1);
     
    /* Add the current file to the current set: */
    /*
    fprintf(stderr, "Adding the file %s (size %i) to the set %i\n",
            first_file_name,
            (int)first_file_size,
            sets_no); */
    current_set_length++;
    current_set_size += first_file_size;
    
    /* If the set is now large enought then output it, and start a new one: */
    if((current_set_size + first_file_size) >= EAM_GENERATED_C_FILES_SET_SIZE){
      output_set(temporary_directory, sets_no, set_beginning, current_set_length);
      current_set_size = 0;
      current_set_length = 0;
      sets_no++;
      set_beginning = eam_cdr(remaining_compilation_units);
    } // if
    
    /* We're done with this file: */
    free(first_file_name);
    remaining_compilation_units = eam_cdr(remaining_compilation_units);    
  } // while

  /* Output the set we were making before exiting the loop, unless it's empty;
     in that case just forget the empty set we were building. It makes no sense
     to create an empty set: */
  if(current_set_length > 0){
    output_set(temporary_directory, sets_no, set_beginning, current_set_length);
    sets_no++;
  } // if
  
  /* Return the number of the sets we made: */
  return sets_no;
}
