/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "import_graph.h"
#include "../common/command_line.h"
#include "../common/string_map.h"
#include "../common/malloc.h"

/* A compilation unit can be either a compilation unit we parsed and we
   fully know, or simply a (non-mangled) name we got from a dumped symbol
   table, of which we can only confirm the existence. By convention there is
   at most only *one* instance of each node in memory, so two nodes can be
   compared for equality "by identity", i.e. just comparing their pointers: */
struct eam_import_graph_node{
  /* The sort of datum held by the structure: */
  enum import_graph_node_type{
    compilation_unit_import_graph_node_type,
    name_import_graph_node_type,
  } type;
  
  /* The actual datum: */
  union{
    char *name; // shared
    struct eam_compilation_unit *compilation_unit; // shared
  }; // anonymous union
  
  /* The list of imported nodes (as struct eam_import_graph_node's): note how
     pointers to each node are unique. Elements are shared: */
  struct eam_list *imported_nodes;

  /* The list of importing nodes (as struct eam_import_graph_node's): note how
     pointers to each node are unique. Elements are shared: */
  struct eam_list *importing_nodes;
  
  /* A flag used to keep track of the current "color" during the graph
     depth-first visit needed to implement topological sort and non-cyclicity
     check: */
  enum{
    undiscovered, discovered, completely_explored,
  } color;
}; // struct

/* The definition of struct eam_import_graph, which was only forward-declared
   in the header. It's a relatively standard implementation of a graph with
   adjacency lists. Note that all names used as keys are unmangled: */
struct eam_import_graph{
  /* A list of parsed eAML compilation units. The whole list is shared: */
  const struct eam_list *compilation_units;
  
  /* The shared symbol table, holding information for compilation_units and
     for the library: */
  const struct eam_symbol_table *symbol_table;
  
  /* A "cache" used to access a compilation unit node in constant time by
     name: */
  struct eam_string_map *name_to_node;
  
  /* A list of all the nodes of the graph (shared). Used to iterate on nodes. */
  struct eam_list *nodes;
}; // struct

static char* node_to_name(const struct eam_import_graph_node *node){
  if(node->type == compilation_unit_import_graph_node_type)
    return node->compilation_unit->name;
  else{
    assert(node->type == name_import_graph_node_type);
    return node->name;
  }
}

static void destroy_stuff_pointed_by_node(const struct eam_import_graph_node *node){
  /* The list elements are shared (in both cases), so we only destroy the
     structure: */
  eam_destroy_list_structure(node->imported_nodes);
  eam_destroy_list_structure(node->importing_nodes);
}

/* Return nonzero iff the object pointed by x is an element of xs: */
static int belongs_to_list(const void *x, const struct eam_list *xs){
  const struct eam_list *p = xs;
  while(! eam_null(p)){
    if(x == eam_car(p))
      return 1;
    p = eam_cdr(p);
  } // while
  return 0;
}

// To do: kill this one
void destroy_node_list_structure(const void *node){
  fprintf(stderr, "Destroying the list starting at %p\n", node);
  eam_destroy_list_structure((const struct eam_list*)node);
}

static void fill_cache_from_compilation_units(struct eam_import_graph *import_graph,
					      const struct eam_list *compilation_units){
  //fprintf(stderr, "fill_cache_from_compilation_units(): compilation_units are %i\n", eam_length(compilation_units));
  const struct eam_list *p = compilation_units;
  while(! eam_null(p)){
    /* Work on the first element: */
    struct eam_compilation_unit *first = (struct eam_compilation_unit *)eam_car(p);
    
    /* Fail with a fatal error if we already have information about the compilation
       unit; they should be always supplied exactly *once*: */
    const struct eam_import_graph_node *current_node =
      eam_access_string_map(import_graph->name_to_node, first->name);
    if(current_node != NULL){
      fprintf(stderr, 
	      "About the compilation unit %s (first as %s, then as eAML file):\n",
	      first->name,
	      ((current_node->type == compilation_unit_import_graph_node_type) ?
	       "eAML file" : "part of a library"));
      eam_command_line_fatal("compilation unit \"%s\" supplied more than once",
                             first->name);
    } // if
    
    /* Make a node and insert a binding for it: */
    struct eam_import_graph_node node;
    node.type = compilation_unit_import_graph_node_type;
    node.compilation_unit = first;
    node.imported_nodes = eam_nil; // there are no edges yet
    node.importing_nodes = eam_nil; // there are no edges yet
    node.color = undiscovered; // the node was not color yet
    eam_insert_into_string_map(import_graph->name_to_node,
			   first->name,
			   &node);
    
    /* Also add the node to the list of the nodes: */
    import_graph->nodes = eam_cons(eam_access_string_map(import_graph->name_to_node,
						 first->name),
			       import_graph->nodes);
    /*
    fprintf(stderr, 
	    "fill_cache_from_compilation_units(): import_graph->nodes is now long %i\n",
	    eam_length(import_graph->nodes));
    */
    /* Go on with the rest: */
    p = eam_cdr(p);
  } // while
}

static void fill_cache_from_symbol_table(struct eam_import_graph *import_graph,
					 const struct eam_symbol_table *symbol_table){
  const struct eam_list *p = eam_symbol_table_to_compilation_units(symbol_table);
  while(! eam_null(p)){
    /* Work on the first element: */
    char *first = (char*)eam_car(p);
    
    /* Fail with a fatal error if we already have information about the compilation
       unit; they should be always supplied exactly *once*: */
    const struct eam_import_graph_node *current_node =
      eam_access_string_map(import_graph->name_to_node, first);
    if(current_node != NULL){
      fprintf(stderr, 
	      "About the compilation unit %s (first as %s, then as eAML file):\n",
	      first,
	      ((current_node->type == compilation_unit_import_graph_node_type) ?
	       "eAML file" : "part of a library"));
      eam_fatal("compilation unit supplied more than once");
    } // if
    
    /* Make a node and insert a binding for it: */
    struct eam_import_graph_node node;
    node.type = name_import_graph_node_type;
    node.name = first;
    node.color = undiscovered; // we've not color the node yet
    node.imported_nodes = eam_nil; // there are no edges yet
    node.importing_nodes = eam_nil; // there are no edges yet
    eam_insert_into_string_map(import_graph->name_to_node, first, &node);
    
    /* Also add the node to the list of the nodes: */
    import_graph->nodes = eam_cons(eam_access_string_map(import_graph->name_to_node, first),
			       import_graph->nodes);
    
    /* Go on with the rest: */
    p = eam_cdr(p);
  } // while
}

static struct eam_import_graph_node* lookup_cache(const struct eam_import_graph *import_graph,
					      const char *name){
  /* Just lookup the table, failing with a fatal error on any reference to an
     unknown name: */
  struct eam_import_graph_node *r = 
    eam_access_string_map(import_graph->name_to_node, name);
  if(r == NULL){
    fprintf(stderr, "About the compilation unit %s:\n", name);
    eam_fatal("reference to unknown compilation unit");
  }
  else
    return r;
}

static struct eam_import_graph* create_empty_import_graph(
                               const struct eam_list *compilation_units,
			       const struct eam_symbol_table *symbol_table){
  /* Make the structure: */
  struct eam_import_graph *r =
    (struct eam_import_graph*)eam_xmalloc(sizeof(struct eam_import_graph));
  
  /* Just copy pointers to the shared structures: */
  r->compilation_units = compilation_units;
  r->symbol_table = symbol_table;
  
  /* Make the "cache" (see the comment above): */
  r->name_to_node =
    eam_create_string_map(sizeof(struct eam_import_graph_node),
		      (eam_string_map_datum_finalizer)destroy_stuff_pointed_by_node);
  
  /* The list of all nodes is just an empty list by now: */
  r->nodes = eam_nil;
  
  /* Fill the cache from the given compilation units: */
  fill_cache_from_compilation_units(r, compilation_units);
  
  /* Fill the cache from the given symbol table: */
  fill_cache_from_symbol_table(r, symbol_table);
  
  /* Return the just-created structure, which still holds no imports: */
  return r;
}

/* Add an edge to the graph. Fail with a fatal error if any of the two given 
   nodes is unknown. */
static void add_import(struct eam_import_graph *import_graph,
		       const char *importing_compilation_unit_name,
		       const char *imported_compilation_unit_name){
  /*  fprintf(stderr, "\n[%s imports %s] ",
	  importing_compilation_unit_name,
	  imported_compilation_unit_name); */
  struct eam_import_graph_node *importing_node = 
    lookup_cache(import_graph, importing_compilation_unit_name);
  struct eam_import_graph_node *imported_node = 
    lookup_cache(import_graph, imported_compilation_unit_name);
  
  /* Add imported_node to the nodes referred by importing_node, unless it's 
     already there (we want exactly one reference): */
  if(! belongs_to_list(imported_node, importing_node->imported_nodes))
    importing_node->imported_nodes = eam_cons(imported_node,
					  importing_node->imported_nodes);
  
  /* Add importing_node to the nodes referred by imported_node, unless it's 
     already there (we want exactly one reference): */
  if(! belongs_to_list(importing_node, imported_node->importing_nodes))
    imported_node->importing_nodes = eam_cons(importing_node,
					  imported_node->importing_nodes);
}

void eam_destroy_import_graph(struct eam_import_graph *import_graph){
  /* Destroy the list of nodes, but not its contents, which are shared by the
     string maps: */
  eam_destroy_list_structure(import_graph->nodes);
  
  /* Of course we don't destroy the shared structuree. Just destroy the string
     map built by us: */
  eam_destroy_string_map(import_graph->name_to_node);
  
  /* Destroy the structure: */
  free(import_graph);
}

/* A nearly canonical depth-first visit, whose only quirk is the updating of
   the given (reversed) list of topologically-sorted nodes: */
static void depth_first_visit(struct eam_import_graph_node *u,
			      struct eam_list **reversed_sorted_nodes){
  u->color = discovered;
  struct eam_list *p;
  for(p = u->imported_nodes; ! eam_null(p); p = eam_cdr(p)){
    struct eam_import_graph_node *v = (struct eam_import_graph_node*)eam_car(p);
    switch(v->color){
    case undiscovered:{
      depth_first_visit(v, reversed_sorted_nodes);
      break;
    }
    case discovered:{
      fprintf(stderr, "About the compilation unit %s:\n", node_to_name(v));
      eam_fatal("there is a cycle in the import graph");
    }
    case completely_explored:{
      /* Do nothing. */
      break;
    }
    } // switch
  } // for
  u->color = completely_explored;
  *reversed_sorted_nodes = eam_cons(u, *reversed_sorted_nodes);
}

/* Return a topologically sorted list of all the nodes (shared, as usual). Fail
   with a fatal error if there's a cycle in the import graph: */
static struct eam_list* topologically_sort_nodes(const struct eam_import_graph* import_graph){
  struct eam_list *reversed_sorted_nodes = eam_nil;

  struct eam_list *p;
  for(p = import_graph->nodes; ! eam_null(p); p = eam_cdr(p)){
    struct eam_import_graph_node *node = (struct eam_import_graph_node*)eam_car(p);
    assert(node->color != discovered);
    if(node->color == undiscovered)
      depth_first_visit(node, &reversed_sorted_nodes);
  } // for
  
  /* Return reversed_sorted_nodes reversed (i.e. straight), and destroy the
     original (i.e. the reversed version): */
  struct eam_list *sorted_nodes = eam_reverse(reversed_sorted_nodes);
  eam_destroy_list_structure(reversed_sorted_nodes);
  return sorted_nodes;
}

struct eam_import_graph* eam_create_import_graph(const struct eam_list *compilation_units,
					 const struct eam_symbol_table *symbol_table){
  /* Make an empty graph: */
  struct eam_import_graph *r =
    create_empty_import_graph(compilation_units, symbol_table);
  
  /* Fill it with all the known imports: */
  const struct eam_list *p = compilation_units;
  while(! eam_null(p)){
    /* Add each import of the first compilation unit: */
    struct eam_compilation_unit *first = (struct eam_compilation_unit*)eam_car(p);
    const struct eam_list *imported_by_first = first->imports;
    while(! eam_null(imported_by_first)){
      struct eam_unmangled_and_mangled *unmangled_and_mangled = eam_car(imported_by_first);
      add_import(r, first->name, unmangled_and_mangled->unmangled);
      imported_by_first = eam_cdr(imported_by_first);
    } // while
    
    /* Go on with the rest: */
    p = eam_cdr(p);
  } // while
  
  /* Topologically sort nodes and check that there are no cycles; replace the
     unsorted node list in the graph with the new sorted version: */
  struct eam_list *sorted_nodes = topologically_sort_nodes(r);
  eam_destroy_list_structure(r->nodes);
  r->nodes = sorted_nodes;
  /*   
  fprintf(stderr, "These are the nodes, topologically sorted: ");
  struct eam_list *q;
  for(q = sorted_nodes; ! eam_null(q); q = eam_cdr(q))
    fprintf(stderr, "%s ", node_to_name((struct eam_import_graph_node*)eam_car(q)));
  fprintf(stderr, "\n");
  */
  
 /* Return the graph we built: */
  return r;
}
