/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "symbol_table.h"
#include "../common/types.h"
#include "../common/command_line.h"
#include <string.h>
#include <assert.h>
#include "../common/malloc.h"
#include "../runtime/registers.h"

/* A single binding, to be stored in the linked list (see below): */
struct eam_binding{
  char *symbol_name; // a string allocated for the purpose
  struct eam_symbol_value symbol_value;
}; // struct

/* The data structure mapping names into symbol_values: */
struct eam_symbol_table{
  /* An hash table mapping strings into symbol_values: */
  struct eam_string_map *map;
  
  /* An hash table mapping a register name, without the "%" prefix, into
     a register number, stored as an integer_t. -1 means "memory
     pseudo-register". */
  struct eam_string_map *register_name_to_register;
  
  /* The register_name_to_register map reversed: an index is "mapped" into
     a string allocated for the purpose, without the "%" prefix. This doesn't
     include memory pseudo-registers. NULL means that a register isn't
     named. */
  char* register_to_register_name[EAM_REGISTERS_NO];
  
  /* List of the names, without the "%" prefix, of memory pseudo-registers.
     Each name is allocated for the purpose: */
  struct eam_list *memory_pseudoregisters;
  
  /* Ah hash table mapping compilation unit unmangled names into nothing really
     useful. This is useful as a quick way to check whether a compilation unit
     has already been included: */
  struct eam_string_map *included_compilation_units;
  
  /* The unmangled and mangled names of the included compilation units. The
     information could be extracted from included_compilation_units, but this
     is useful since symbol tables can be requested for an "iteratable" 
     structure holding this. The names held in the lists are newly-alloceted,
     not shared: */
  struct eam_list *included_compilation_units_unmangled_names;
  struct eam_list *included_compilation_units_mangled_names;
  
  /* Lists of malloc()ed strings of various sizes; elements are pointed to by
     some elements of struct eam_symbol_value. The simpler solution involving 
     copies of these strings in symbol_values is not viable because our hash
     table maps objects into *constant size* data. */
  struct eam_list *compilation_unit_unmangled_names;
  struct eam_list *compilation_unit_mangled_names;
  
  /* A (not sorted) linked list holding bindings for the exported symbols; this
     is useful for dumping symbol tables into files: */
  struct eam_list *exported_bindings;
  
  /* The largest currently-used id in the translation unit we're currently
     working on. This should be reset every time we include a new compilation
     unit: */
  int last_used_id;
}; // struct

static char* symbol_type_to_string(const enum eam_symbol_type type)
  __attribute__(( unused )); // only needed in verbose mode
static char* symbol_type_to_string(const enum eam_symbol_type type){
  switch(type){
  case eam_primitive_symbol_type: return "primitive";
  case eam_global_symbol_type:    return "global";
  case eam_label_symbol_type:     return "label";
  default:                    assert(0);
  } // switch
}

struct eam_symbol_table* eam_create_symbol_table(void){
  /* Allocate our struct: */
  struct eam_symbol_table *r = (struct eam_symbol_table*)
    eam_xmalloc(sizeof(struct eam_symbol_table));
  
  /* Fill it: */
  r->map = eam_create_string_map_of_size(sizeof(struct eam_symbol_value), 
				     65536, // there can be a lot of symbols
				     eam_default_string_map_datum_finalizer);
  r->included_compilation_units = eam_create_string_map(1 /* unused*/,
						    eam_default_string_map_datum_finalizer);
  r->compilation_unit_unmangled_names = eam_nil;
  r->compilation_unit_mangled_names = eam_nil;
  r->exported_bindings = eam_nil;
  r->included_compilation_units_unmangled_names = eam_nil;
  r->included_compilation_units_mangled_names = eam_nil;

  /* Fill register-naming structures: */
  r->register_name_to_register =
    eam_create_string_map_of_size(sizeof(integer_t), 
                                  EAM_REGISTERS_NO * 3,
                                  eam_default_string_map_datum_finalizer);
  int i;
  for(i = 0; i < EAM_REGISTERS_NO; i++)
    r->register_to_register_name[i] = NULL;
  r->memory_pseudoregisters = eam_nil;

  /* Return a pointer to our struct: */
  return r;
}

void eam_destroy_symbol_table(struct eam_symbol_table *symbol_table){
  /* Destroy the hash map: */
  eam_destroy_string_map(symbol_table->map);

  /* Destroy the included compilation unit hash map: */
  eam_destroy_string_map(symbol_table->included_compilation_units);
  
  /* Destroy the linked lists and all their elements. This is safe bacause all
     the elements of all the lists are individual malloced() objects: */
  eam_destroy_whole_list(symbol_table->exported_bindings);
  eam_destroy_whole_list(symbol_table->compilation_unit_unmangled_names);
  eam_destroy_whole_list(symbol_table->compilation_unit_mangled_names);
  eam_destroy_whole_list(symbol_table->included_compilation_units_unmangled_names);
  eam_destroy_whole_list(symbol_table->included_compilation_units_mangled_names);
  
  /* Destroy register-naming structures: */
  eam_destroy_string_map(symbol_table->register_name_to_register);
  int i;
  for(i = 0; i < EAM_REGISTERS_NO; i++)
    if(symbol_table->register_to_register_name[i] != NULL)
      free(symbol_table->register_to_register_name[i]);
  eam_destroy_whole_list(symbol_table->memory_pseudoregisters);
  
  /* Destroy our trivial struct: */
  free(symbol_table);
}

int eam_is_symbol_table_defined_on(const struct eam_symbol_table *table,
		  const char *symbol){
  return eam_is_string_map_defined_on(table->map, symbol);
}

struct eam_symbol_value* eam_lookup_symbol_table(const struct eam_symbol_table *table,
					 const char *symbol){
  struct eam_symbol_value *r = (struct eam_symbol_value*)
    eam_access_string_map(table->map, symbol);
  if(r == NULL){
    fprintf(stderr, "About the symbol \"%s\":\n", symbol);
    eam_fatal("could not find symbol in lookup");
  } // if
  else
    return r;
}

static void fill_symbol_table_from_register_naming(
   struct eam_symbol_table *table,
   struct eam_register_naming *naming){
  /* Check whether the name the user wants to give the register is already
     taken: */
  if(eam_is_string_map_defined_on(table->register_name_to_register,
                                  naming->name))
    eam_fatal("%%%s is being named twice", naming->name);
  
  /* Ok, add the naming. This can't fail if it's a memory pseudo-register,
     but if we are naming an indexed register we have to check it's not
     already named in some other way: */
  if(naming->definition->type == eam_memory_register_definition_type){
    char *copy_of_the_name = eam_xmalloc(strlen(naming->name) + 1);
    integer_t register_index = -1; // memory pseudo-register
    strcpy(copy_of_the_name, naming->name);
    eam_insert_into_string_map(table->register_name_to_register,
                               naming->name,
                               &register_index);
    table->memory_pseudoregisters =
      eam_cons(copy_of_the_name, table->memory_pseudoregisters);
  }
  else if(naming->definition->type == eam_indexed_register_register_definition_type){
    char *copy_of_the_name = eam_xmalloc(strlen(naming->name) + 1);
    if(naming->definition->register_no >= EAM_REGISTERS_NO)
      eam_fatal("referring the register %%r"INTEGER_T_FORMAT
                " (registers are only "INTEGER_T_FORMAT")",
                (integer_t)naming->definition->register_no,
                (integer_t)EAM_REGISTERS_NO
                );
    strcpy(copy_of_the_name, naming->name);
    eam_insert_into_string_map(table->register_name_to_register,
                               naming->name,
                               &(naming->definition->register_no));
    if(table->register_to_register_name[naming->definition->register_no] != NULL)
      eam_fatal("%%"INTEGER_T_FORMAT" is being named twice",
                (integer_t)naming->definition->register_no);
    table->register_to_register_name[naming->definition->register_no] =
      copy_of_the_name;
  }
  else
    assert(0);
}

static void fill_symbol_table_from_register_namings(
   struct eam_symbol_table *table,
   struct eam_list *namings){
  struct eam_list *p;
  for(p = namings; ! eam_null(p); p = eam_cdr(p))
    fill_symbol_table_from_register_naming(table, eam_car(p));
}

static struct eam_symbol_value* bind_in_symbol_table(struct eam_symbol_table *table,
						 const char *symbol,
						 const int exported,
						 const enum eam_symbol_type type,
						 const char *compilation_unit_unmangled_name,
						 const char *compilation_unit_mangled_name,
						 const int identifier){
  struct eam_symbol_value *current_value = eam_access_string_map(table->map, symbol);
  if(current_value != NULL){
    /* Do nothing if the current binding is *identical* to the new one: */
    if((current_value->type == type) &&
       (current_value->exported == exported) &&
       (current_value->identifier == identifier) &&
       (! strcmp(current_value->compilation_unit_unmangled_name,
		 compilation_unit_unmangled_name)) &&
       (! strcmp(current_value->compilation_unit_mangled_name,
		 compilation_unit_mangled_name))){
      fprintf(stderr, 
	      "WARNING: redundant bindings of \"%s\" from \"%s\" and \"%s\".\n", 
	      symbol,
	      current_value->compilation_unit_unmangled_name,
	      compilation_unit_unmangled_name);
      return current_value;
    } // if
    
    /* If we arrived here then there is a conflict: */
    fprintf(stderr, "Conflicting bindings of \"%s\" from \"%s\" and \"%s\":\n", 
	    symbol,
	    current_value->compilation_unit_unmangled_name,
	    compilation_unit_unmangled_name);
    eam_fatal("symbol bound twice");
  } // if
  else{
    /* Add compilation unit names to the lists; we don't care about duplication: */
    const size_t compilation_unit_unmangled_name_size =
      strlen(compilation_unit_unmangled_name);
    char *new_unmangled_name =
      (char*)eam_xmalloc(compilation_unit_unmangled_name_size + 1);
    strcpy(new_unmangled_name, compilation_unit_unmangled_name);
    table->compilation_unit_unmangled_names = 
      eam_cons(new_unmangled_name, table->compilation_unit_unmangled_names);
    const size_t compilation_unit_mangled_name_size =
      strlen(compilation_unit_mangled_name);
    char *new_mangled_name =
      (char*)eam_xmalloc(compilation_unit_mangled_name_size + 1);
    strcpy(new_mangled_name, compilation_unit_mangled_name);
    table->compilation_unit_mangled_names = 
      eam_cons(new_mangled_name, table->compilation_unit_mangled_names);

    /* Our temporary value, to be copied into the hash: */
    struct eam_symbol_value value;
    value.type = type;
    value.exported = exported;
    value.identifier = identifier;
    value.compilation_unit_unmangled_name = new_unmangled_name;
    value.compilation_unit_mangled_name = new_mangled_name;

    /* Insert the binding into the hash: */
    eam_insert_into_string_map(table->map, symbol, &value);
    
    /* Insert the binding also into the linked list, if it's to be exported: */
    if(exported){
      struct eam_binding *binding = // symbol_name is stored just after the struct
	(struct eam_binding*)eam_xmalloc(sizeof(struct eam_binding) + strlen(symbol) + 1);
      binding->symbol_name = ((char*)binding) + sizeof(struct eam_binding);
      strcpy(binding->symbol_name, symbol); // copy name
      binding->symbol_value = value;        // copy what the name is mapped into
      table->exported_bindings = eam_cons(binding, table->exported_bindings);
    } // if
    
    #ifdef VERBOSE_DEBUG
    fprintf(stderr, "Binding %s as a%s %s from %s, as %i\n",
            symbol,
            (exported ? "n exported" : " not exported"),
            symbol_type_to_string(type),
            compilation_unit_unmangled_name,
            value.identifier);
    #endif // #ifdef VERBOSE_DEBUG

    /* Return a pointer to the stored in-table binding, not to our
       temporary copy: */
    return (struct eam_symbol_value*)eam_access_string_map(table->map, symbol);
  } // else
}

static void fill_symbol_table_from_definition(struct eam_symbol_table* symbol_table,
					      const struct eam_definition *definition,
					      const struct eam_compilation_unit *compilation_unit,
					      const int only_exported){
  switch(definition->type){
  case eam_immediate_type_definition_type:{
    /* Do nothing. To do: think of this. */
    break;
  } // case
  case eam_primitive_definition_type:{
    /* We make a binding when the symbol is exported, or when we also accept
       non-exported symbols: */
    if(definition->primitive_definition->exported || ! only_exported){
      symbol_table->last_used_id++;
      bind_in_symbol_table(symbol_table,
			   definition->primitive_definition->name,
			   definition->primitive_definition->exported,
			   eam_primitive_symbol_type,
			   compilation_unit->name,
			 compilation_unit->mangled_name,
			   symbol_table->last_used_id);
    } // if
    break;
  } // case
  case eam_global_definition_type:{
    /* We make a binding when the symbol is exported, or when we also accept
       non-exported symbols: */
    if(definition->global_definition->exported || ! only_exported){
      symbol_table->last_used_id++;
      bind_in_symbol_table(symbol_table,
			   definition->global_definition->name,
			   definition->global_definition->exported,
			   eam_global_symbol_type,
			   compilation_unit->name,
			   compilation_unit->mangled_name,
			   symbol_table->last_used_id);
      break;
    } // if
  } // case
  } // switch
}

static void fill_symbol_table_from_definitions(struct eam_symbol_table* symbol_table,
					       const struct eam_list *definitions,
					       const struct eam_compilation_unit *compilation_unit,
					       const int only_exported){
  while(! eam_null(definitions)){
    fill_symbol_table_from_definition(symbol_table,
				      (struct eam_definition*)eam_car(definitions),
				      compilation_unit,
				      only_exported);
    definitions = eam_cdr(definitions);
  } // while
}

static void fill_symbol_table_from_command(
   struct eam_symbol_table* symbol_table,
   const struct eam_command *command,
   const struct eam_compilation_unit *compilation_unit,
   const int only_exported){
  /* We're only interested in labels... */
  if(command->type == eam_label_command_type)
    /* ...when they are exported, or when we also accept non-exported ones: */
    if(command->label_command.exported || ! only_exported){
      symbol_table->last_used_id++;
      bind_in_symbol_table(symbol_table,
			   command->label_command.label,
			   command->label_command.exported,
			   eam_label_symbol_type,
			   compilation_unit->name,
			   compilation_unit->mangled_name,
			   symbol_table->last_used_id);
    } // inner if
}

static void fill_symbol_table_from_commands(
   struct eam_symbol_table* symbol_table,
   const struct eam_list *commands,
   const struct eam_compilation_unit *compilation_unit,
   const int only_exported){
  while(! eam_null(commands)){
    fill_symbol_table_from_command(symbol_table,
				   (struct eam_command*)eam_car(commands),
				   compilation_unit,
				   only_exported);
    commands = eam_cdr(commands);
  } // while
}

void eam_fill_symbol_table_from_compilation_unit(
        struct eam_symbol_table *symbol_table,
	const struct eam_compilation_unit *compilation_unit,
	const int only_exported){
  char unused; // the object included compilation units names are mapped into
  
  /* Do nothing if we have already included the given compilation unit: */
  if(eam_is_string_map_defined_on(symbol_table->included_compilation_units,
			      compilation_unit->name))
    return;

  /* We're starting to work with a new compilation unit: reset the counter,
     so that the first symbol gets the first id: */
  symbol_table->last_used_id = 0;
  
  /* Note that imported compilation units are *not* processed here. */
  
  /* Add register namings, also if only using the interface: */
  fill_symbol_table_from_register_namings(symbol_table,
                                          compilation_unit->register_namings);
  
  /* Extract exported symbols from definitions: */
  fill_symbol_table_from_definitions(symbol_table,
				     compilation_unit->definitions,
				     compilation_unit,
				     only_exported);

  /* Extract label names from commands: */
  fill_symbol_table_from_commands(symbol_table,
				  compilation_unit->commands,
				  compilation_unit,
				  only_exported);
  
  /* Remember that we included the given compilation unit: */
  eam_insert_into_string_map(symbol_table->included_compilation_units,
			 compilation_unit->name,
			 &unused); // safe because this will not be referenced
  char *unmangled_name = (char*)eam_xmalloc(strlen(compilation_unit->name) + 1);
  strcpy(unmangled_name, compilation_unit->name);
  symbol_table->included_compilation_units_unmangled_names = 
    eam_cons(unmangled_name, symbol_table->included_compilation_units_unmangled_names);
  char *mangled_name = (char*)eam_xmalloc(strlen(compilation_unit->mangled_name) + 1);
  strcpy(mangled_name, compilation_unit->mangled_name);
  symbol_table->included_compilation_units_mangled_names = 
    eam_cons(mangled_name, symbol_table->included_compilation_units_mangled_names);
}



char* eam_lookup_symbol_table_into_string(const struct eam_symbol_table *symbol_table,
                                          const char *symbol){
  /* Pointer to the last-used buffer: */
  static char *string = NULL;
  
  /* Lookup the symbol: */
  const struct eam_symbol_value *symbol_value =
    eam_lookup_symbol_table(symbol_table, symbol);
  
  /* Compute (a majoration of) the size needed for the buffer, resize it and 
     fill it: */
  const size_t size = 
    strlen(symbol) +
    strlen(symbol_value->compilation_unit_mangled_name) +
    500;
  string = realloc(string, size); // this is correct also when string == NULL
  switch(symbol_value->type){
  case eam_primitive_symbol_type:{
    sprintf(string, "/*%s*/%s"EAM_PRIMITIVE_ROOT"%i",
	    symbol,
	    symbol_value->compilation_unit_mangled_name,
	    symbol_value->identifier);
    break;
  }
  case eam_global_symbol_type:{
    sprintf(string, "/*%s*/%s"EAM_GLOBAL_ROOT"%i",
	    symbol,
	    symbol_value->compilation_unit_mangled_name,
	    symbol_value->identifier);
    break;
  }
  case eam_label_symbol_type:{
    sprintf(string, "/*%s*/%s"EAM_LABEL_ROOT"%i",
	    symbol,
	    symbol_value->compilation_unit_mangled_name,
	    symbol_value->identifier);
    break;
  }
  default:
    assert(0);
  } // switch
  
  /* Return a pointer to the string we prepared: */
  return string;
}

const struct eam_list* eam_symbol_table_to_compilation_units(
                      const struct eam_symbol_table *symbol_table){
  return symbol_table->included_compilation_units_unmangled_names;
}

const struct eam_list* eam_symbol_table_to_mangled_compilation_units(
                      const struct eam_symbol_table *symbol_table){
  return symbol_table->included_compilation_units_mangled_names;
}

/* Return a nonzero value iff the eAM register with the given index has been
   named: */
integer_t eam_is_register_named_in_symbol_table(
   const struct eam_symbol_table *table,
   const integer_t index){
  assert(index >= 0);
  assert(index < EAM_REGISTERS_NO);
  return table->register_to_register_name[index] != NULL;
}

/* Return a nonzero value iff the given eAM register name is bound: */
integer_t eam_is_register_name_bound_in_symbol_table(
   const struct eam_symbol_table *symbol_table,
   const char *name){
  return eam_is_string_map_defined_on(symbol_table->register_name_to_register,
                                      name);
}

/* Return the index of the given named register, or fail with a fatal error
   if the register name isn't bound. -1 is returned for memory pseudo-registers.
   The "%" prefix must be omitted: */
integer_t eam_name_to_register_in_symbol_table(
   const struct eam_symbol_table *table,
   const char *name){
  if(! eam_is_string_map_defined_on(table->register_name_to_register, name))
    eam_fatal("the register name %%%s isn't bound", name);
  return *((integer_t*)
           (eam_access_string_map(table->register_name_to_register, name)));
}

/* Return the name of the given indexed register as a pointe to an internal
   string (not to be freed by the caller), or fail with a fatal error if
   the register name isn't bound. Fail with a fatal error if the given
   register isn't named. The "%" prefix is omitted: */
char* eam_register_to_name_in_symbol_table(
   const struct eam_symbol_table *table,
   const integer_t register_index){
  if(table->register_to_register_name[register_index] == NULL)
    eam_fatal("the register %%r"INTEGER_T_FORMAT" isn't named",
              register_index);
  return table->register_to_register_name[register_index];
}
