/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005 Luca Saiu
Updated in 2006 by Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef EAM_SYMBOL_TABLE_H_
#define EAM_SYMBOL_TABLE_H_

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "eaml.h"
#include "../common/list.h"
#include "../common/string_map.h"

/* A symbol table is the implementation of the mapping name |--> <id, type>;
   such a mapping is kept for each exported identifier. A separate symbol
   table is used for each compilation unit. */

/* To do: the entire concept of dumping and undumping symbol tables should be
   revisited. Repositories (not yet implemented) are a much better and more
   scalable solution. */

/* Mnemonic roots for generated symbols: they form part the name of each
   symbol in the generated C code: */
#define EAM_LONG_MNEMONICS
#ifdef EAM_LONG_MNEMONICS
  #define EAM_PRIMITIVE_ROOT  "_primitive_"
  #define EAM_LABEL_ROOT      "_label_"
  #define EAM_GLOBAL_ROOT     "_global_"
  #define EAM_INITIALIZE_ROOT "initialize"
  #define FINALIZE_ROOT       "finalize"
  #define EAM_PINS_NO_ROOT    "_pins_no"
#else
  #define EAM_PRIMITIVE_ROOT  "p"
  #define EAM_LABEL_ROOT      "l"
  #define EAM_GLOBAL_ROOT     "g"
  #define EAM_INITIALIZE_ROOT "i"
  #define FINALIZE_ROOT   "f"
  #define EAM_PINS_NO_ROOT    "q"
#endif // #define EAM_LONG_MNEMONICS

/* The "sort" of referred object: */
enum eam_symbol_type{
  eam_primitive_symbol_type,
  eam_global_symbol_type,
  eam_label_symbol_type,
}; // enum

/* The object type each symbol is mapped into: */
struct eam_symbol_value{ // the type which symbol tables map identifiers into
  enum eam_symbol_type type;
  int exported; // a "boolean" with a nonzero value iff the symbol is exported
  int identifier; // a unique (up to the compilation unit) name
  char *compilation_unit_unmangled_name; // pointer to a shared string
  char *compilation_unit_mangled_name;   // pointer to a shared string
}; // struct

/* The data structure mapping names into symbol_values, as an abstract data
   type: */
struct eam_symbol_table;

/* Make a new empty symbol table: */
struct eam_symbol_table* eam_create_symbol_table(void);

/* Add information from the given compilation unit to the given symbol table;
   if only_exported is nonzero then only exported symbols are added. Do
   nothing if the given compilation unit is already included in the given
   symbol table. */
void eam_fill_symbol_table_from_compilation_unit(
        struct eam_symbol_table *symbol_table,
	const struct eam_compilation_unit *compilation_unit,
	const int only_exported);

/* Deallocate a symbol table: */
void eam_destroy_symbol_table(struct eam_symbol_table *symbol_table);

/* Return a nonzero value iff the given symbol is bound: */
int eam_is_symbol_table_defined_on(const struct eam_symbol_table *table,
                                   const char *symbol)
  __attribute__(( pure ));

/* Return a reference to an internal list (which should not be altered or
   destroyed) holding the unmangled names of all included compilation units: */
const struct eam_list* eam_symbol_table_to_compilation_units(
   const struct eam_symbol_table *symbol_table)
  __attribute__(( pure ));

/* Return a reference to an internal list (which should not be altered or
   destroyed) holding the mangled names of all included compilation units: */
const struct eam_list* eam_symbol_table_to_mangled_compilation_units(
   const struct eam_symbol_table *symbol_table)
  __attribute__(( pure ));

/* Return the binding for the given symbol, or fail with a fatal error if the
   symbol is unbound: */
struct eam_symbol_value* eam_lookup_symbol_table(
   const struct eam_symbol_table *table,
   const char *symbol)
  __attribute__(( pure ));

/* Return a pointer to an internally allocated string representing the identifier
   into the given symbol should be output in C files. Each returned string is
   destroyed at the next call. Obviously non-reentrant, but nice to have: */
char* eam_lookup_symbol_table_into_string(
   const struct eam_symbol_table *symbol_table,
   const char *symbol)
  __attribute__(( pure ));

/* Return a nonzero value iff the eAM register with the given index has been
   named: */
integer_t eam_is_register_named_in_symbol_table(
   const struct eam_symbol_table *symbol_table,
   const integer_t index)
  __attribute__(( pure ));

/* Return a nonzero value iff the given eAM register name is bound: */
integer_t eam_is_register_name_bound_in_symbol_table(
   const struct eam_symbol_table *symbol_table,
   const char *name)
  __attribute__(( pure ));

/* Return the index of the given named register, or fail with a fatal error
   if the register name isn't bound. -1 is returned for memory pseudo-registers.
   The "%" prefix must be omitted: */
integer_t eam_name_to_register_in_symbol_table(
   const struct eam_symbol_table *symbol_table,
   const char *name)
  __attribute__(( pure ));

/* Return the name of the given indexed register as a pointe to an internal
   string (not to be freed by the caller), or fail with a fatal error if
   the register name isn't bound. Fail with a fatal error if the given
   register isn't named. The "%" prefix is omitted: */
char* eam_register_to_name_in_symbol_table(
   const struct eam_symbol_table *symbol_table,
   const integer_t register_index)
  __attribute__(( pure ));

#endif // #ifndef EAM_SYMBOL_TABLE_H_
